#include "png.h"

#include "../codec.h"
#include "mbuf.h"

#include "libpng.h"

//========================================================================================
//
// Compressor
//
//========================================================================================
typedef struct {
	png_structp png_ptr;
	png_infop info_ptr;
	size_t streamSize;
	mbuf_t* streamBufHead;
	mbuf_t* streamBufTail;
	codec_mem_t m_cm;
} tCContext;

//----------------------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
png_voidp user_malloc_fn(png_structp png_ptr, png_size_t size)
{
	return (png_voidp)codec_mem_alloc((codec_mem_t*)NULL, size);
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
void user_free_fn(png_structp png_ptr, png_voidp ptr)
{
	codec_mem_free((codec_mem_t*)NULL, ptr);
}

//----------------------------------------------------------------------------------------
//
// error/warning callback and so on
//
//----------------------------------------------------------------------------------------
static void user_error_fn(png_structp png_ptr, png_const_charp error_msg)
{
	tCContext* ctx = (tCContext*)png_ptr;

	throw expCODEC(const_cast<char*>(error_msg));

}
static void user_warning_fn(png_structp png_ptr, png_const_charp waning_msg)
{
}

//----------------------------------------------------------------------------------------
//
// write callback and so on
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void init_outbuf(tCContext* ctx)
{
	ctx->streamSize = 0;
	ctx->streamBufTail = ctx->streamBufHead = mbuf_malloc();
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void user_write_data(png_structp png_ptr, png_bytep data, png_size_t length)
{
	tCContext* ctx = (tCContext*)png_get_io_ptr(png_ptr);

	mbuf_t* mbuf = ctx->streamBufTail;
	while (length > 0) {
		png_size_t count = sizeof(mbuf->buf) - mbuf->size;
		if (count >= length) {
			memcpy(mbuf->buf + mbuf->size, data, length);
			mbuf->size += length;
			ctx->streamSize += length;
			return;
		}
		memcpy(mbuf->buf + mbuf->size, data, count);
		mbuf->size += count;
		data += count;
		length -= count;
		ctx->streamSize += count;

		mbuf_t* new_buf = mbuf_malloc();
		new_buf->next = (mbuf_t*)NULL;
		new_buf->size = 0;
		mbuf->next = new_buf;
		mbuf = ctx->streamBufTail = new_buf;
	}
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void user_flush_data(png_structp png_ptr)
{
}

//========================================================================================
//
//========================================================================================
size_t cpng_ctx_size(void)
{
	return sizeof(tCContext);
}

//========================================================================================
//
//========================================================================================
void init_cpng(void* vctx)
{
	tCContext* ctx = (tCContext*)vctx;

	png_structp png_ptr = ctx->png_ptr = png_create_write_struct_2(PNG_LIBPNG_VER_STRING, (png_voidp)NULL, (png_error_ptr)NULL, (png_error_ptr)NULL, (png_voidp)ctx, user_malloc_fn, user_free_fn);
	png_infop info_ptr = ctx->info_ptr = png_create_info_struct(png_ptr);

	init_outbuf(ctx);
	png_set_write_fn(png_ptr, (png_voidp)ctx, user_write_data, user_flush_data);
	png_set_error_fn(png_ptr, (png_voidp)ctx, user_error_fn, user_warning_fn);
	png_set_filter(png_ptr, 0, PNG_FILTER_NONE);
}

//========================================================================================
//
//========================================================================================
void setup_cpng(void* vctx, const unsigned short width, const unsigned short height, const int colorType, const int comp_level)
{
	tCContext* ctx = (tCContext*)vctx;
	png_structp png_ptr = ctx->png_ptr;
	png_infop info_ptr = ctx->info_ptr;

	png_set_IHDR(png_ptr, info_ptr, width, height, colorType == 2 ? 16 : 8, colorType == 1 ? PNG_COLOR_TYPE_RGB : PNG_COLOR_TYPE_GRAY, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

	if (colorType != 2) {
		png_color_8 sig_bit;
		sig_bit.gray = sig_bit.red = sig_bit.green = sig_bit.blue = sig_bit.alpha = 8;
		png_set_sBIT(png_ptr, info_ptr, &sig_bit);
	}

	if (comp_level >= 0)
		png_set_compression_level(png_ptr, comp_level);

	png_write_info(png_ptr, info_ptr);
}

//========================================================================================
//
//========================================================================================
void proc_cpng(void* vctx, void* row_pointer, unsigned int rows)
{
	tCContext* ctx = (tCContext*)vctx;
	png_structp png_ptr = ctx->png_ptr;
	png_infop info_ptr = ctx->info_ptr;

	png_write_rows(png_ptr, (png_bytepp)row_pointer, rows);
}

//========================================================================================
//
//========================================================================================
size_t finish_cpng(void* vctx)
{
	tCContext* ctx = (tCContext*)vctx;
	png_structp png_ptr = ctx->png_ptr;
	png_infop info_ptr = ctx->info_ptr;

	png_write_end(png_ptr, info_ptr);

	return ctx->streamSize;
}
//========================================================================================
//
//========================================================================================
void flush_cpng(void* vctx, void* pECB)
{
	tCContext* ctx = (tCContext*)vctx;

	mbuf_t* mbuf = ctx->streamBufHead;
	while (mbuf != NULL) {
		codec_write_client(pECB, mbuf->buf, mbuf->size);
		mbuf_t* next = mbuf->next;
		mbuf_free(mbuf);
		mbuf = next;
	}
}

//========================================================================================
//
//========================================================================================
void destroy_cpng(void* vctx)
{
	tCContext* ctx = (tCContext*)vctx;
	png_structp png_ptr = ctx->png_ptr;
	png_infop info_ptr = ctx->info_ptr;

	png_destroy_write_struct(&png_ptr, &info_ptr);
}

//========================================================================================
//
// Decompressor
//
//========================================================================================

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
typedef struct {
	png_structp png_ptr;
	png_infop info_ptr;
	png_bytep dataPtr;
	png_bytep curPtr;
	size_t dataSize;
} tDContext;

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void user_read_fn(png_structp png_ptr, png_bytep buf, png_size_t size)
{
	tDContext* ctx = (tDContext*)png_get_io_ptr(png_ptr);

	CopyMemory(buf, ctx->curPtr, size);
	ctx->curPtr += size;
}


//========================================================================================
//
//========================================================================================
size_t dpng_ctx_size(void)
{
	return sizeof(tDContext);
}
//========================================================================================
//
//========================================================================================
void init_dpng(void* vctx, f_decode_input_func input_func, void* input_ctx) throw (expCODEC)
{
	tDContext* ctx = (tDContext*)vctx;

	ctx->png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, (void*)ctx, user_error_fn, user_warning_fn);
	if (ctx->png_ptr == NULL) {
		throw expCODEC("png_create_read_struct() failed");
	}

	ctx->info_ptr = png_create_info_struct(ctx->png_ptr);
	if (ctx->info_ptr == NULL) {
		png_destroy_read_struct(&ctx->png_ptr, (png_infopp)NULL, (png_infopp)NULL);
		throw expCODEC("png_create_info_struct() failed");
	}

	input_func(input_ctx, 0, (const void**)&ctx->dataPtr, &ctx->dataSize, 0);
	ctx->curPtr = ctx->dataPtr;

	png_set_read_fn(ctx->png_ptr, (void*)ctx, user_read_fn);
	
}
//========================================================================================
//
//========================================================================================
void setup_dpng(void* vctx, unsigned short scale_denom, unsigned short* width, unsigned short* height) throw (expCODEC)
{
	tDContext* ctx = (tDContext*)vctx;

	png_read_info(ctx->png_ptr, ctx->info_ptr);

	png_uint_32 pWidth, pHeight;
	int bit_depth, color_type, interlace_type;

	png_get_IHDR(ctx->png_ptr, ctx->info_ptr, &pWidth, &pHeight, &bit_depth, &color_type, &interlace_type, (int*)NULL, (int*)NULL);

	*width = pWidth;
	*height = pHeight;
}

//========================================================================================
//
//========================================================================================
int continue_dpng(void* vctx) throw (expCODEC)
{
	tDContext* ctx = (tDContext*)vctx;

	return ctx->png_ptr->height - ctx->png_ptr->row_number;
//   height of image in pixels - number of rows in current pass 
}
//========================================================================================
//
//========================================================================================
void proc_dpng(void* vctx, void* row_pointer, unsigned int rows) throw (expCODEC)
{
	tDContext* ctx = (tDContext*)vctx;

	png_read_rows(ctx->png_ptr, (png_bytepp)NULL, (unsigned char**)row_pointer, rows);
}
//========================================================================================
//
//========================================================================================
void finish_dpng(void* vctx) throw (expCODEC)
{
	tDContext* ctx = (tDContext*)vctx;

	png_read_end(ctx->png_ptr, ctx->info_ptr);
	png_destroy_read_struct(&ctx->png_ptr, &ctx->info_ptr, (png_infopp)NULL);
}

