
/* exportStream
**
** Purpose:
**      Export a DICOM object into the stream format suitable
**      for network transmission or disk storage.
**
** Parameter Dictionary:
**	callerObject		Handle to caller's DICOM object
**	opt			Bit mask giving options for exporting data
**	buffer			Pointer to caller's buffer to hold next chunk
**				of DCM stream data
**	bufferlength		Length of caller's buffer to hold stream data
**	callback		Callback routine to be called.
**	ctx			Pointer to context variable we maintain to keep
**				track of our location in export process.
**	sequenceLevel		Current level in the sequence hierarchy
**
** Return Values:
**
**	DCM_FILEACCESSERROR
**	DCM_ILLEGALOBJECT
**	DCM_LISTFAILURE
**	DCM_NORMAL
**	DCM_NULLOBJECT
**	DCM_CALLBACKABORTED
**
** Notes:
**
** Algorithm:
**	Description of the algorithm (optional) and any other notes.
*/

static CONDITION
exportStream(DCM_OBJECT ** callerObject, unsigned long opt,
	     void *buffer, U32 bufferlength, CONDITION(*callback) (),
	     void *ctx, int sequenceLevel)
{
	PRIVATE_OBJECT** object;
	PRV_GROUP_ITEM* groupItem;
	PRV_ELEMENT_ITEM* elementItem;
	DCM_ELEMENT element;
	int byteOrder;
	int lastFlag = 0;
	unsigned char *src, *dst;
	U32 c, bytesExported = 0, rtnLength, remainingData, exportLength;
	CONDITION cond;
	DCM_SEQUENCE_ITEM* sequenceItem;
	DCM_ELEMENT itemMarker = {
			DCM_DLMITEM, DCM_DLM, "", 1, DCM_UNSPECIFIEDLENGTH, NULL
		},
		itemDelimiter = {
			DCM_DLMITEMDELIMITATIONITEM, DCM_DLM, "", 1, 0, NULL
		},
		sequenceDelimiter = {
			DCM_DLMSEQUENCEDELIMITATIONITEM, DCM_DLM, "", 1, 0, NULL
		};
	CTNBOOLEAN unspecifiedSQLength = FALSE,
		explicitVR = FALSE,
		part10File = FALSE;
	unsigned long fileOptions = 0;
	unsigned long argOptions = opt & DCM_FILEFORMATMASK;

	object = (PRIVATE_OBJECT **) callerObject;
	cond = checkObject(object, "exportStream");
	if (cond != DCM_NORMAL)
		return cond;

	if ((opt & DCM_FILEFORMATMASK) == DCM_PART10FILE) {
		part10File = TRUE;
		opt &= ~DCM_ORDERMASK;
		opt |= DCM_EXPLICITLITTLEENDIAN;
		cond = setFileOptions(callerObject, &fileOptions);
		if (cond != DCM_NORMAL)
			return cond;
	}
	if ((opt & DCM_ORDERMASK) == 0)
		return COND_PushCondition(DCM_ILLEGALOPTION,
			DCM_Message(DCM_ILLEGALOPTION), "Byte order",
			"exportStream");

	switch (opt & DCM_ORDERMASK) {
	case DCM_ORDERNATIVE:
		byteOrder = NATIVE_ORDER;
		break;
	case DCM_ORDERLITTLEENDIAN:
		byteOrder = LITTLE_ORDER;
		break;
	case DCM_EXPLICITLITTLEENDIAN:
		byteOrder = LITTLE_ORDER;
		explicitVR = TRUE;
		break;
	case DCM_ORDERBIGENDIAN:
		byteOrder = BIG_ORDER;
		break;
	case DCM_EXPLICITBIGENDIAN:
		byteOrder = BIG_ORDER;
		explicitVR = TRUE;
		break;
	default:
		byteOrder = LITTLE_ORDER;
		break;
	}

	/*  We are making this step mandatory for now (smm)*/

	opt &= ~DCM_SEQUENCELENGTHMASK;
	opt |= DCM_UNSPECIFIEDLENGTHFLAG;

	/*  End of something that is out of place */

	if ((opt & DCM_SEQUENCELENGTHMASK) == DCM_UNSPECIFIEDLENGTHFLAG)
		unspecifiedSQLength = TRUE;

	dst = (unsigned char *) buffer;
	c = bufferlength;

	if (part10File) {
		cond = exportPreamble(object, dst, c, &rtnLength);
		if (cond != DCM_NORMAL)
			return cond;

		dst += rtnLength;
		c -= rtnLength;
		bytesExported += rtnLength;
	}
	if (sequenceLevel != 0) {
		if (!unspecifiedSQLength)
			itemMarker.length = (*object)->objectSize;
		exportFixedFields(&itemMarker, dst, bufferlength, byteOrder,
		explicitVR, &rtnLength);
		dst += rtnLength;
		c -= rtnLength;
		bytesExported += rtnLength;
	}
	groupItem = LST_Head(&(*object)->groupList);

	/*  Take this code out to allow empty groups. */
#if 0
	if (groupItem == NULL)
		return COND_PushCondition(DCM_LISTFAILURE,
		DCM_Message(DCM_LISTFAILURE), "exportStream");
#endif
	if (groupItem != NULL)
		(void) LST_Position(&(*object)->groupList, groupItem);

	while (groupItem != NULL) {
		if (part10File && groupItem->group != DCM_GROUPFILEMETA) {
#ifdef _scsb_ext_
			if (fileOptions != argOptions) {
			}
#else
			if (opt != fileOptions) {
				opt = fileOptions;
				cond = extractFileOptions(opt, &part10File,
				&explicitVR, &byteOrder);
				if (cond != DCM_NORMAL)
					return cond;
			}
#endif
		}
		elementItem = LST_Head(&groupItem->elementList);
		if (elementItem != NULL)
			(void) LST_Position(&groupItem->elementList, elementItem);
		if (DCM_TAG_ELEMENT(elementItem->element.tag) == 0x0000) {
#ifdef _scsb_ext_
			if (groupItem->group == 0x0000 || groupItem->group == 0x0002) {
				*elementItem->element.d.ul = computeGroupLength(groupItem, explicitVR);
			} else if (!dcm_group_length_enabled) {
				elementItem = LST_Next(&groupItem->elementList);
			}
#else
			U32 l;
			l = computeGroupLength(groupItem, explicitVR);
			*elementItem->element.d.ul = l;

			/* We have some problems computing group length for groups with sequences.
			** For now, just remove this attribute, except for group 0000 and 0002.
			*/
			if (groupItem->group != 0x0000 && groupItem->group != 0x0002)
				elementItem = LST_Next(&groupItem->elementList);
#endif
		}
		while (elementItem != NULL) {
			if (elementItem->element.tag == DCM_MAKETAG(0x003a, 0x01000)) {
				printf("Found waveform for export\n");
			}
			if (c <= 20) {
				cond = callback(buffer, bytesExported, 0, ctx);
				if (cond != DCM_NORMAL)
					return COND_PushCondition(DCM_CALLBACKABORTED,
						DCM_Message(DCM_CALLBACKABORTED), "exportStream");

				bytesExported = 0;
				c = bufferlength;
				dst = (unsigned char *) buffer;
			}
			element = elementItem->element;
			if (element.representation == DCM_SQ) {
				if (unspecifiedSQLength)
					element.length = DCM_UNSPECIFIEDLENGTH;

				exportFixedFields(&element, dst, bufferlength, byteOrder,
				explicitVR, &rtnLength);
				} else {
				element.length = elementItem->paddedDataLength;
				exportFixedFields(&element, dst, bufferlength, byteOrder,
					explicitVR, &rtnLength);
			}
			dst += rtnLength;
			c -= rtnLength;
			bytesExported += rtnLength;

			remainingData = element.length;
			src = element.d.ot;
			elementItem->currentOffset = elementItem->dataOffset;

			if (element.representation == DCM_SQ) {

				cond = callback(buffer, bytesExported, 0, ctx);
				if (cond != DCM_NORMAL)
					return COND_PushCondition(DCM_CALLBACKABORTED,
						DCM_Message(DCM_CALLBACKABORTED), "exportStream");

				bytesExported = 0;
				c = bufferlength;
				dst = (unsigned char *) buffer;

				if (element.d.sq != NULL) {
					sequenceItem = LST_Head(&element.d.sq);
					if (sequenceItem != NULL)
						(void) LST_Position(&element.d.sq, sequenceItem);
					while (sequenceItem != NULL) {
						cond = exportStream(&sequenceItem->object, opt,
							buffer, bufferlength, callback, ctx,
							sequenceLevel + 1);
						if (cond != DCM_NORMAL)
							return cond;
						sequenceItem = LST_Next(&element.d.sq);
					}
				}
				if (element.length == DCM_UNSPECIFIEDLENGTH) {
					sequenceDelimiter.length = 0;
					exportFixedFields(&sequenceDelimiter, dst, bufferlength,
						byteOrder, explicitVR, &rtnLength);
					dst += rtnLength;
					c -= rtnLength;
					bytesExported += rtnLength;
				}
			} else {
				while (remainingData > 0) {
					if (debug)
						fprintf(stderr, "Export: (%08x) %d\n",
							element.tag, element.length);
					if (element.d.ot != NULL)
						remainingData = element.length -
							(src - ((unsigned char *) element.d.ot));
					else
						remainingData = element.length -
							(elementItem->currentOffset - elementItem->dataOffset);

					exportLength = (remainingData < c) ? remainingData : c;
					cond = exportData(object, elementItem, src, dst,
						exportLength, byteOrder, &rtnLength);
					if (cond != DCM_NORMAL)
						return cond;

					src += rtnLength;
					dst += rtnLength;
					bytesExported += rtnLength;
					c -= rtnLength;

					if (c <= 20) {
						cond = callback(buffer, bytesExported, 0, ctx);
						if (cond != DCM_NORMAL)
							return COND_PushCondition(DCM_CALLBACKABORTED,
							      DCM_Message(DCM_CALLBACKABORTED), "exportStream");

						bytesExported = 0;
						c = bufferlength;
						dst = (unsigned char *) buffer;
					}
				}
			}
			elementItem = LST_Next(&groupItem->elementList);
		}
		groupItem = LST_Next(&(*object)->groupList);
	}

	if ((sequenceLevel != 0) && unspecifiedSQLength) {
		if (c <= 20) {
			cond = callback(buffer, bytesExported, 0, ctx);
			if (cond != DCM_NORMAL)
				return COND_PushCondition(DCM_CALLBACKABORTED,
					DCM_Message(DCM_CALLBACKABORTED), "exportStream");

			bytesExported = 0;
			c = bufferlength;
			dst = (unsigned char *) buffer;
		}
		exportFixedFields(&itemDelimiter, dst, bufferlength, byteOrder,
		explicitVR, &rtnLength);
		dst += rtnLength;
		c -= rtnLength;
		bytesExported += rtnLength;
	}
	lastFlag = (sequenceLevel == 0) ? 1 : 0;
	cond = callback(buffer, bytesExported, lastFlag, ctx);
	if (cond != DCM_NORMAL)
		return COND_PushCondition(DCM_CALLBACKABORTED,
			DCM_Message(DCM_CALLBACKABORTED), "exportStream");

	return DCM_NORMAL;
}
