/*
          Copyright (C) 1993, 1994, RSNA and Washington University

          The software and supporting documentation for the Radiological
          Society of North America (RSNA) 1993, 1994 Digital Imaging and
          Communications in Medicine (DICOM) Demonstration were developed
          at the
                  Electronic Radiology Laboratory
                  Mallinckrodt Institute of Radiology
                  Washington University School of Medicine
                  510 S. Kingshighway Blvd.
                  St. Louis, MO 63110
          as part of the 1993, 1994 DICOM Central Test Node project for, and
          under contract with, the Radiological Society of North America.

          THIS SOFTWARE IS MADE AVAILABLE, AS IS, AND NEITHER RSNA NOR
          WASHINGTON UNIVERSITY MAKE ANY WARRANTY ABOUT THE SOFTWARE, ITS
          PERFORMANCE, ITS MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
          USE, FREEDOM FROM ANY COMPUTER DISEASES OR ITS CONFORMITY TO ANY
          SPECIFICATION. THE ENTIRE RISK AS TO QUALITY AND PERFORMANCE OF
          THE SOFTWARE IS WITH THE USER.

          Copyright of the software and supporting documentation is
          jointly owned by RSNA and Washington University, and free access
          is hereby granted as a license to use this software, copy this
          software and prepare derivative works based upon this software.
          However, any distribution of this software source code or
          supporting documentation or derivative works (source code and
          supporting documentation) must include the three paragraphs of
          the copyright notice.
*/
/* Copyright marker.  Copyright will be inserted above.  Do not remove */

/*
**				DICOM 93
**		     Electronic Radiology Laboratory
**		   Mallinckrodt Institute of Radiology
**		Washington University School of Medicine
**
** Module Name(s):	DCM_OpenFile
**			DCM_CreateObject
**			DCM_CloseObject
**			DCM_AddElement
**			DCM_RemoveElement
**			DCM_GetElementValue
**			DCM_GetElementSize
**			DCM_ScanParseObject
**			DCM_ImportStream
**			DCM_ExportStream
**			DCM_GetObjectSize
**			DCM_DumpElements
**			DCM_Debug
**			DCM_WriteFile
**			DCM_ModifyElements
**			DCM_ParseObject
**			DCM_RemoveGroup
**			DCM_GetSequenceList
**			DCM_ComputeExportSize
**	private functions
**			newElementItem
**			findCreateGroup
**			insertNewElement
**			updateObjectType
**			updateSpecialElements
**			exportFixedFields
**			exportData
**			fileSize
**			swapInPlace
**			checkObject
**			writeFile
**			countBytes
**			exportStream
**			verifyFormat
**			readFile
** Author, Date:	Stephen M. Moore, 26-Apr-93
** Intent:
**	This file contains the routines which implement a facility for
**	manipulating DICOM V3 information objects.  These routines parse
**	and construct NEMA objects (or streams).  Users are able to
**	read/write/examine individual attributes.  The package uses those
**	individual elements to construct an internal memory representation of
**	an object.  This representation is a linked list of the individual
**	attributes.  The user of the package can add attributes to an object,
**	remove an element from an object, query the object about an attribute,
**	and convert the object to and from its "stream" representation.
**	In addition, the package can parse a file which contains a stream
**	and create its internal object.
** Last Update:		$Author: scsb $, $Date: 2002/10/15 19:37:44 $
** Source File:		$RCSfile: readFile.c,v $
** Revision:		$Revision: 1.1.1.1 $
** Status:		$State: Exp $
*/

static char rcsid[] = "$Revision: 1.1.1.1 $ $RCSfile: readFile.c,v $";

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#ifndef MACOS
#include <stdlib.h>
#endif
#ifdef MALLOC_DEBUG
#include "malloc.h"
#endif
#ifdef GCCSUNOS
#include <unistd.h>
#endif

#ifdef MACOS
#include <files.h>
#elif defined _MSC_VER
#include <io.h>
#include <fcntl.h>
#else
#include <sys/file.h>
#endif

#ifdef MACOS
#include <fcntl.h>
#elif defined _MSC_VER
#include <fcntl.h>
#elif defined SOLARIS
#include <sys/fcntl.h>
#endif

#include <sys/types.h>
#ifdef MACOS
#include <stat.h>
#else
#include <sys/stat.h>
#endif
#include "dicom.h"
#include "condition.h"
#include "lst.h"
#include "dicom_uids.h"
#include "dicom_objects.h"
#include "dcmprivate.h"

#define BYTEORDER_SAME			1
#define BYTEORDER_REVERSE		2
#define	NATIVE_ORDER			BYTEORDER_SAME

#ifdef BIG_ENDIAN_ARCHITECTURE
#define	LITTLE_ORDER	BYTEORDER_REVERSE
#define	BIG_ORDER	BYTEORDER_SAME
#endif

#ifdef LITTLE_ENDIAN_ARCHITECTURE
#define	LITTLE_ORDER	BYTEORDER_SAME
#define	BIG_ORDER	BYTEORDER_REVERSE
#endif

static CTNBOOLEAN debug = FALSE;/* Flag for debugging messages to stdout */

/* readFile
**
** Purpose:
**	Read DICOM object from a file
**
** Parameter Dictionary:
**	name			Name of the file
**	callerBuf		Buffer from which to read the object
**	fd			File descriptor
**	size			Size of the file
**	fileOffset		Offset in the file from which point read starts
**	recursionLevel		Level of recursion
**	opt			Indicates in what byte order to read
**	callerObject		The object into which the contents are stored
**	scannedLength		Length of data scanned
**	remainOpenFlag		Indicates whether the file remains open
**
** Return Values:
**
**	DCM_ELEMENTCREATEFAILED
**	DCM_ELEMENTLENGTHERROR
**	DCM_ELEMENTOUTOFORDER
**	DCM_FILEACCESSERROR
**	DCM_ILLEGALSTREAMLENGTH
**	DCM_LISTFAILURE
**	DCM_NORMAL
**	DCM_OBJECTCREATEFAILED
**	DCM_UNEVENELEMENTLENGTH
**
** Notes:
**
** Algorithm:
**	Description of the algorithm (optional) and any other notes.
*/
static CONDITION
readFile(char *name, unsigned char *callerBuf, int fd, long size,
	 off_t fileOffset, int recursionLevel,
	 unsigned long opt, DCM_OBJECT ** callerObject,
	 U32 * scannedLength, CTNBOOLEAN * remainOpenFlag,
	 void *ctx,
	 CONDITION(*rd) (void *ctx, void *buf, int toRead, int *bytesRead),
	 CONDITION(*sk) (void *ctx, int offset, int flag))
	{
	CONDITION cond;
	int byteOrder;
	long lastGroup = -1,
	     lastElement = -1;
	U32 sequenceLength,
	    localLength;
	PRIVATE_OBJECT** object;
	PRV_GROUP_ITEM* groupItem = NULL;
	unsigned short group,
	               element,
	               tagGroup,
	               tagElement;
	DCM_ELEMENT e,
	            tagE;
	CTNBOOLEAN pixelFlag,
	           convertFlag = FALSE,
	           done = FALSE,
	           knownLength = TRUE,
	           sequenceDone = FALSE,
	           createGroupFlag,
	           explicitVR = FALSE;
	unsigned char buf[8],
	              *ptr;
	int nBytes;
	PRV_ELEMENT_ITEM* elementItem = NULL;
	DCM_OBJECT* sequenceObject;
	DCM_SEQUENCE_ITEM* sequenceItem;
	CTNBOOLEAN fileFlag = TRUE;

	if (callerBuf != NULL) {
		ptr = callerBuf;
		fileFlag = FALSE;
	} else
		ptr = buf;

	switch (opt & DCM_ORDERMASK) {
	case DCM_ORDERNATIVE:
		byteOrder = NATIVE_ORDER;
		break;
	case DCM_ORDERLITTLEENDIAN:
		byteOrder = LITTLE_ORDER;
		break;
	case DCM_EXPLICITLITTLEENDIAN:
		byteOrder = LITTLE_ORDER;
		explicitVR = TRUE;
		break;
	case DCM_ORDERBIGENDIAN:
		byteOrder = BIG_ORDER;
		break;
	case DCM_EXPLICITBIGENDIAN:
		byteOrder = BIG_ORDER;
		explicitVR = TRUE;
		break;
	default:
		byteOrder = NATIVE_ORDER;
		break;
	}
	if ((opt & DCM_CONVERTMASK) == DCM_FORMATCONVERSION)
		convertFlag = TRUE;

	if (scannedLength != NULL)
		*scannedLength = 0;

	cond = DCM_CreateObject(callerObject, opt);
	if (cond != DCM_NORMAL)
		return cond;

	object = (PRIVATE_OBJECT**)callerObject;
	if (fileFlag)
		strcpy((*object)->fileName, name);

	(*object)->fd = -1;
	(*object)->rd = rd;
	(*object)->sk = sk;
	(*object)->userCtx = ctx;
	if (size == (long) DCM_UNSPECIFIEDLENGTH)
		knownLength = FALSE;

	if ((fileFlag) && ((opt & DCM_DELETEMASK) == DCM_DELETEONCLOSE) && (recursionLevel == 0))
		(*object)->deleteFlag = TRUE;

	if (knownLength && (size == 0))
		done = TRUE;

	while (!done) {

		if ((size < 8) && knownLength) {
			if (debug)
				(void) DCM_DumpElements(callerObject, 0);
			(void) DCM_CloseObject(callerObject);
			return COND_PushCondition(DCM_ILLEGALSTREAMLENGTH, DCM_Message(DCM_ILLEGALSTREAMLENGTH), size, "readFile");
		}
		if (fileFlag) {
			if (fd != -1) {
				nBytes = read(fd, buf, 4);
			} else {
				cond = (*object)->rd((*object)->userCtx, buf, 4, &nBytes);
			}

			if (nBytes != 4)
				return COND_PushCondition(DCM_FILEACCESSERROR, DCM_Message(DCM_FILEACCESSERROR), name,"readFile");
			ptr = buf;
		}
		if (knownLength)
			size -= 4;
		fileOffset += (off_t) 4;
		if (scannedLength != NULL)
			(*scannedLength) += 4;
		(*object)->objectSize += 4;

		if (byteOrder == BYTEORDER_SAME) {
			GET_SHORT_SAME_ORDER(ptr, group);
			GET_SHORT_SAME_ORDER(ptr + 2, element);
			e.tag = DCM_MAKETAG(group, element);
		} else {
			GET_SHORT_REVERSE_ORDER(ptr, group);
			GET_SHORT_REVERSE_ORDER(ptr + 2, element);
			e.tag = DCM_MAKETAG(group, element);
		}
		ptr += 4;

		if (explicitVR) {
			if (fileFlag) {
				if (fd != -1) {
					nBytes = read(fd, (char*)buf, 4);
				} else {
					cond = (*object)->rd((*object)->userCtx, buf, 4, &nBytes);
				}

				if (nBytes != 4)
					return COND_PushCondition(DCM_FILEACCESSERROR, DCM_Message(DCM_FILEACCESSERROR), name, "readFile");
				ptr = buf;
			}
			if (knownLength)
				size -= 4;
			fileOffset += (off_t) 4;
			if (scannedLength != NULL)
				(*scannedLength) += 4;
			(*object)->objectSize += 4;
			if ((strncmp((char *) ptr, "OB", 2) == 0)
			 || (strncmp((char *) ptr, "OW", 2) == 0)
			 || (strncmp((char *) ptr, "SQ", 2) == 0)) {
			} else {
			}
		} else {
			if (fileFlag) {
				if (fd != -1) {
					nBytes = read(fd, buf, 4);
				} else {
					cond = (*object)->rd((*object)->userCtx, buf, 4, &nBytes);
				}

				if (nBytes != 4)
					return COND_PushCondition(DCM_FILEACCESSERROR, DCM_Message(DCM_FILEACCESSERROR), name, "readFile");
				ptr = buf;
			}
			if (knownLength)
				size -= 4;
			fileOffset += (off_t) 4;
			if (scannedLength != NULL)
				(*scannedLength) += 4;
			(*object)->objectSize += 4;

			if (byteOrder == BYTEORDER_SAME) {
				GET_LONG_SAME_ORDER(ptr, e.length);
			} else {
				GET_LONG_REVERSE_ORDER(ptr, e.length);
			}
			ptr += 4;
		}

		if (((e.length & 1) != 0) && (e.length != DCM_UNSPECIFIEDLENGTH)) {
			if (debug)
				(void) DCM_DumpElements(callerObject, 0);
			(void) DCM_CloseObject(callerObject);
			return COND_PushCondition(DCM_UNEVENELEMENTLENGTH, DCM_Message(DCM_UNEVENELEMENTLENGTH), group, element, e.length, "readFile");
		}
		if ((e.length != (U32) DCM_UNSPECIFIEDLENGTH) && (e.length > (U32) size)) {
			if (debug)
				(void) DCM_DumpElements(callerObject, 0);
			(void) DCM_CloseObject(callerObject);
			return COND_PushCondition(DCM_ELEMENTLENGTHERROR, DCM_Message(DCM_ELEMENTLENGTHERROR), group, element, e.length, size, "readFile");
		}
		if ((e.tag == DCM_DLMITEMDELIMITATIONITEM)
		 || (e.tag == DCM_DLMSEQUENCEDELIMITATIONITEM)) {
			return DCM_NORMAL;
		}
		cond = DCM_LookupElement(&e);
		if (cond != DCM_NORMAL)
			(void) COND_PopCondition(0);
		if (e.representation == DCM_CTX)
			ctxSensitiveLookup(object, &e);

		if (e.representation == DCM_SQ) {
			cond = newElementItem(&e, FALSE, &elementItem);
			if (cond != DCM_NORMAL)
				return cond;
			elementItem->element.d.sq = LST_Create();
			if (elementItem->element.d.sq == NULL)
				return COND_PushCondition(DCM_LISTFAILURE, DCM_Message(DCM_LISTFAILURE), "readFile");

			localLength = elementItem->element.length;
			sequenceDone = (localLength == 0);

			while (!sequenceDone) {
			if (debug)
				fprintf(stderr, "Sequence Length: %d %x\n", localLength, localLength);
			if (fileFlag) {
				if (fd != -1) {
					nBytes = read(fd, buf, 8);
				} else {
					cond = (*object)->rd((*object)->userCtx, buf, 8, &nBytes);
				}
				if (nBytes != 8)
					return COND_PushCondition(DCM_FILEACCESSERROR, DCM_Message(DCM_FILEACCESSERROR), name, "readFile");
			ptr = buf;
			}
			if (size != (long) DCM_UNSPECIFIEDLENGTH)
				size -= 8;
			fileOffset += (off_t) 8;
			if (scannedLength != NULL)
				(*scannedLength) += 8;
			(*object)->objectSize += 8;
			if (localLength != DCM_UNSPECIFIEDLENGTH)
				localLength -= 8;

			if (byteOrder == BYTEORDER_SAME) {
				GET_SHORT_SAME_ORDER(ptr, tagGroup);
				GET_SHORT_SAME_ORDER(ptr + 2, tagElement);
				tagE.tag = DCM_MAKETAG(tagGroup, tagElement);
				GET_LONG_SAME_ORDER(ptr + 4, tagE.length);
			} else {
				GET_SHORT_REVERSE_ORDER(ptr, tagGroup);
				GET_SHORT_REVERSE_ORDER(ptr + 2, tagElement);
				tagE.tag = DCM_MAKETAG(tagGroup, tagElement);
				GET_LONG_REVERSE_ORDER(ptr + 4, tagE.length);
			}
			ptr += 8;
			if (debug)
				fprintf(stderr, "Sequence item: %4x %4x %d (%x)\n", tagGroup, tagElement, tagE.length, tagE.length);
			if (tagE.tag == DCM_DLMITEM) {
				/*		    if (size != DCM_UNSPECIFIEDLENGTH)
				size -= 8;
				*/
				/*		    fileOffset += 8;
				*/
				cond = readFile(name, (fileFlag) ? NULL : ptr, fd, tagE.length, fileOffset, recursionLevel + 1, opt, &sequenceObject, &sequenceLength, remainOpenFlag, ctx, rd, sk);
				if (cond == DCM_NORMAL) {
					sequenceItem = malloc(sizeof(*sequenceItem));
					if (sequenceItem == NULL)
						return COND_PushCondition(DCM_MALLOCFAILURE, DCM_Message(DCM_MALLOCFAILURE), sizeof(*sequenceItem), "readFile");

					sequenceItem->object = sequenceObject;
					cond = LST_Enqueue(&elementItem->element.d.sq, sequenceItem);
					if (cond != LST_NORMAL)
						return COND_PushCondition(DCM_LISTFAILURE, DCM_Message(DCM_LISTFAILURE), "readFile");
					if (size != (long) DCM_UNSPECIFIEDLENGTH)
						size -= sequenceLength;
					fileOffset += (off_t) sequenceLength;
					if (scannedLength != NULL)
						*scannedLength += sequenceLength;
					(*object)->objectSize += sequenceLength;
					if (localLength != DCM_UNSPECIFIEDLENGTH)
						localLength -= sequenceLength;
					ptr += sequenceLength;
				} else
					return cond;
			} else {
				sequenceDone = TRUE;
			}
			if (localLength == 0)
				sequenceDone = TRUE;
			}
		} else {
			pixelFlag = (e.tag == DCM_PXLPIXELDATA);
			cond = newElementItem(&e, (pixelFlag == FALSE), &elementItem);
			if (cond != DCM_NORMAL) {
				(void) DCM_CloseObject(callerObject);
				return cond;
			}
			if (pixelFlag) {
				if (fileFlag)
					*remainOpenFlag = TRUE;
				elementItem->byteOrder = byteOrder;
				elementItem->dataOffset = fileOffset;
				elementItem->currentOffset = 0;
				if (fileFlag)
					elementItem->element.d.ot = NULL;
				else
					elementItem->element.d.ot = (void *) ptr;
				if ((*object)->pixelBitsAllocated == 8)
					elementItem->element.representation = DCM_OB;
				else
					elementItem->element.representation = DCM_OW;
				if (fileFlag) {
					if (fd != -1) {
						(void) lseek(fd, (off_t) elementItem->element.length, SEEK_CUR);
					} else {
						(*object)->sk((*object)->userCtx,
						elementItem->element.length, SEEK_CUR);
					}
					(*object)->fd = fd;
				}
			} else {
				if (fileFlag) {
					if (fd != -1) {
						nBytes = read(fd, elementItem->element.d.ot,
						(int) elementItem->element.length);
					} else {
						cond = (*object)->rd((*object)->userCtx, elementItem->element.d.ot, (long) elementItem->element.length, &nBytes);
					}
					if (nBytes != (int) elementItem->element.length) {
						(void) DCM_CloseObject(callerObject);
						return COND_PushCondition(DCM_FILEACCESSERROR, DCM_Message(DCM_FILEACCESSERROR), name, "readFile");
					}
				} else {
					(void) memcpy(elementItem->element.d.ot, ptr, elementItem->element.length);
					ptr += elementItem->originalDataLength;
				}

#ifdef LITTLE_ENDIAN_ARCHITECTURE
				if (elementItem->element.representation == DCM_AT)
					swapATGroupElement(&elementItem->element);
#endif
				if (byteOrder != BYTEORDER_SAME)
					swapInPlace(object, &elementItem->element);
				if (convertFlag) {
					cond = verifyFormat(elementItem);
					if (cond != DCM_NORMAL)
						return cond;
				}
			}
			if (size != (long) DCM_UNSPECIFIEDLENGTH)
				size -= elementItem->originalDataLength;
			fileOffset += (off_t) elementItem->originalDataLength;
			if (scannedLength != NULL)
				(*scannedLength) += elementItem->originalDataLength;

			elementItem->paddedDataLength = elementItem->element.length;
			if (elementItem->paddedDataLength & 1)
				elementItem->paddedDataLength += 1;
			(*object)->objectSize += elementItem->paddedDataLength;
		}

		computeVM(object, &elementItem->element);

		if ((long) DCM_TAG_GROUP(e.tag) == lastGroup) {
			if ((long) DCM_TAG_ELEMENT(e.tag) <= lastElement)
				return COND_PushCondition(DCM_ELEMENTOUTOFORDER, DCM_Message(DCM_ELEMENTOUTOFORDER), group, element, "readFile");
		} else if ((long) DCM_TAG_GROUP(e.tag) > lastGroup) {
		} else {
			return COND_PushCondition(DCM_ELEMENTOUTOFORDER, DCM_Message(DCM_ELEMENTOUTOFORDER), group, element, "readFile");
		}
		lastGroup = (long) group;
		lastElement = (long) element;

		if (groupItem == NULL)
			createGroupFlag = TRUE;
		else if (groupItem->group != group)
			createGroupFlag = TRUE;
		else
			createGroupFlag = FALSE;

		if (createGroupFlag == TRUE) {
			groupItem = malloc(sizeof(*groupItem));
			if (groupItem == NULL) {
				(void) DCM_CloseObject(callerObject);
				return COND_PushCondition(DCM_ELEMENTCREATEFAILED, DCM_Message(DCM_ELEMENTCREATEFAILED), "readFile");
			}
			groupItem->group = group;
			groupItem->baseLength = 0;
			groupItem->longVRAttributes = 0;
			groupItem->elementList = LST_Create();
			if (groupItem->elementList == NULL) {
				(void) DCM_CloseObject(callerObject);
				return COND_PushCondition(DCM_LISTFAILURE, DCM_Message(DCM_LISTFAILURE), "readFile");
			}
			if (LST_Enqueue(&(*object)->groupList, groupItem) != LST_NORMAL) {
				(void) DCM_CloseObject(callerObject);
				return COND_PushCondition(DCM_LISTFAILURE, DCM_Message(DCM_LISTFAILURE), "readFile");
			}
		}
		if (element != 0x0000)
			groupItem->baseLength += 8 + elementItem->paddedDataLength;
		if ((element == 0x0000) && ((*object)->groupLengthFlag == FALSE)) {
			free(elementItem);
		} else {
			cond = LST_Enqueue(&groupItem->elementList, elementItem);
			if (cond != LST_NORMAL) {
				(void) DCM_CloseObject(callerObject);
				return COND_PushCondition(DCM_LISTFAILURE, DCM_Message(DCM_LISTFAILURE), "readFile");
			}
			cond = updateObjectType(object, &elementItem->element);	/* repair */

			cond = updateSpecialElements(object, elementItem);	/* repair */
		}

		if (size == 0)
			done = TRUE;

#ifdef DEBUG
		if (debug) {
		/*lint -e644 */
			(void) fprintf(stderr, "Address: %px Group %2x, element %2x, length %ld ", elementItem, DCM_TAG_GROUP(elementItem->element.tag), DCM_TAG_ELEMENT(elementItem->element.tag), elementItem->element.length);
		/*lint +e644 */
			(void) fprintf(stderr, "Object size: %d\n", (*object)->objectSize);
		}
#endif
	}

	groupItem = LST_Head(&(*object)->groupList);
	if (groupItem != NULL) {
		(void) LST_Position(&(*object)->groupList, groupItem);
		while (groupItem != NULL) {
			elementItem = LST_Head(&groupItem->elementList);
			if (elementItem != NULL) {
				if (DCM_TAG_ELEMENT(elementItem->element.tag) == 0x0000) {
					*elementItem->element.d.ul = groupItem->baseLength;
				}
			}
			groupItem = LST_Next(&(*object)->groupList);
		}
	}

	return DCM_NORMAL;
}

