/*
          Copyright (C) 1993, 1994, RSNA and Washington University

          The software and supporting documentation for the Radiological
          Society of North America (RSNA) 1993, 1994 Digital Imaging and
          Communications in Medicine (DICOM) Demonstration were developed
          at the
                  Electronic Radiology Laboratory
                  Mallinckrodt Institute of Radiology
                  Washington University School of Medicine
                  510 S. Kingshighway Blvd.
                  St. Louis, MO 63110
          as part of the 1993, 1994 DICOM Central Test Node project for, and
          under contract with, the Radiological Society of North America.

          THIS SOFTWARE IS MADE AVAILABLE, AS IS, AND NEITHER RSNA NOR
          WASHINGTON UNIVERSITY MAKE ANY WARRANTY ABOUT THE SOFTWARE, ITS
          PERFORMANCE, ITS MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
          USE, FREEDOM FROM ANY COMPUTER DISEASES OR ITS CONFORMITY TO ANY
          SPECIFICATION. THE ENTIRE RISK AS TO QUALITY AND PERFORMANCE OF
          THE SOFTWARE IS WITH THE USER.

          Copyright of the software and supporting documentation is
          jointly owned by RSNA and Washington University, and free access
          is hereby granted as a license to use this software, copy this
          software and prepare derivative works based upon this software.
          However, any distribution of this software source code or
          supporting documentation or derivative works (source code and
          supporting documentation) must include the three paragraphs of
          the copyright notice.
*/
/* Copyright marker.  Copyright will be inserted above.  Do not remove */

/*
**				DICOM 93
**		     Electronic Radiology Laboratory
**		   Mallinckrodt Institute of Radiology
**		Washington University School of Medicine
**
** Module Name(s):
**			SRV_ReceiveCommand
**			SRV_ReceiveDataSet
**			SRV_SendCommand
**			SRV_SendDataSet
**			SRV_TestForCancel
**	private modules
**			SRVPRV_ReadNextPDV
**			writeCallback
**			dequeCommand
**			enqueCommand
** Author, Date:	Stephen M. Moore, 15-Apr-93
** Intent:		This module contains general routines which are used
**			in our implementation of service classes.  These
**			routines allow users to request and accept service
**			classes, build and manipulate the public DUL
**			structures, send and receive messages, and request
**			unique Message IDs.
** Last Update:		$Author: scsb $, $Date: 2003/04/28 19:18:46 $
** Source File:		$RCSfile: scsb_send.cpp,v $
** Revision:		$Revision: 1.2 $
** Status:		$State: Exp $
*/

static char rcsid[] = "$Revision: 1.2 $ $RCSfile: scsb_send.cpp,v $";

#include "ctn_os.h"

#include "dicom.h"
#include "dicom_uids.h"
#include "condition.h"
#include "lst.h"
#include "dulprotocol.h"
#include "dicom_objects.h"
#include "dicom_messages.h"
#include "dicom_services.h"
#include "private.h"

/* SRV_SendDataSet
**
** Purpose:
**	Send a DICOM dataset to a peer using an established Association.
**
** Parameter Dictionary:
**	association	Key which describes the Association used for
**			transmitting the command.
**	context		Presentation context to be used when sending the
**			command.
**	object		Address of DICOM object which contains the dataset to
**			be sent.
**	callback	User callback function which is called periodically
**			while the dataset is sent accross the network. This
**			allows the application to monitor the rate of
**			transmission and cancel the transmission.
**	callbackCtx	User context information which is passed to the
**			callback function as a parameter.
**	length		Length in bytes of the amount of data to transmit
**			over the network before calling the user's callback
**			function.
**
** Return Values:
**
**	SRV_NORMAL
**	SRV_NOTRANSFERSYNTAX
**	SRV_OBJECTACCESSFAILED
**	SRV_SENDFAILED
**
** Algorithm:
**	Description of the algorithm (optional) and any other notes.
*/

CONDITION SRV_SendDataSetEx(DUL_ASSOCIATIONKEY** association, DUL_PRESENTATIONCONTEXT* context, const char* dataset_ptr, unsigned long dataset_size)
{
	CONDITION cond;

	unsigned long maxPDUsize;
	cond = DUL_AssociationParameter(association, DUL_K_MAX_PDV_XMIT, DUL_K_INTEGER, &maxPDUsize, sizeof(maxPDUsize));
	if (cond != DUL_NORMAL)
		return COND_PushCondition(SRV_SENDFAILED, SRV_Message(SRV_SENDFAILED), "as unable to get maxPDUsize", "SRV_SendDataSetEx");

	const unsigned int fudgeFactor = 20;
	if (maxPDUsize < fudgeFactor)
		return COND_PushCondition(SRV_SENDFAILED, SRV_Message(SRV_SENDFAILED), "due to invalid maxPDUsize", "SRV_SendDataSetEx");

	const unsigned int pdv_data_size_max = maxPDUsize - fudgeFactor;
	const unsigned int lastFragmentSize = dataset_size % pdv_data_size_max;
	const unsigned int pdvCount = dataset_size / pdv_data_size_max + (lastFragmentSize == 0 ? 0 : 1);

	DUL_PDV* const pdvArray = (DUL_PDV*)_alloca(sizeof(DUL_PDV) * pdvCount);
	if (pdvArray == NULL)
		return COND_PushCondition(SRV_MALLOCFAILURE, SRV_Message(SRV_MALLOCFAILURE), sizeof(DUL_PDV) * pdvCount, "SRV_SendDataSetEx");

	DUL_PDV* pdv = pdvArray;
	for (int i = 0; i < pdvCount; i++, pdv++) {
		if (i == pdvCount - 1) {
		    pdv->fragmentLength = lastFragmentSize != 0 ? lastFragmentSize : pdv_data_size_max;
		    pdv->lastPDV = 1;
		} else {
		    pdv->fragmentLength = pdv_data_size_max;
		    pdv->lastPDV = 0;
		}
		pdv->presentationContextID = context->presentationContextID;
		pdv->pdvType = DUL_DATASETPDV;
		pdv->data = (void*)(dataset_ptr + pdv_data_size_max * i);
	}

	DUL_PDVLIST pdvList;
	ZeroMemory(&pdvList, sizeof(pdvList));
	pdvList.count = pdvCount;
	pdvList.pdv = pdvArray;

    cond = DUL_WritePDVs(association, &pdvList);
    if (cond != DUL_NORMAL)
		return COND_PushCondition(SRV_SENDFAILED, SRV_Message(SRV_SENDFAILED), "DUL_WritePDVs", "SRV_SendDataSetEx");

	return SRV_NORMAL;
}
