/*
          Copyright (C) 1993, 1994, RSNA and Washington University

          The software and supporting documentation for the Radiological
          Society of North America (RSNA) 1993, 1994 Digital Imaging and
          Communications in Medicine (DICOM) Demonstration were developed
          at the
                  Electronic Radiology Laboratory
                  Mallinckrodt Institute of Radiology
                  Washington University School of Medicine
                  510 S. Kingshighway Blvd.
                  St. Louis, MO 63110
          as part of the 1993, 1994 DICOM Central Test Node project for, and
          under contract with, the Radiological Society of North America.

          THIS SOFTWARE IS MADE AVAILABLE, AS IS, AND NEITHER RSNA NOR
          WASHINGTON UNIVERSITY MAKE ANY WARRANTY ABOUT THE SOFTWARE, ITS
          PERFORMANCE, ITS MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
          USE, FREEDOM FROM ANY COMPUTER DISEASES OR ITS CONFORMITY TO ANY
          SPECIFICATION. THE ENTIRE RISK AS TO QUALITY AND PERFORMANCE OF
          THE SOFTWARE IS WITH THE USER.

          Copyright of the software and supporting documentation is
          jointly owned by RSNA and Washington University, and free access
          is hereby granted as a license to use this software, copy this
          software and prepare derivative works based upon this software.
          However, any distribution of this software source code or
          supporting documentation or derivative works (source code and
          supporting documentation) must include the three paragraphs of
          the copyright notice.
*/
/* Copyright marker.  Copyright will be inserted above.  Do not remove */

/*
**				DICOM 93
**		     Electronic Radiology Laboratory
**		   Mallinckrodt Institute of Radiology
**		Washington University School of Medicine
**
** Module Name(s):	SRV_CStoreRequest
**			SRV_CStoreResponse
**	private modules
**			localCallback
**			genericImageSize
** Author, Date:	Stephen M. Moore, 22-Apr-93
** Intent:		This module contains routines which implement the
**			storage service class (but user and provider).
** Last Update:		$Author: scsb $, $Date: 2003/04/28 19:18:46 $
** Source File:		$RCSfile: scsb_storage.cpp,v $
** Revision:		$Revision: 1.2 $
** Status:		$State: Exp $
*/

static char rcsid[] = "$Revision: 1.2 $ $RCSfile: scsb_storage.cpp,v $";

#include "ctn_os.h"

#include "dicom.h"
#include "dicom_uids.h"
#include "condition.h"
#include "lst.h"
#include "utility.h"
#include "dulprotocol.h"
#include "dicom_objects.h"
#include "dicom_messages.h"
#include "dicom_services.h"
#include "private.h"

static unsigned long genericImageSize(char *SOPClass);
typedef struct {
    CONDITION(*callback) ();
    MSG_C_STORE_REQ *storeRequest;
    MSG_C_STORE_RESP *storeResponse;
    void *userCtx;
}   STORAGE_CTX;
static CONDITION
localCallback(unsigned long bytesTransmitted, unsigned long totalBytes,
	      STORAGE_CTX * storageCtx);
typedef struct {
  char *xferSyntax;
  unsigned long options;
} XFER_SYNTAX_MAP;

static
XFER_SYNTAX_MAP xferSyntaxMap[] = {
	{DICOM_TRANSFERLITTLEENDIAN, DCM_ORDERLITTLEENDIAN },
	{DICOM_TRANSFERLITTLEENDIANEXPLICIT, DCM_EXPLICITLITTLEENDIAN },
	{DICOM_TRANSFERBIGENDIANEXPLICIT, DCM_EXPLICITBIGENDIAN}
};

/*
static unsigned long
exportOptionsFromXferSyntax(const char* xferSyntax)
{
  int index = 0;

  for (index = 0; index < DIM_OF(xferSyntaxMap); index++) {
    if (strcmp(xferSyntax, xferSyntaxMap[index].xferSyntax) == 0)
      return xferSyntaxMap[index].options;
  }
  return 0;
}
*/


static CONDITION
callbackFunction(void* buf, U32 bytesExported, int lastFlag, void* ctx)
{
  int fd = *(int*)ctx;

  if (write(fd, buf, bytesExported) != (int)bytesExported) {
    return 0;
  }
  return DCM_NORMAL;
}

static int
writePart10MetaHeader(int fd, DCM_FILE_META* fileMeta)
{
  DCM_OBJECT* obj = 0;
  char buffer[2048];

  CONDITION cond = DCM_CreateObject(&obj, 0);
  if (cond != DCM_NORMAL) {
    return -1;
  }

  cond = DCM_SetFileMeta(&obj, fileMeta);
  if (cond != DCM_NORMAL) {
    return -1;
  }

  cond = DCM_ExportStream(&obj, 
			    DCM_EXPLICITLITTLEENDIAN | DCM_PART10FILE,
			    buffer, sizeof(buffer), callbackFunction, &fd);
  if (cond != DCM_NORMAL) {
    return -1;
  }

  return 0;
}

static int
setFileMeta(DCM_FILE_META* fileMeta, MSG_C_STORE_REQ* msg,
	    const char* xferSyntax)
{
  memset(fileMeta->preamble, 0, sizeof(fileMeta->preamble));
  fileMeta->fileMetaInformationVersion[0] = 0x01;
  fileMeta->fileMetaInformationVersion[1] = 0x00;
  strcpy(fileMeta->mediaStorageSOPClassUID, msg->classUID);
  strcpy(fileMeta->mediaStorageSOPInstanceUID, msg->instanceUID);
  strcpy(fileMeta->transferSyntaxUID,  xferSyntax);
  strcpy(fileMeta->implementationClassUID, MIR_IMPLEMENTATIONCLASSUID);
  strcpy(fileMeta->implementationVersionName, MIR_IMPLEMENTATIONVERSIONNAME);
  strcpy(fileMeta->sourceApplicationEntityTitle, "XX");
  strcpy(fileMeta->privateInformationCreatorUID, "");

  return 0;
}

/* SRV_CStoreRequest
**
** Purpose:
**	SRV_CStoreRequest assists an application that wants to be an SCU of
**	one of the storage SOP classes.  This function constructs a
**	C-STORE-REQ Message and sends it to the peer application which
**	is acting as the SCP for the storage SOP class.  After the request
**	message is sent, SRV_CStoreRequest sends the data set which
**	contains the object of the store request.
**
**	The user specifies the data set for the operation by placing
**	a legal DICOM Information Object in the MSG_C_STORE_REQ structure
**	or by including a file name in the structure that points to a
**	DICOM Information Object.
**
**	The function calculates the numbers of bytes that are present
**	in the data set and calls the user callback function during the
**	send process.  The callback function is called after each 10%
**	of image transmission.
**
**	The arguments to the callback function are:
**		MSG_C_STORE_REQ	*storeRequest
**		MSG_C_STORE_RESP *storeResponse
**		char		*abstractSyntax
**		unsigned long	bytesTransmitted
**		unsigned long	totalBytes
**		void		*callbackCtx
**	where
**		storeRequest	Pointer to MSG structure with C-STORE request.
**		storeRespponse	Pointer to MSG structure with C-STORE response.
**		abstractSyntax	A character string which identifies the
**				abstract syntax of the SOP Class of the
**				object to be stored.
**		bytesTransmitted Number of bytes transmitted so far.
**		totalBytes	Total number of bytes in data set.
**		callbackCtx	User's callbackCtx argument which is used to
**				maintain context information in the callback
**				function.
**
**	On each invocation of the callback function, the user should examine
**	the storeResponse pointer.  This pointer will be NULL during the
**	store process.  After SRV_CStoreRequest completes the process of
**	sending the image, it waits for the C-STORE RESPONSE from the peer.
**	When this reponse is receive, the callback function is called a
**	final time with the response message.
**
**	The callback function should return SRV_NORMAL.  Any other return
**	value will cause the SRV facility to abort the Association.
**
** Parameter Dictionary:
**	association	The key which is used to access an existing
**			association.
**	params		The list of parameters for the association.  This
**			list includes the list of presentation contexts.
**	storeRequest	Pointer to structure in caller's memory which contains
**			the STORE request.
**	storeResponse	Address of pointer to response message. This function
**			will create a response message and return the
**			address of the structure to the caller.
**	callback	Address of user routine which is called during the
**			send process and when the response message is received.
**	callbackCtx	User context information which is supplied during call
**			to callback function.
**      dirName         Name of directory where files may be created.
**
** Return Values:
**
**	SRV_CALLBACKABORTEDSERVICE
**	SRV_ILLEGALPARAMETER
**	SRV_NOCALLBACK
**	SRV_NORMAL
**	SRV_NOSERVICEINASSOCIATION
**	SRV_OBJECTBUILDFAILED
**	SRV_REQUESTFAILED
**	SRV_UNEXPECTEDCOMMAND
**
** Algorithm:
**	Description of the algorithm (optional) and any other notes.
*/

CONDITION SRV_CStoreRequestEx(DUL_ASSOCIATIONKEY** association, DUL_ASSOCIATESERVICEPARAMETERS* params, MSG_C_STORE_REQ* storeRequest, MSG_C_STORE_RESP* storeResponse, const char* dataset_ptr, unsigned long dataset_size)
{
	CONDITION cond;

	if (storeRequest->type != MSG_K_C_STORE_REQ)
		return COND_PushCondition(SRV_ILLEGALPARAMETER, SRV_Message(SRV_ILLEGALPARAMETER), "type", "STORE Request", "SRV_CStoreRequestEx");
	if (storeRequest->dataSetType == DCM_CMDDATANULL)
		return COND_PushCondition(SRV_ILLEGALPARAMETER, SRV_Message(SRV_ILLEGALPARAMETER), "dataSetType", "STORE Request", "SRV_CStoreRequestEx");

	DUL_PRESENTATIONCONTEXT* presentationCtx = SRVPRV_PresentationContext(params, storeRequest->classUID);
	if (presentationCtx == NULL)
		return COND_PushCondition(SRV_NOSERVICEINASSOCIATION, SRV_Message(SRV_NOSERVICEINASSOCIATION), storeRequest->classUID, "SRV_CStoreRequestEx");

	DCM_OBJECT* commandObject;
	cond = MSG_BuildCommand(storeRequest, &commandObject);
	if (cond != MSG_NORMAL)
		return COND_PushCondition(SRV_OBJECTBUILDFAILED, SRV_Message(SRV_OBJECTBUILDFAILED), "STORE Request", "SRV_CStoreRequestEx");

	cond = SRV_SendCommand(association, presentationCtx, &commandObject);
	(void) DCM_CloseObject(&commandObject);
	if (cond != SRV_NORMAL)
	return COND_PushCondition(SRV_REQUESTFAILED,
		       SRV_Message(SRV_REQUESTFAILED), "SRV_CStoreRequestEx");

	cond = SRV_SendDataSetEx(association, presentationCtx, dataset_ptr, dataset_size);
	if (cond != SRV_NORMAL)
		return COND_PushCondition(SRV_REQUESTFAILED, SRV_Message(SRV_REQUESTFAILED), "SRV_CStoreRequestEx");

	void* message;
	MSG_TYPE messageType;
	unsigned short command;
	DUL_PRESENTATIONCONTEXTID ctxID;
	cond = SRV_ReceiveCommand(association, params, DUL_BLOCK, 0, &ctxID, &command, &messageType, &message);
	if (cond != SRV_NORMAL)
		return COND_PushCondition(SRV_REQUESTFAILED, SRV_Message(SRV_REQUESTFAILED), "SRV_CStoreRequestEx");

	if (messageType != MSG_K_C_STORE_RESP) {
		MSG_Free(&message);
		return COND_PushCondition(SRV_UNEXPECTEDCOMMAND, SRV_Message(SRV_UNEXPECTEDCOMMAND), (int) command, "SRV_CStoreRequestEx");
	}
	MSG_C_STORE_RESP* response = (MSG_C_STORE_RESP*)message;
	if (storeResponse != NULL)
		*storeResponse = *response;

	MSG_Free(&message);

	return SRV_NORMAL;
}

