/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "UtilClasses.h"

#include "dcmdict.h"
#include "CDcmScanner.h"

#include "DataElementCache.h"


//----------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static HRESULT append_value(IXMLDOMDocument* dom, IXMLDOMNode* elm, IXMLDOMElement* valueTemplate, BSTR& valueStr)
{
	AutoComPtr<IXMLDOMNode> newValue;
	HRESULT hr = valueTemplate->cloneNode(VARIANT_TRUE, newValue.r());
	if (FAILED(hr))
		return hr;
	
	AutoComPtr<IXMLDOMText> text;
	hr = dom->createTextNode(valueStr, text.r());
	if (FAILED(hr))
		return hr;

	hr = newValue->appendChild(text.p(), NULL);
	if (FAILED(hr))
		return hr;

	hr = elm->appendChild(newValue.p(), NULL);
	if (FAILED(hr))
		return hr;	

	return S_OK;
}
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static HRESULT string_value(const char* data, int len, bool multiValue, bool trimSpace, bool multiByte, IXMLDOMDocument* dom, IXMLDOMNode* newElement, IXMLDOMElement* valueTemplate, BSTR& valueStr)
{
	const unsigned char* cp = (const unsigned char*)data;
	const unsigned char* ep = cp + len;

	while (cp < ep) {
		if (trimSpace) {
			while (cp < ep && (!multiValue || *cp != '\\') && *cp <= ' ')
				cp++;
			if (cp == ep) {
				break;
			}
		}
		const unsigned char* bp = cp;
		while (cp < ep && (!multiValue || *cp != '\\'))
			cp++;
		const unsigned char* tp = cp;
		if (trimSpace) {
			while (tp > bp && *(tp - 1) <= ' ')
				tp--;
		}

		if (tp == bp) {
			SysReAllocStringLen(&valueStr, NULL, 0);
		} else if (multiByte) {
			const int len = multi_byte_to_wide_char((const char*)bp, tp - bp, NULL, 0);
			if (len == 0)
				return HRESULT_FROM_WIN32(GetLastError());
			if (!SysReAllocStringLen(&valueStr, NULL, len))
				return E_OUTOFMEMORY;
			if (multi_byte_to_wide_char((const char*)bp, tp - bp, valueStr, len) == 0)
				return HRESULT_FROM_WIN32(GetLastError());
		} else {
			if (!SysReAllocStringLen(&valueStr, NULL, tp - bp))
				return E_OUTOFMEMORY;
			for (wchar_t* wp = valueStr; bp < tp; bp++, wp++)
				*wp = (wchar_t)*bp;
		}

		HRESULT hr = append_value(dom, newElement, valueTemplate, valueStr);
		if (FAILED(hr))
			return hr;

		if (cp < ep && *cp == '\\')
			cp++;
	}

	return S_OK;
}

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static HRESULT binary_numeric_value(const char* data, int len, VARTYPE vt, IXMLDOMDocument* dom, IXMLDOMNode* newElement, IXMLDOMElement* valueTemplate, VARIANT& var, BSTR& valueStr)
{
	const char* cp = data;
	const char* ep = cp + len;

	var.vt = vt;

	while (cp < ep) {
		switch (vt) {
		case VT_I2:
			var.iVal = *reinterpret_cast<const short*>(cp);
			cp += 2;
			break;
		case VT_I4:
			var.lVal = *reinterpret_cast<const long*>(cp);
			cp += 4;
			break;
		case VT_UI2:
			var.uiVal = *reinterpret_cast<const unsigned short*>(cp);
			cp += 2;
			break;
		case VT_UI4:
			var.ulVal = *reinterpret_cast<const unsigned long*>(cp);
			cp += 4;
			break;
		case VT_R4:
			var.fltVal = *reinterpret_cast<const float*>(cp);
			cp += 4;
			break;
		case VT_R8:
			var.dblVal = *reinterpret_cast<const double*>(cp);
			cp += 8;
			break;
		}

		HRESULT hr = VarFormatNumber(&var, -1, 0, 0, 0, 0, &valueStr);
		if (FAILED(hr))
			return hr;
		
		hr = append_value(dom, newElement, valueTemplate, valueStr);
		if (FAILED(hr))
			return hr;

	}

	return S_OK;
}
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static void hexwstr(unsigned int val, wchar_t* str, unsigned int len)
{
	wchar_t* cp = str + len - 1;
	do {
		unsigned int ch = val & 0xF;
		if (ch < 10)
			ch += L'0';
		else
			ch += L'A' - 10;
		*cp-- = ch;
		val = val >> 4;
	} while (cp >= str);

	*(str + len) = L'\0';
}

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static HRESULT add_element(const tDcmElementHdr& hdr, const char* cp, DWORD id, IXMLDOMDocument* dom, IXMLDOMElement* listElm, IXMLDOMElement* elementTemplate, IXMLDOMAttribute* idAttr, IXMLDOMAttribute* tagAttr, IXMLDOMElement* valueTemplate, VARIANT& var, BSTR& valueStr)
{
	HRESULT hr;

	var.vt = VT_UI4;
	var.ulVal = id;
	hr = idAttr->put_value(var);
	if (FAILED(hr))
		return hr;

	wchar_t tagStr[12];
	hexwstr(hdr.ltag(), tagStr, 8);
	if (!SysReAllocString(&valueStr, tagStr))
		return E_OUTOFMEMORY;
	var.vt = VT_BSTR;
	var.bstrVal = valueStr;;
	hr = tagAttr->put_value(var);
	if (FAILED(hr))
		return hr;

	AutoComPtr<IXMLDOMNode> newElement;
	hr = elementTemplate->cloneNode(VARIANT_TRUE, newElement.r());
	if (FAILED(hr))
		return hr;
	
	hr = listElm->appendChild(newElement.p(), NULL);
	if (FAILED(hr))
		return hr;

	if (hdr.value_length() != 0) {
		if (hdr.m_vr[0] == 'A' && (hdr.m_vr[1] == 'E' || hdr.m_vr[1] == 'S')) { // AE , AS
			hr = string_value(hdr.value_ptr(cp), hdr.value_length(), true, true, false, dom, newElement.p(), valueTemplate, valueStr);
		} else if (hdr.m_vr[0] == 'C') {
			if (hdr.m_vr[1] == 'S') { // CS
				hr = string_value(hdr.value_ptr(cp), hdr.value_length(), true, true, false, dom, newElement.p(), valueTemplate, valueStr);
			} else if (hdr.m_vr[1] == 'T') { // CTX
				hr = binary_numeric_value(hdr.value_ptr(cp), hdr.value_length(), VT_I2, dom, newElement.p(), valueTemplate, var, valueStr);
			} else {
				hr = S_OK;
			}
		} else if (hdr.m_vr[0] == 'D' && (hdr.m_vr[1] == 'A' || hdr.m_vr[1] == 'S' || hdr.m_vr[1] == 'T')) { // DA, DS, DT
			hr = string_value(hdr.value_ptr(cp), hdr.value_length(), true, true, false, dom, newElement.p(), valueTemplate, valueStr);
		} else if (hdr.m_vr[0] == 'I' && hdr.m_vr[1] == 'S') { // IS
			hr = string_value(hdr.value_ptr(cp), hdr.value_length(), true, true, false, dom, newElement.p(), valueTemplate, valueStr);
		} else if (hdr.m_vr[0] == 'L') {
			if (hdr.m_vr[1] == 'O') { // LO
				hr = string_value(hdr.value_ptr(cp), hdr.value_length(), true, false, true, dom, newElement.p(), valueTemplate, valueStr);
			} else if (hdr.m_vr[1] == 'T') { // LT
				hr = string_value(hdr.value_ptr(cp), hdr.value_length(), false, false, true, dom, newElement.p(), valueTemplate, valueStr);
			} else {
				hr = S_OK;
			}
		} else if (hdr.m_vr[0] == 'P' && hdr.m_vr[1] == 'N') { // PN
			hr = string_value(hdr.value_ptr(cp), hdr.value_length(), true, true, true, dom, newElement.p(), valueTemplate, valueStr);
		} else if (hdr.m_vr[0] == 'S') {
			if (hdr.m_vr[1] == 'H') { // SH
				hr = string_value(hdr.value_ptr(cp), hdr.value_length(), true, true, true, dom, newElement.p(), valueTemplate, valueStr);
			} else if (hdr.m_vr[1] == 'T') { // ST
				hr = string_value(hdr.value_ptr(cp), hdr.value_length(), false, false, true, dom, newElement.p(), valueTemplate, valueStr);
			} else if (hdr.m_vr[1] == 'L') { // SL
				hr = binary_numeric_value(hdr.value_ptr(cp), hdr.value_length(), VT_I4, dom, newElement.p(), valueTemplate, var, valueStr);
			} else if (hdr.m_vr[1] == 'S') { // SS
				hr = binary_numeric_value(hdr.value_ptr(cp), hdr.value_length(), VT_I2, dom, newElement.p(), valueTemplate, var, valueStr);
			} else {
				hr = S_OK;
			}
		} else if (hdr.m_vr[0] == 'T' && hdr.m_vr[1] == 'M') { // TM
			hr = string_value(hdr.value_ptr(cp), hdr.value_length(), true, true, false, dom, newElement.p(), valueTemplate, valueStr);
		} else if (hdr.m_vr[0] == 'U') {
			if (hdr.m_vr[1] == 'I') { // UI
				hr = string_value(hdr.value_ptr(cp), hdr.value_length(), true, true, false, dom, newElement.p(), valueTemplate, valueStr);
			} else if (hdr.m_vr[1] == 'T') { // UT
				hr = string_value(hdr.value_ptr(cp), hdr.value_length(), false, false, true, dom, newElement.p(), valueTemplate, valueStr);
			} else if (hdr.m_vr[1] == 'L') { // UL
				hr = binary_numeric_value(hdr.value_ptr(cp), hdr.value_length(), VT_UI4, dom, newElement.p(), valueTemplate, var, valueStr);
			} else if (hdr.m_vr[1] == 'S') { // US
				hr = binary_numeric_value(hdr.value_ptr(cp), hdr.value_length(), VT_UI2, dom, newElement.p(), valueTemplate, var, valueStr);
			} else {
				hr = S_OK;
			}
		}
		if (FAILED(hr))
			return hr;
	}

	return S_OK;
}


//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
template < class BLOCK_T >
static HRESULT block_list_to_dom(BLOCK_T* bp, IXMLDOMDocument* dom)
{
	HRESULT hr;

	BSTR str1 = SysAllocStringLen(NULL, 1);
	AutoSysFreeString af_str1(str1);

	BSTR str2 = SysAllocStringLen(NULL, 2);
	AutoSysFreeString af_str2(str2);

	BSTR valueStr = SysAllocStringLen(NULL, 256);
	AutoSysFreeString af_valueStr(valueStr);

	VARIANT var;
	VariantInit(&var);

	*str1 = L'E';
	AutoComPtr<IXMLDOMElement> elementTemplate;
	hr = dom->createElement(str1, elementTemplate.r());
	if (FAILED(hr))
		return hr;

	*str1 = L'i';
	AutoComPtr<IXMLDOMAttribute> idAttr;
	hr = dom->createAttribute(str1, idAttr.r());
	if (FAILED(hr))
		return hr;

	hr = elementTemplate->setAttributeNode(idAttr.p(), NULL);
	if (FAILED(hr))
		return hr;

	*str1 = L't';
	AutoComPtr<IXMLDOMAttribute> tagAttr;
	hr = dom->createAttribute(str1, tagAttr.r());
	if (FAILED(hr))
		return hr;

	hr = elementTemplate->setAttributeNode(tagAttr.p(), NULL);
	if (FAILED(hr))
		return hr;

	*str1 = L'v';
	AutoComPtr<IXMLDOMElement> valueTemplate;
	hr = dom->createElement(str1, valueTemplate.r());
	if (FAILED(hr))
		return hr;

	lstrcpy(str2, L"EL");
	AutoComPtr<IXMLDOMElement> listElm;
	hr = dom->createElement(str2, listElm.r());
	if (FAILED(hr))
		return hr;

	DWORD offset = 0;

	do {
		const char* const block_begin = reinterpret_cast<const char*>(bp->m_data);
		const char* const block_end = block_begin + bp->m_bytesInBlock;

		CDcmScannerF scanner(block_begin, block_end);

		tDcmElementHdr hdr;
		const char* cp = block_begin;

		while (cp < block_end) {
			int sr = scanner.NextElement(cp, hdr, true);
			if (sr != CDcmScannerF::S_CONTINUE)
				break;

			hr = ::add_element(hdr, cp, (cp - block_begin) + offset, dom, listElm.p(), elementTemplate.p(), idAttr.p(), tagAttr.p(), valueTemplate.p(), var, valueStr);
			if (FAILED(hr))
				return hr;

			cp = hdr.next_ptr(cp);
		}

		offset += bp->m_bytesInBlock;
		bp = bp->m_next;
	} while (bp != NULL);

	AutoComPtr<IXMLDOMElement> documentElement;
	hr = dom->get_documentElement(documentElement.r());
	if (FAILED(hr))
		return hr;

	hr = documentElement->appendChild(listElm.p(), NULL);
	if (FAILED(hr))
		return hr;
	
	return S_OK;
}
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
template < class BLOCK_T >
static void free_blocks_(BLOCK_T& first_block)
{
	BLOCK_T* mb = first_block.m_next;
	while (mb != NULL) {
		BLOCK_T* next = mb->m_next;
		delete mb;
		mb = next;
	}
	first_block.m_next = NULL;
}

//----------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
void CDataElementCache::avlnode_t::avlrotright(avlnode_t **n)
{
	avlnode_t *tmp = *n;

	*n = (*n)->left;
	tmp->left = (*n)->right;
	(*n)->right = tmp;
}
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
void CDataElementCache::avlnode_t::avlrotleft(avlnode_t **n)
{
	avlnode_t *tmp = *n;

	*n = (*n)->right;
	tmp->right = (*n)->left;
	(*n)->left = tmp;
}

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
CDataElementCache::avlnode_t::AVLRES CDataElementCache::avlnode_t::avlleftgrown(avlnode_t **n)
{
	switch ((*n)->skew) {
	case AVL_LEFT:
		if ((*n)->left->skew == AVL_LEFT) {
			(*n)->skew = (*n)->left->skew = AVL_NONE;
			avlrotright(n);
		}	
		else {
			switch ((*n)->left->right->skew) {
			case AVL_LEFT:
				(*n)->skew = AVL_RIGHT;
				(*n)->left->skew = AVL_NONE;
				break;

			case AVL_RIGHT:
				(*n)->skew = AVL_NONE;
				(*n)->left->skew = AVL_LEFT;
				break;

			default:
				(*n)->skew = AVL_NONE;
				(*n)->left->skew = AVL_NONE;
			}
			(*n)->left->right->skew = AVL_NONE;
			avlrotleft(& (*n)->left);
			avlrotright(n);
		}
		return AVL_OK;

	case AVL_RIGHT:
		(*n)->skew = AVL_NONE;
		return AVL_OK;
	
	default:
		(*n)->skew = AVL_LEFT;
		return AVL_BALANCE;
	}
}
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
CDataElementCache::avlnode_t::AVLRES CDataElementCache::avlnode_t::avlrightgrown(avlnode_t **n)
{
	switch ((*n)->skew) {
	case AVL_LEFT:					
		(*n)->skew = AVL_NONE;
		return AVL_OK;

	case AVL_RIGHT:
		if ((*n)->right->skew == AVL_RIGHT) {	
			(*n)->skew = (*n)->right->skew = AVL_NONE;
			avlrotleft(n);
		}
		else {
			switch ((*n)->right->left->skew) {
			case AVL_RIGHT:
				(*n)->skew = AVL_LEFT;
				(*n)->right->skew = AVL_NONE;
				break;

			case AVL_LEFT:
				(*n)->skew = AVL_NONE;
				(*n)->right->skew = AVL_RIGHT;
				break;

			default:
				(*n)->skew = AVL_NONE;
				(*n)->right->skew = AVL_NONE;
			}
			(*n)->right->left->skew = AVL_NONE;
			avlrotright(& (*n)->right);
			avlrotleft(n);
		}
		return AVL_OK;

	default:
		(*n)->skew = AVL_RIGHT;
		return AVL_BALANCE;
	}
}

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static int dataelement_compare(const tDcmElementHdr& ha, LPCVOID va, LPCVOID vb)
{
	tDcmElementHdr hb;
	hb.ReadHdr((const char* const)vb, true);

	int r = ha.gtag() - hb.gtag();
	if (r != 0)
		return r;

	r = ha.etag() - hb.etag();
	if (r != 0)
		return r;

	r = ha.value_length() - hb.value_length();
	if (r != 0)
		return r;

	r = memcmp(ha.value_ptr((const char* const)va), hb.value_ptr((const char* const)vb), ha.value_length());

	return r;
}
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
CDataElementCache::avlnode_t::AVLRES CDataElementCache::avlnode_t::avlinsert(avlnode_t** n, const tDcmElementHdr& hdr, LPCVOID d, avlnode_buffer_t& avlbuf, CDataElementCache* elementcache, DWORD& offset)
{
	AVLRES tmp;

	if (!(*n)) {
		avlnode_t* node = avlbuf.get_new_node();
		if (node == NULL)
			return AVL_ERROR;
		if (elementcache != NULL) {
			HRESULT hr = elementcache->append(hdr, d, offset, node->d);
			if (FAILED(hr))
				return AVL_ERROR;
		} else {
			node->d = d;
		}
		node->offset = offset;
		node->left = node->right = NULL;
		node->skew = AVL_NONE;
		*n = node;
		return AVL_BALANCE;
	}

	int cmp = dataelement_compare(hdr, d, (*n)->d);
	if (cmp < 0) {
		if ((tmp = avlinsert(&(*n)->left, hdr, d, avlbuf, elementcache, offset)) == AVL_BALANCE) {
			return avlleftgrown(n);
		}
		return tmp;
	} else if (cmp > 0) {
		if ((tmp = avlinsert(&(*n)->right, hdr, d, avlbuf, elementcache, offset)) == AVL_BALANCE) {
			return avlrightgrown(n);
		}
		return tmp;
	} else {
		offset = (*n)->offset;
		return AVL_HIT;
	}
}

//----------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
CDataElementCache::avlnode_buffer_t::~avlnode_buffer_t()
{
	free_blocks_<block_t>(m_first_block);
}
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
CDataElementCache::avlnode_t* CDataElementCache::avlnode_buffer_t::get_new_node(void)
{
	if (m_num_nodes_in_current_block == kNodesInBlock) {
		void* ptr = CoTaskMemAlloc(sizeof(block_t));
		if (ptr == NULL)
			return NULL;
		block_t* mb = new (ptr) block_t;
		m_current_block->m_next = mb;
		m_current_block = mb;
		m_num_nodes_in_current_block = 0;
	}
	avlnode_t* node = m_current_block->m_data + m_num_nodes_in_current_block++;
	return node;
}
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
void CDataElementCache::avlnode_buffer_t::recalc_pointer(const char* cache_begin)
{
	block_t* bp = &m_first_block;
	do {
		const DWORD count = m_current_block == bp ? m_num_nodes_in_current_block : kNodesInBlock;
		avlnode_t* node = bp->m_data;
		const avlnode_t* const end = node + count;
		for (; node < end; node++)
			node->d = static_cast<LPCVOID>(cache_begin + node->offset);
		bp = bp->m_next;
	} while (bp != NULL);
}

//======================================================================
//
//
//
//======================================================================
const char* CDataElementCache::sig = "DECh";
const wchar_t* CDataElementCache::filename = L".DataElm";
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
HRESULT CDataElementCacheFile::close(HRESULT hr)
{
	if (m_lpMap != NULL) {
		UnmapViewOfFile(m_lpMap);
		m_lpMap = NULL;
	}
	if (m_hMap != NULL) {
		CloseHandle(m_hMap);
		m_hMap = NULL;
	}
	if (m_hFile != INVALID_HANDLE_VALUE) {
		CloseHandle(m_hFile);
		m_hFile = INVALID_HANDLE_VALUE;
	}

	return hr;
}

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
void CDataElementCache::avlnode_buffer_t::free_blocks()
{
	free_blocks_<block_t>(m_first_block);
	m_first_block.m_next = NULL;
	m_current_block = &m_first_block;
	m_num_nodes_in_current_block = 0;
	m_num_nodes = 0;
}

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
void CDataElementCache::free_blocks()
{
	free_blocks_<block_t>(m_first_block);
	m_first_block.m_next = NULL;
	m_first_block.m_bytesInBlock = 0;
	m_current_block = &m_first_block;	
}

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
HRESULT CDataElementCache::append(const tDcmElementHdr& hdr, LPCVOID cp, DWORD& offset, LPCVOID& dataInCache)
{
	const unsigned long block_data_size = sizeof(m_current_block->m_data);
	const unsigned short value_length = hdr.value_length();
	const unsigned long size = 4 + 2 + 2 + value_length;

	if (size > block_data_size)
		return E_UNEXPECTED;

	const DWORD bufRemain = block_data_size - m_current_block->m_bytesInBlock;
	if (bufRemain < size) {
		void* ptr = CoTaskMemAlloc(sizeof(block_t));
		if (ptr == NULL)
			return E_OUTOFMEMORY;
		m_current_block->m_next = new (ptr) block_t;
		m_current_block = m_current_block->m_next;
	}

	BYTE* dp = m_current_block->m_data + m_current_block->m_bytesInBlock;
	CopyMemory(dp, cp, 4); dp += 4;
	CopyMemory(dp, hdr.m_vr, 2); dp += 2;
	CopyMemory(dp, &value_length, 2); dp += 2;
	CopyMemory(dp, hdr.value_ptr((char*)cp), value_length);

	offset = m_dataSize;
	dataInCache = static_cast<LPCVOID>(m_current_block->m_data + m_current_block->m_bytesInBlock);

	m_current_block->m_bytesInBlock += size;
	m_dataSize += size;

	return S_OK;
}

//======================================================================
//
//======================================================================
HRESULT CDataElementCache::CheckinElement(const tDcmElementHdr& hdr, LPCVOID cp, DWORD& offset)
{
	avlnode_t::AVLRES r = avlnode_t::avlinsert(&m_avltree_root, hdr, cp, m_avlnode_buffer, this, offset);
	if (r == avlnode_t::AVL_ERROR)
		return E_OUTOFMEMORY;

	return S_OK;
}

//======================================================================
//
//======================================================================
HRESULT CDataElementCache::Dump(const wchar_t* filePath)
{
	HANDLE hFile = CreateFile(filePath, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
		return HRESULT_FROM_WIN32(GetLastError());
	AutoCloseFileHandle ac_hFile(hFile);

	const DWORD dwFileSize = GetFileSize(hFile, NULL);
	if (dwFileSize == 0) {
		header_t h;
		ZeroMemory(&h, sizeof(h));
		CopyMemory(h.m_sig, sig, 4);
		h.m_headerSize = sizeof(header_t);
		h.m_dataSize = 0;
		DWORD bytesWritten;
		if (!WriteFile(hFile, &h, h.m_headerSize, &bytesWritten, NULL))
			return HRESULT_FROM_WIN32(GetLastError());
		if (SetFilePointer(hFile, 0, NULL, FILE_BEGIN) == INVALID_SET_FILE_POINTER)
			return HRESULT_FROM_WIN32(GetLastError());
	} else if (dwFileSize < sizeof(header_t)) {
		return E_UNEXPECTED;
	}

	OVERLAPPED ov = { 0, 0, 0, 0, 0 };
	if (!LockFileEx(hFile, LOCKFILE_EXCLUSIVE_LOCK, 0, sizeof(header_t), 0, &ov))
		return HRESULT_FROM_WIN32(GetLastError());
	AutoUnlockFileEx au(hFile, sizeof(header_t), 0, &ov);

	header_t header;

	DWORD bytesRead;
	if (!ReadFile(hFile, &header, sizeof(header), &bytesRead, NULL))
		return HRESULT_FROM_WIN32(GetLastError());

	if (SetFilePointer(hFile, 0, NULL, FILE_BEGIN) == INVALID_SET_FILE_POINTER)
		return HRESULT_FROM_WIN32(GetLastError());

	header.m_dataSize = 0;
	DWORD bytesWritten;
	if (!WriteFile(hFile, &header, sizeof(header), &bytesWritten, NULL))
		return HRESULT_FROM_WIN32(GetLastError());

	block_t* bp = &m_first_block;
	do {
		DWORD bytesWritten;
		if (!WriteFile(hFile, bp->m_data, bp->m_bytesInBlock, &bytesWritten, NULL)) {
			HRESULT hr = HRESULT_FROM_WIN32(GetLastError());
			if (SetFilePointer(hFile, sizeof(header), NULL, FILE_BEGIN) != INVALID_SET_FILE_POINTER)
				SetEndOfFile(hFile);
			return hr;
		}
		header.m_dataSize += bytesWritten;
		bp = bp->m_next;
	} while (bp != NULL);

	if (!SetEndOfFile(hFile))
		return HRESULT_FROM_WIN32(GetLastError());

	if (SetFilePointer(hFile, 0, NULL, FILE_BEGIN) == INVALID_SET_FILE_POINTER)
		return HRESULT_FROM_WIN32(GetLastError());

	if (!WriteFile(hFile, &header, sizeof(header), &bytesWritten, NULL))
		return HRESULT_FROM_WIN32(GetLastError());
	
	return S_OK;
}

//======================================================================
//
//======================================================================
HRESULT CDataElementCache::GetDOM(IXMLDOMDocument* dom)
{
	return block_list_to_dom<block_t>(&m_first_block, dom);
}

//======================================================================
//
//
//
//======================================================================
//======================================================================
//
//======================================================================
HRESULT CDataElementCacheFile::Open(const wchar_t* filePath)
{
	m_dataSize = 0;

	m_hFile = CreateFile(filePath, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (m_hFile == INVALID_HANDLE_VALUE)
		return close(HRESULT_FROM_WIN32(GetLastError()));

	const DWORD dwFileSize = GetFileSize(m_hFile, NULL);
	if (dwFileSize == 0) {
		header_t h;
		CopyMemory(h.m_sig, sig, 4);
		h.m_headerSize = sizeof(header_t);
		h.m_dataSize = 0;
		DWORD bytesWritten;
		if (!WriteFile(m_hFile, &h, h.m_headerSize, &bytesWritten, NULL))
			return close(HRESULT_FROM_WIN32(GetLastError()));
	} else if (dwFileSize < sizeof(header_t)) {
		return close(E_UNEXPECTED);
	}

	m_hMap = CreateFileMapping(m_hFile, NULL, PAGE_READWRITE, 0, 0, NULL);
	if (m_hMap == NULL)
		return close(HRESULT_FROM_WIN32(GetLastError()));

	m_lpMap = MapViewOfFile(m_hMap, FILE_MAP_ALL_ACCESS, 0, 0, 0);
	if (m_lpMap == NULL)
		return close(HRESULT_FROM_WIN32(GetLastError()));

	if (dwFileSize == 0)
		return S_OK;

	const header_t* header = static_cast<header_t*>(m_lpMap);
	if (memcmp(header->m_sig, sig, sizeof(header->m_sig)) != 0)
		return close(E_UNEXPECTED);
	if (dwFileSize < header->m_headerSize + header->m_dataSize)
		return close(E_UNEXPECTED);

	const char* const block_begin = static_cast<char*>(m_lpMap) + header->m_headerSize;
	const char* const block_end = block_begin + header->m_dataSize;

	CDcmScannerF scanner(block_begin, block_end);

	tDcmElementHdr hdr;
	const char* cp = block_begin;

	while (cp < block_end) {
		int sr = scanner.NextElement(cp, hdr, true);
		if (sr != CDcmScannerF::S_CONTINUE)
			break;

		DWORD offset = cp - block_begin;
		avlnode_t::AVLRES r = avlnode_t::avlinsert(&m_avltree_root, hdr, cp, m_avlnode_buffer, NULL, offset);
		if (r == avlnode_t::AVL_ERROR)
			return close(E_OUTOFMEMORY);

		cp = hdr.next_ptr(cp);
	}

	m_dataSize = header->m_dataSize;

	return S_OK;
}

//======================================================================
//
//======================================================================
HRESULT CDataElementCacheFile::Commit(bool bClose)
{
	if (m_first_block.m_bytesInBlock == 0)
		return S_FALSE;

	OVERLAPPED ov = { 0, 0, 0, 0, 0 };
	if (!LockFileEx(m_hFile, LOCKFILE_EXCLUSIVE_LOCK, 0, sizeof(header_t), 0, &ov))
		return close(HRESULT_FROM_WIN32(GetLastError()));
	AutoUnlockFileEx au(m_hFile, sizeof(header_t), 0, &ov);

	header_t* header = static_cast<header_t*>(m_lpMap);
	const long pos = header->m_headerSize + header->m_dataSize;

	UnmapViewOfFile(m_lpMap); m_lpMap = NULL;
	CloseHandle(m_hMap); m_hMap = NULL;

	if (SetFilePointer(m_hFile, pos, NULL, FILE_BEGIN) == INVALID_SET_FILE_POINTER)
		return close(HRESULT_FROM_WIN32(GetLastError()));

	block_t* bp = &m_first_block;
	do {
		DWORD bytesWritten;
		if (!WriteFile(m_hFile, bp->m_data, bp->m_bytesInBlock, &bytesWritten, NULL)) {
			HRESULT hr = HRESULT_FROM_WIN32(GetLastError());
			if (SetFilePointer(m_hFile, pos, NULL, FILE_BEGIN) != INVALID_SET_FILE_POINTER)
				SetEndOfFile(m_hFile);
			return close(hr);
		}
		bp = bp->m_next;
	} while (bp != NULL);

	free_blocks();

	if (!SetEndOfFile(m_hFile))
		return close(HRESULT_FROM_WIN32(GetLastError()));

	if (bClose) {
		HRESULT hr = S_OK;
		if (SetFilePointer(m_hFile, offsetof(header_t, m_dataSize), NULL, FILE_BEGIN) == INVALID_SET_FILE_POINTER) {
			hr = HRESULT_FROM_WIN32(GetLastError());
		} else {
			DWORD bytesWritten;
			if (!WriteFile(m_hFile, &m_dataSize, sizeof(m_dataSize), &bytesWritten, NULL)) {
				hr = HRESULT_FROM_WIN32(GetLastError());
			}
		}
		Close();
		return hr;
	}

	m_hMap = CreateFileMapping(m_hFile, NULL, PAGE_READWRITE, 0, 0, NULL);
	if (m_hMap == NULL)
		return close(HRESULT_FROM_WIN32(GetLastError()));

	m_lpMap = MapViewOfFile(m_hMap, FILE_MAP_ALL_ACCESS, 0, 0, 0);
	if (m_lpMap == NULL)
		return close(HRESULT_FROM_WIN32(GetLastError()));

	header = static_cast<header_t*>(m_lpMap);
	header->m_dataSize = m_dataSize;

	const char* const cache_begin = static_cast<char*>(m_lpMap) + header->m_headerSize;
	m_avlnode_buffer.recalc_pointer(cache_begin);
	
	return S_OK;
}

//======================================================================
//
//======================================================================
void CDataElementCacheFile::Close()
{
	m_avlnode_buffer.free_blocks();
	m_avltree_root = NULL;
	free_blocks();
	close(S_OK);
}

//======================================================================
//
//======================================================================
HRESULT CDataElementCacheFile::LoadDOM(const wchar_t* path, IXMLDOMDocument* dom)
{
	HANDLE hFile = CreateFile(path, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
		return HRESULT_FROM_WIN32(GetLastError());
	AutoCloseFileHandle ac_hFile(hFile);

	if (GetFileSize(hFile, NULL) < sizeof(header_t))
		return E_UNEXPECTED;

	OVERLAPPED ov = { 0, 0, 0, 0, 0 };
	if (!LockFileEx(hFile, 0, 0, sizeof(header_t), 0, &ov))
		return HRESULT_FROM_WIN32(GetLastError());
	AutoUnlockFileEx au(hFile, sizeof(header_t), 0, &ov);

	HANDLE hMap = CreateFileMapping(hFile, NULL, PAGE_READONLY, 0, 0, NULL);
	if (hMap == NULL)
		return HRESULT_FROM_WIN32(GetLastError());
	AutoCloseHandle ac_hMap(hMap);

	LPVOID lpMap = MapViewOfFile(hMap, FILE_MAP_READ, 0, 0, 0);
	if (lpMap == NULL)
		return HRESULT_FROM_WIN32(GetLastError());
	AutoUnmapViewOfFile au_lpMap(lpMap);

	const header_t* const header = static_cast<header_t*>(lpMap);
	if (memcmp(header->m_sig, sig, sizeof(header->m_sig)) != 0)
		return E_UNEXPECTED;
	if (GetFileSize(hFile, NULL) < header->m_headerSize + header->m_dataSize)
		return E_UNEXPECTED;

	struct block_t {
		block_t* m_next;
		DWORD m_bytesInBlock;
		BYTE* m_data;
	} block = { NULL, header->m_dataSize, (BYTE*)lpMap + header->m_headerSize };

	return block_list_to_dom<block_t>(&block, dom);
}

