/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "UtilClasses.h"

//======================================================================
//
// Response Buffer
//
//======================================================================
//======================================================================
//
//======================================================================
void CStackResponseBuffer::add(const char* str, size_t len)
{
	if (len == 0)
		len = strlen(str);

	m_buffer.append(str, len);
}

//======================================================================
//
//======================================================================
DWORD CStackResponseBuffer::flush(EXTENSION_CONTROL_BLOCK* ecb, const char* content_type, const char* extra_header_fields)
{
	EXTENSION_CONTROL_BLOCK *pECB = (EXTENSION_CONTROL_BLOCK*)ecb;

	//--------------------
	// Send HTTP Response
	//--------------------
	DWORD len;
	char httpHeader[512];
	sprintf(httpHeader, "Content-Type: %s\r\n\Content-Length: %d\r\n", content_type, m_buffer.data_size());
	if (extra_header_fields != NULL)
		strcat(httpHeader, extra_header_fields);
	strcat(httpHeader, "\r\n");
	len = strlen(httpHeader);
	if (pECB->ServerSupportFunction(pECB->ConnID, HSE_REQ_SEND_RESPONSE_HEADER,
		                      (LPVOID)"200 OK", &len, (LPDWORD)httpHeader) == false) {
		sprintf((char*)pECB->lpszLogData, "ServerSupportFunction() Error as %d", GetLastError());

		pECB->dwHttpStatusCode = HSE_STATUS_ERROR;
		return HSE_STATUS_ERROR;
	}

	//---------------
	// Send Contents
	//---------------
	size_t buffer_offset = 0;
	while (true) {
		size_t chunk_size;
		const char* chunk = m_buffer.peek(buffer_offset, chunk_size);
		if (chunk == NULL)
			break;
		buffer_offset += chunk_size;

		DWORD bytesWritten = chunk_size;
		pECB->WriteClient(pECB->ConnID, (LPVOID)chunk, &bytesWritten, HSE_IO_SYNC); 
	}

	pECB->dwHttpStatusCode = HSE_STATUS_SUCCESS;
	return HSE_STATUS_SUCCESS;
}

//======================================================================
//
//======================================================================
DWORD CStackResponseBuffer::flush(FILE* fp, const char* content_type, const char* extra_header_fields)
{
	fprintf(fp, "Content-Type: %s\r\n\Content-Length: %d\r\n", content_type, m_buffer.data_size());
	if (extra_header_fields != NULL)
		fprintf(fp, extra_header_fields);
	fprintf(fp, "\r\n");

	//---------------
	// Send Contents
	//---------------
	size_t buffer_offset = 0;
	while (true) {
		size_t chunk_size;
		const char* chunk = m_buffer.peek(buffer_offset, chunk_size);
		if (chunk == NULL)
			break;
		buffer_offset += chunk_size;

		fwrite(chunk, chunk_size, 1, fp);
	}

	return 0;
}

//======================================================================
//
// Virtual Memory Buffer
//
//======================================================================
//======================================================================
//
//======================================================================
CVirtualBuffer::CVirtualBuffer()
	: m_pvBaseAddress(NULL), m_dwReservedSize(0), m_dwLogicalCommitedSize(0), m_dwDataSize(0), m_dwCommitedSize(0), m_dwLastError(NOERROR)
{
}
	
//======================================================================
//
//======================================================================
CVirtualBuffer::~CVirtualBuffer()
{
	Release();
}
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
LPVOID CVirtualBuffer::commit_memory(const DWORD dwRequestedSize)
{
	DWORD dwNewDataSize = m_dwDataSize + dwRequestedSize;
	if (dwNewDataSize > m_dwCommitedSize) {
		DWORD dwToBeCommited = ((dwNewDataSize - m_dwCommitedSize + kPageSize - 1) / kPageSize) * kPageSize;
		LPVOID pvCommitBase = (LPVOID)((char*)m_pvBaseAddress + m_dwCommitedSize);
		LPVOID pv = VirtualAlloc(pvCommitBase, dwToBeCommited, MEM_COMMIT, PAGE_READWRITE);
		if (pv == NULL) {
			m_dwLastError = GetLastError();
			return NULL;
		}
		m_dwCommitedSize += dwToBeCommited;
	}

	LPVOID pv = (LPVOID)((char*)m_pvBaseAddress + m_dwDataSize);
	m_dwDataSize = dwNewDataSize;

	return pv;
}
//======================================================================
//
//======================================================================
BOOL CVirtualBuffer::Reserve(const DWORD dwReserveSize, const DWORD dwPreCommit)
{
	if (m_dwReservedSize) {
		m_dwLastError = 1;
		return FALSE;
	}

	LPVOID pv = VirtualAlloc(NULL, dwReserveSize, MEM_RESERVE, PAGE_READWRITE);
	if (pv == NULL) {
		m_dwLastError = GetLastError();
		return FALSE;
	}

	m_pvBaseAddress = pv;
	m_dwReservedSize = dwReserveSize;

	if (dwPreCommit != 0) {
		pv = VirtualAlloc(m_pvBaseAddress, dwPreCommit, MEM_COMMIT, PAGE_READWRITE);
		if (pv == NULL) {
			m_dwLastError = GetLastError();
			return FALSE;
		}
		m_dwCommitedSize = dwPreCommit;
	}

	return TRUE;
}

//======================================================================
//
//======================================================================
VOID CVirtualBuffer::Release(void)
{
	Decommit();
	if (m_pvBaseAddress) {
		VirtualFree(m_pvBaseAddress, m_dwReservedSize, MEM_RELEASE);
	}
	if (m_pvBaseAddress) {
		ZeroMemory(this, sizeof(CVirtualBuffer));
	}
}

//======================================================================
//
//======================================================================
LPVOID CVirtualBuffer::Append(LPCVOID pvData, const DWORD dwDataSize)
{
	LPVOID pv = this->commit_memory(dwDataSize);
	if (pv == NULL)
		return NULL;
	CopyMemory(pv, pvData, dwDataSize);
	return pv;
}

//======================================================================
//
//======================================================================
BOOL CVirtualBuffer::Commit(void)
{
	m_dwLogicalCommitedSize = m_dwDataSize;
	return TRUE;
}

//======================================================================
//
//======================================================================
void CVirtualBuffer::Decommit(void)
{
	if (m_dwCommitedSize) {
		VirtualFree(m_pvBaseAddress, m_dwCommitedSize, MEM_DECOMMIT);
	}
	m_dwLogicalCommitedSize = m_dwDataSize = m_dwCommitedSize = 0;
}

//======================================================================
//
//======================================================================
BOOL CVirtualBuffer::RollBack(void)
{
	m_dwDataSize = m_dwLogicalCommitedSize;
	// currently not actually decommit virtual memory, simply back the pointer
	return TRUE;
}

//======================================================================
//
//======================================================================
BOOL CVirtualBuffer::Read(LPVOID* ppvData, DWORD* pdwDataSize)
{
	*ppvData = m_pvBaseAddress;
	*pdwDataSize = m_dwDataSize;
	return TRUE;
}

//======================================================================
//
//======================================================================
VOID CVirtualBuffer::Protect(void)
{
	DWORD old;
	VirtualProtect(m_pvBaseAddress, m_dwCommitedSize, PAGE_READONLY, &old);
}

//======================================================================
//
//======================================================================
DWORD CVirtualBuffer::GetOffset(LPVOID pv)
{
	return (char*)pv - (char*)m_pvBaseAddress;
}

//======================================================================
//
//======================================================================
LPVOID CVirtualBuffer::GetPointer(DWORD offset)
{
	return (LPVOID)((char*)m_pvBaseAddress + offset);
}

//======================================================================
//
//======================================================================
DWORD CVirtualBuffer::GetDataSize(void)
{
	return m_dwDataSize;
}

//======================================================================
//
//======================================================================
DWORD CVirtualBuffer::lastError(void) const
{
	return m_dwLastError;
}

//======================================================================
//
// IStream implementation -- by Virtual Memory Buffer
//
//======================================================================
//======================================================================
//
//======================================================================
void* CVirtualBufferStream::operator new(size_t size)
{
	return CoTaskMemAlloc(size);
}
//======================================================================
//
//======================================================================
void CVirtualBufferStream::operator delete(void* deadObject, size_t size)
{
	CoTaskMemFree(deadObject);
}
//======================================================================
//
//======================================================================
CVirtualBufferStream::CVirtualBufferStream(DWORD size)
	: m_cRef(1), m_vb(), m_offset(0)
{
	m_vb.Reserve(size);
}
//======================================================================
//
//======================================================================
CVirtualBufferStream::~CVirtualBufferStream()
{
	m_vb.Release();
}

//======================================================================
//
//======================================================================
STDMETHODIMP_(ULONG) CVirtualBufferStream::AddRef(void)
{
	return InterlockedIncrement(&m_cRef);
}

//======================================================================
//
//======================================================================
STDMETHODIMP_(ULONG) CVirtualBufferStream::Release(void)
{
	LONG count = InterlockedDecrement(&m_cRef);

	if (count == 0)
		delete this;
	
	return count;
}

//======================================================================
//
//======================================================================
STDMETHODIMP CVirtualBufferStream::QueryInterface(REFIID riid, void** ppv)
{
	if (ppv == NULL)
		return E_INVALIDARG;
	if (riid == IID_IUnknown || riid == IID_IStream) {
		*ppv = this;
		AddRef();
		return S_OK;
	}
	*ppv = NULL;
	return E_NOINTERFACE;
}

//======================================================================
//
//======================================================================
STDMETHODIMP CVirtualBufferStream::Write(VOID const* pv, ULONG cb, ULONG* pcbWritten)
{
	if (m_vb.Append(pv, cb) == NULL)
		return STG_E_MEDIUMFULL;

	m_vb.Commit();
	*pcbWritten = cb;

	return S_OK;
}
//======================================================================
//
//======================================================================
STDMETHODIMP CVirtualBufferStream::Read(void *pv, ULONG cb, ULONG *pcbRead)
{
	*pcbRead = 0;

	ULONG remain = m_vb.GetDataSize() - m_offset;
	if (remain == 0)
		return S_FALSE;

	*pcbRead = remain < cb ? remain : cb;

	CopyMemory(pv, m_vb.GetPointer(m_offset), *pcbRead);

	m_offset += *pcbRead;

	return S_OK;
}

//======================================================================
//
//======================================================================
STDMETHODIMP CVirtualBufferStream::Seek(LARGE_INTEGER dbMove, DWORD dwOrigin, ULARGE_INTEGER *pbNewPosition)
{
	if (dbMove.HighPart != 0)
		return STG_E_INVALIDFUNCTION;

	pbNewPosition->HighPart = 0;

	switch (dwOrigin) {
	case STREAM_SEEK_SET:
		pbNewPosition->LowPart = dbMove.LowPart;
		break;
	case STREAM_SEEK_CUR:
		pbNewPosition->LowPart = m_offset + dbMove.LowPart;
		break;
	case STREAM_SEEK_END:
		pbNewPosition->LowPart = m_vb.GetDataSize() + dbMove.LowPart;
		break;
	}

	if (pbNewPosition->LowPart > m_vb.GetDataSize())
		return STG_E_INVALIDFUNCTION;

	m_offset = pbNewPosition->LowPart;

	return S_OK;
}

//======================================================================
//
// IStream implementation -- by File
//
//======================================================================
//======================================================================
//
//======================================================================
CStackFileStream::CStackFileStream(HANDLE& hFile)
	: m_cRef(1), m_hFile(hFile)
{
}
//======================================================================
//
//======================================================================
CStackFileStream::~CStackFileStream()
{
}
//======================================================================
//
//======================================================================
STDMETHODIMP CStackFileStream::QueryInterface(REFIID riid, void** ppv)
{
	if (ppv == NULL)
		return E_INVALIDARG;
	if (riid == IID_IUnknown || riid == IID_IStream) {
		*ppv = this;
		AddRef();
		return S_OK;
	}
	*ppv = NULL;
	return E_NOINTERFACE;
}

//======================================================================
//
//======================================================================
STDMETHODIMP CStackFileStream::Write(VOID const* pv, ULONG cb, ULONG* pcbWritten)
{
	return WriteFile(m_hFile, pv, cb, pcbWritten, NULL) ? S_OK : GetLastError();
}
//======================================================================
//
//======================================================================
STDMETHODIMP CStackFileStream::Read(void *pv, ULONG cb, ULONG *pcbRead)
{
	return ReadFile(m_hFile, pv, cb, pcbRead, NULL) ? S_OK : GetLastError();
}
//======================================================================
//
//======================================================================
STDMETHODIMP CStackFileStream::Seek(LARGE_INTEGER dbMove, DWORD dwOrigin, ULARGE_INTEGER *pbNewPosition)
{
	DWORD dwMoveMethod;
	switch (dwOrigin) {
	case STREAM_SEEK_SET:
		dwMoveMethod = FILE_BEGIN;
		break;
	case STREAM_SEEK_CUR:
		dwMoveMethod = FILE_CURRENT;
		break;
	case STREAM_SEEK_END:
		dwMoveMethod = FILE_END;
		break;
	}

	if (SetFilePointerEx(m_hFile, dbMove, (PLARGE_INTEGER)pbNewPosition, dwMoveMethod)) {
		return S_OK;
	} else {
		return HRESULT_FROM_WIN32(GetLastError());
	}
}


//======================================================================
//
// IStream implementation -- by Link List of Block Memory Buffer
//
//======================================================================
//======================================================================
//
//======================================================================
void* CBlockBufferStream::operator new(size_t size)
{
	return CoTaskMemAlloc(size);
}
//======================================================================
//
//======================================================================
void CBlockBufferStream::operator delete(void* deadObject, size_t size)
{
	CoTaskMemFree(deadObject);
}
//======================================================================
//
//======================================================================
CBlockBufferStream::CBlockBufferStream()
	: m_cRef(1), m_data_size(0), m_offset(0), m_current_mb(&m_block), m_buf_offset(0)
{
}
//======================================================================
//
//======================================================================
CBlockBufferStream::~CBlockBufferStream()
{
	block_t* mb = m_block.m_next;
	while (mb != NULL) {
		block_t* next = mb->m_next;
		CoTaskMemFree(mb);
		mb = next;
	}
}

//======================================================================
//
//======================================================================
STDMETHODIMP_(ULONG) CBlockBufferStream::AddRef(void)
{
	return InterlockedIncrement(&m_cRef);
}

//======================================================================
//
//======================================================================
STDMETHODIMP_(ULONG) CBlockBufferStream::Release(void)
{
	LONG count = InterlockedDecrement(&m_cRef);

	if (count == 0)
		delete this;
	
	return count;
}

//======================================================================
//
//======================================================================
STDMETHODIMP CBlockBufferStream::QueryInterface(REFIID riid, void** ppv)
{
	if (ppv == NULL)
		return E_INVALIDARG;
	if (riid == IID_IUnknown || riid == IID_IStream) {
		*ppv = this;
		AddRef();
		return S_OK;
	}
	*ppv = NULL;
	return E_NOINTERFACE;
}

//======================================================================
//
//======================================================================
STDMETHODIMP CBlockBufferStream::Write(VOID const* pv, ULONG cb, ULONG* pcbWritten)
{
	BYTE* src = (BYTE*)pv;
	*pcbWritten = 0;

	while (cb > 0) {
		if (m_buf_offset == kBlockSize) {
			if (m_current_mb->m_next != NULL) {
				m_current_mb = m_current_mb->m_next;
			} else {
				block_t* mb = (block_t*)CoTaskMemAlloc(sizeof(block_t));
				if (mb == NULL) {
					return STG_E_MEDIUMFULL;
				}
				mb->m_next = NULL;
				m_current_mb->m_next = mb;
				m_current_mb = mb;
			}
			m_buf_offset = 0;
		}
		ULONG buf_remain = kBlockSize - m_buf_offset;

		ULONG copy_count = buf_remain < cb ? buf_remain : cb;

		CopyMemory(m_current_mb->m_data + m_buf_offset, src, copy_count);

		src += copy_count;

		*pcbWritten += copy_count;

		m_buf_offset += copy_count;
		m_offset += copy_count;
		if (m_offset > m_data_size)
			m_data_size = m_offset;

		cb -= copy_count;
	}

	return S_OK;
}
//======================================================================
//
//======================================================================
STDMETHODIMP CBlockBufferStream::Read(void *pv, ULONG cb, ULONG *pcbRead)
{
	BYTE* dst = (BYTE*)pv;
	*pcbRead = 0;

	if (cb == 0)
		return S_OK;

	ULONG data_remain = m_data_size - m_offset;

	if (data_remain == 0)
		return S_FALSE;

	while (cb > 0 && data_remain > 0) {
		if (m_buf_offset == kBlockSize) {
			if (m_current_mb->m_next == NULL)
				break;
			m_current_mb = m_current_mb->m_next;
			m_buf_offset = 0;
		}
		ULONG buf_remain = kBlockSize - m_buf_offset;
		if (data_remain < buf_remain)
			buf_remain = data_remain;

		ULONG copy_count = buf_remain < cb ? buf_remain : cb;

		CopyMemory(dst, m_current_mb->m_data + m_buf_offset, copy_count);

		dst += copy_count;

		*pcbRead += copy_count;

		m_buf_offset += copy_count;
		m_offset += copy_count;

		cb -= copy_count;
		data_remain -= copy_count;
	}

	return S_OK;
}

//======================================================================
//
//======================================================================
STDMETHODIMP CBlockBufferStream::Seek(LARGE_INTEGER dbMove, DWORD dwOrigin, ULARGE_INTEGER *pbNewPosition)
{
	if (dbMove.HighPart != 0)
		return STG_E_INVALIDFUNCTION;

	pbNewPosition->HighPart = 0;

	switch (dwOrigin) {
	case STREAM_SEEK_SET:
		pbNewPosition->LowPart = dbMove.LowPart;
		break;
	case STREAM_SEEK_CUR:
		pbNewPosition->LowPart = m_offset + dbMove.LowPart;
		break;
	case STREAM_SEEK_END:
		pbNewPosition->LowPart = m_data_size + dbMove.LowPart;
		break;
	}

	if (pbNewPosition->LowPart > m_data_size)
		return STG_E_INVALIDFUNCTION;

	if (pbNewPosition->LowPart >= m_offset - m_buf_offset
	 && pbNewPosition->LowPart <= m_offset - m_buf_offset + kBlockSize) {
		m_buf_offset += pbNewPosition->LowPart - m_offset;
		m_offset = pbNewPosition->LowPart;
	} else {
		if (pbNewPosition->LowPart < m_offset) {
			m_offset = 0;
			m_buf_offset = 0;
			m_current_mb = &m_block;
		}
		while (m_offset < pbNewPosition->LowPart) {
			ULONG buf_remain = kBlockSize - m_buf_offset;
			ULONG diff = pbNewPosition->LowPart - m_offset;
			if (diff > buf_remain) {
				m_current_mb = m_current_mb->m_next;
				m_buf_offset = 0;
				m_offset += buf_remain;
			} else {
				m_offset += diff;
				m_buf_offset += diff;
			}
		}
	}

	return S_OK;
}
