/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "UtilClasses.h"

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static HRESULT cstr2bstr(const char* cstr, int length, BSTR& bstr, SimpleErrorInfo* ei) throw (SimpleErrorInfo)
{
	bool is_mb = false;
	const unsigned char* cp = (const unsigned char*)cstr;
	if (length < 0) {
		length = 0;
		for (; *cp != '\0'; cp++) {
			length++;
			is_mb |= *cp >= 0x7F;
		}
	} else {
		const unsigned char* const ep = cp + length;
		for (; cp < ep; cp++) {
			is_mb |= *cp >= 0x7F;
		}
	}

	if (length == 0) {
		SysReAllocString(&bstr, L"");
		return S_OK;
	}

	if (is_mb) {
		int wclen = multi_byte_to_wide_char(cstr, length, NULL, 0);
		if (wclen == 0) {
			const HRESULT hr = HRESULT_FROM_WIN32(GetLastError());
			if (ei == xduThrowSimpleErrorInfo) {
				throw SimpleErrorInfo(hr, L"MultiByteToWideChar(cstr, %1!d!, NULL, 0)", length);
			} else if (ei != xduDontSetSimpleErrorInfo) {
				ei->set(hr, L"MultiByteToWideChar()");
			}
			return hr;
		}
		if (!SysReAllocStringLen(&bstr, NULL, wclen)) {
			if (ei == xduThrowSimpleErrorInfo) {
				throw SimpleErrorInfo(E_OUTOFMEMORY, L"SysReAllocStringLen(,,%1!d!)", wclen);
			} else if (ei != xduDontSetSimpleErrorInfo) {
				ei->set(E_OUTOFMEMORY, L"SysReAllocStringLen(,,%1!d!)", wclen);
			}
			return E_OUTOFMEMORY;
		}
		wclen = multi_byte_to_wide_char(cstr, length, bstr, wclen);
		if (wclen == 0) {
			const HRESULT hr = HRESULT_FROM_WIN32(GetLastError());
			if (ei == xduThrowSimpleErrorInfo) {
				throw SimpleErrorInfo(hr, L"MultiByteToWideChar(cstr,  %1!d!, bstr, wclen)", length);
			} else if (ei != xduDontSetSimpleErrorInfo) {
				ei->set(hr, L"MultiByteToWideChar()");
			}
			return hr;
		}
	} else {
		if (!SysReAllocStringLen(&bstr, NULL, length)) {
			if (ei == xduThrowSimpleErrorInfo) {
				throw SimpleErrorInfo(E_OUTOFMEMORY, L"SysReAllocStringLen(,,%1!d!)", length);
			} else if (ei != xduDontSetSimpleErrorInfo) {
				ei->set(E_OUTOFMEMORY, L"SysReAllocStringLen(,,%1!d!)", length);
			}
			return E_OUTOFMEMORY;
		}
		const char* sp = cstr;
		const char* ep = cstr + length;
		wchar_t* dp = bstr;
		while (sp < ep)
			*dp++ = *sp++;
	}

	return S_OK;
}

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static HRESULT makeAttribute(IXMLDOMDocument* dom, IXMLDOMElement* elm, const wchar_t* name, const wchar_t* info, BSTR& bstr, IXMLDOMAttribute** newAttribute, SimpleErrorInfo* ei)  throw (SimpleErrorInfo)
{
	HRESULT hr;

	bstr = SysAllocString(name);
	if (bstr == NULL) {
		if (ei == xduThrowSimpleErrorInfo) {
			throw SimpleErrorInfo(E_OUTOFMEMORY, L"SysAllocString(%1)", name);
		}else if (ei != xduDontSetSimpleErrorInfo) {
			ei->set(E_OUTOFMEMORY, L"SysAllocString(%1)", name);
			ei->add(L"xduMakeAttribute(%1, %2)", name, info);
		}
		return E_OUTOFMEMORY;
	}

	AutoComPtr<IXMLDOMAttribute> attr;
	hr = dom->createAttribute(bstr, attr.r());
	if (FAILED(hr)) {
		if (ei == xduThrowSimpleErrorInfo) {
			throw SimpleErrorInfo(hr, L"IXMLDOMDocument::createAttribute()");
		}else if (ei != xduDontSetSimpleErrorInfo) {
			ei->set(hr, L"IXMLDOMDocument::createAttribute()");
			ei->add(L"xduMakeAttribute(%1, %2)", name, info);
		}
		return hr;
	}
	hr = elm->setAttributeNode(attr.p(), NULL);
	if (FAILED(hr)) {
		if (ei == xduThrowSimpleErrorInfo) {
			throw SimpleErrorInfo(hr, L"IXMLDOMElement::setAttributeNode()");
		} else if (ei != xduDontSetSimpleErrorInfo) {
			ei->set(hr, L"IXMLDOMElement::setAttributeNode()");
			ei->add(L"xduMakeAttribute(%1, %2)", name, info);
		}
		return hr;
	}

	attr->AddRef();
	*newAttribute = attr.p();
	
	return S_OK;
}

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static HRESULT put_value(IXMLDOMAttribute* attr, VARIANT& var, const wchar_t* name, const wchar_t* info, SimpleErrorInfo* ei)  throw (SimpleErrorInfo)
{
	HRESULT hr = attr->put_value(var);
	if (FAILED(hr)) {
		if (ei == xduThrowSimpleErrorInfo) {
			throw SimpleErrorInfo(hr, L"IXMLDOMAttribute::put_value()");
		} else if (ei != xduDontSetSimpleErrorInfo) {
			ei->set(hr, L"IXMLDOMAttribute::put_value()");
			ei->add(L"xduMakeAttribute(%1, %2)", name, info);
		}
		return hr;
	}
	return S_OK;
}

//======================================================================
//
//======================================================================
HRESULT xduMakeAttribute(IXMLDOMDocument* dom, IXMLDOMElement* elm, const wchar_t* name, const wchar_t* value, SimpleErrorInfo* ei) throw (SimpleErrorInfo)
{
	try {
		HRESULT hr;

		const wchar_t* const info = L"<const wchar_t*>";

		BSTR bstr = NULL;
		AutoSysFreeString af_bstr(bstr);
		AutoComPtr<IXMLDOMAttribute> attr;

		hr = makeAttribute(dom, elm, name, info, bstr, attr.r(), ei);
		if (FAILED(hr))
			return hr;

		if (!SysReAllocString(&bstr, value)) {
			if (ei == xduThrowSimpleErrorInfo) {
				throw SimpleErrorInfo(E_OUTOFMEMORY, L"SysReAllocString(%1)", value);
			} else if (ei != xduDontSetSimpleErrorInfo) {
				ei->set(E_OUTOFMEMORY, L"SysReAllocString(%1)", value);
				ei->add(L"xduMakeAttribute(%1, <const wchar_t*>)", name);
			}
			return E_OUTOFMEMORY;
		}

		VARIANT var;
		VariantInit(&var);

		var.vt = VT_BSTR;
		var.bstrVal = bstr;

		hr = put_value(attr.p(), var, name, info, ei);
		if (FAILED(hr))
			return hr;

		return S_OK;

	} catch (SimpleErrorInfo& ei) {
		ei.add(L"xduMakeAttribute(%1, <const wchar_t*>)", name);
		throw;
	}

	return E_UNEXPECTED;
}


//======================================================================
//
//======================================================================
HRESULT xduMakeAttribute(IXMLDOMDocument* dom, IXMLDOMElement* elm, const wchar_t* name, const char* value, SimpleErrorInfo* ei) throw (SimpleErrorInfo)
{
	try {
		HRESULT hr;

		const wchar_t* const info = L"<const char*>";

		BSTR bstr = NULL;
		AutoSysFreeString af_bstr(bstr);
		AutoComPtr<IXMLDOMAttribute> attr;

		hr = makeAttribute(dom, elm, name, info, bstr, attr.r(), ei);
		if (FAILED(hr))
			return hr;

		hr = cstr2bstr(value, -1, bstr, ei);
		if (FAILED(hr))
			return hr;

		VARIANT var;
		VariantInit(&var);

		var.vt = VT_BSTR;
		var.bstrVal = bstr;

		hr = put_value(attr.p(), var, name, info, ei);
		if (FAILED(hr))
			return hr;

		return S_OK;

	} catch (SimpleErrorInfo& ei) {
		ei.add(L"xduMakeAttribute(%1, <const char*>)", name);
		throw;
	}

	return E_UNEXPECTED;
}

//======================================================================
//
//======================================================================
HRESULT xduMakeAttribute(IXMLDOMDocument* dom, IXMLDOMElement* elm, const wchar_t* name, unsigned int value, SimpleErrorInfo* ei) throw (SimpleErrorInfo)
{
	try {
		HRESULT hr;

		const wchar_t* const info = L"<unsigned int>";

		BSTR bstr = NULL;
		AutoSysFreeString af_bstr(bstr);
		AutoComPtr<IXMLDOMAttribute> attr;

		hr = makeAttribute(dom, elm, name, info, bstr, attr.r(), ei);
		if (FAILED(hr))
			return hr;

		VARIANT var;
		VariantInit(&var);

		var.vt = VT_UI4;
		var.ulVal = value;

		hr = put_value(attr.p(), var, name, info, ei);
		if (FAILED(hr))
			return hr;

		return S_OK;

	} catch (SimpleErrorInfo& ei) {
		ei.add(L"xduMakeAttribute(%1, <unsigned int>)", name);
		throw;
	}

	return E_UNEXPECTED;
}

//======================================================================
//
//======================================================================
HRESULT xduMakeAttribute(IXMLDOMDocument* dom, IXMLDOMElement* elm, const wchar_t* name, double value, SimpleErrorInfo* ei) throw (SimpleErrorInfo)
{
	try {
		HRESULT hr;

		const wchar_t* const info = L"<double>";

		BSTR bstr = NULL;
		AutoSysFreeString af_bstr(bstr);
		AutoComPtr<IXMLDOMAttribute> attr;

		hr = makeAttribute(dom, elm, name, info, bstr, attr.r(), ei);
		if (FAILED(hr))
			return hr;

		VARIANT var;
		VariantInit(&var);

		var.vt = VT_R8;
		var.dblVal = value;

		hr = put_value(attr.p(), var, name, info, ei);
		if (FAILED(hr))
			return hr;

		return S_OK;

	} catch (SimpleErrorInfo& ei) {
		ei.add(L"xduMakeAttribute(%1, <double>)", name);
		throw;
	}

	return E_UNEXPECTED;
}

//======================================================================
//
//======================================================================
HRESULT xduMakeAttribute(IXMLDOMDocument* dom, IXMLDOMElement* elm, const wchar_t* name, const SYSTEMTIME* time, SimpleErrorInfo* ei) throw (SimpleErrorInfo)
{
	try {
		HRESULT hr;

		const wchar_t* const info = L"<SYSTEMTIME>";

		BSTR bstr = NULL;
		AutoSysFreeString af_bstr(bstr);
		AutoComPtr<IXMLDOMAttribute> attr;

		hr = makeAttribute(dom, elm, name, info, bstr, attr.r(), ei);
		if (FAILED(hr))
			return hr;

		SYSTEMTIME sysTime;
		if (time == NULL) {
			GetLocalTime(&sysTime);
			time = &sysTime;
		}

		double varTime;
		SystemTimeToVariantTime(const_cast<SYSTEMTIME*>(time), &varTime);

		VARIANT var;
		VariantInit(&var);

		var.vt = VT_DATE;
		var.date = varTime;

		hr = put_value(attr.p(), var, name, info, ei);
		if (FAILED(hr))
			return hr;

		return S_OK;

	} catch (SimpleErrorInfo& ei) {
		ei.add(L"xduMakeAttribute(%1, <double>)", name);
		throw;
	}

	return E_UNEXPECTED;
}

//======================================================================
//
//======================================================================
HRESULT xduMakeElement(IXMLDOMDocument* dom, IXMLDOMElement* parent, const wchar_t* name, IXMLDOMElement** newElement, SimpleErrorInfo* ei) throw (SimpleErrorInfo)
{
	try {
		HRESULT hr;

		AutoComPtr<IXMLDOMElement> tmpElm;
		if (parent == NULL) {
			hr = dom->get_documentElement(tmpElm.r());
			if (FAILED(hr)) {
				if (ei == xduThrowSimpleErrorInfo) {
					throw SimpleErrorInfo(hr, L"IXMLDOMDocument::get_documentElement()");
				} else if (ei != xduDontSetSimpleErrorInfo) {
					ei->set(hr, L"IXMLDOMDocument::get_documentElement()");
					ei->add(L"xduMakeElement(%1)", name);
				}
				return hr;
			}
			parent = tmpElm.p();
		}

		BSTR bstr = SysAllocString(name);
		if (bstr == NULL) {
			if (ei == xduThrowSimpleErrorInfo) {
				throw SimpleErrorInfo(E_OUTOFMEMORY, L"SysAllocString(%1)");
			} else if (ei != xduDontSetSimpleErrorInfo) {
				ei->set(E_OUTOFMEMORY, L"SysAllocString(%1)");
				ei->add(L"xduMakeElement(%1)", name);
			}
			return E_OUTOFMEMORY;
		}
		AutoSysFreeString af_bstr(bstr);

		AutoComPtr<IXMLDOMElement> elm;
		hr = dom->createElement(bstr, elm.r());
		if (FAILED(hr)) {
			if (ei == xduThrowSimpleErrorInfo) {
				throw SimpleErrorInfo(hr, L"IXMLDOMDocument::createElement()");
			} else if (ei != xduDontSetSimpleErrorInfo) {
				ei->set(hr, L"IXMLDOMDocument::createElement()");
				ei->add(L"xduMakeElement(%1)", name);
			}
			return hr;
		}

		hr = parent->appendChild(elm.p(), NULL);
		if (FAILED(hr)) {
			if (ei == xduThrowSimpleErrorInfo) {
				throw SimpleErrorInfo(hr, L"IXMLDOMElement::appendChild()");
			} else if (ei != xduDontSetSimpleErrorInfo) {
				ei->set(hr, L"IXMLDOMElement::appendChild()");
				ei->add(L"xduMakeElement(%1)", name);
			}
			return hr;
		}

		if (newElement != NULL) {
			elm->AddRef();
			*newElement = elm.p();
		}
	} catch (SimpleErrorInfo& ei) {
		ei.add(L"xduMakeElement(%1)", name);
		throw;
	}
	return S_OK;

}

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static HRESULT xduAddTextIntrinsic(IXMLDOMDocument* dom, IXMLDOMElement* elm, const void* text, int length, bool isWC, SimpleErrorInfo* ei) throw (SimpleErrorInfo)
{
	HRESULT hr;

	BSTR bstr = NULL;
	AutoSysFreeString af_bstr(bstr);

	if (isWC) {
		if (length < 0) {
			bstr = SysAllocString(static_cast<const wchar_t*>(text));
			if (bstr == NULL) {
				if (ei == xduThrowSimpleErrorInfo)
					throw SimpleErrorInfo(E_OUTOFMEMORY, L"SysAllocString()");
				else if (ei != xduDontSetSimpleErrorInfo)
					ei->set(E_OUTOFMEMORY, L"SysAllocString()");
				return E_OUTOFMEMORY;
			}
		} else {
			bstr = SysAllocStringLen(static_cast<const wchar_t*>(text), length);
			if (bstr == NULL) {
				if (ei == xduThrowSimpleErrorInfo)
					throw SimpleErrorInfo(E_OUTOFMEMORY, L"SysAllocStringLen()");
				else if (ei != xduDontSetSimpleErrorInfo)
					ei->set(E_OUTOFMEMORY, L"SysAllocStringLen()");
				return E_OUTOFMEMORY;
			}
		}
	} else {
		hr = cstr2bstr(static_cast<const char*>(text), length, bstr, ei);
		if (FAILED(hr))
			return hr;
	}

	AutoComPtr<IXMLDOMText> textElm;
	hr = dom->createTextNode(bstr, textElm.r());
	if (FAILED(hr)) {
		if (ei == xduThrowSimpleErrorInfo)
			throw SimpleErrorInfo(hr, L"IXMLDOMDocument::createTextNode()");
		else if (ei != xduDontSetSimpleErrorInfo)
			ei->set(hr, L"IXMLDOMDocument::createTextNode()");
		return hr;
	}

	hr = elm->appendChild(textElm.p(), NULL);
	if (FAILED(hr)) {
		if (ei == xduThrowSimpleErrorInfo)
			throw SimpleErrorInfo(hr, L"IXMLDOMElement::appendChild()");
		else if (ei != xduDontSetSimpleErrorInfo)
			ei->set(hr, L"IXMLDOMElement::appendChild()");
		return hr;
	}
	return S_OK;
}

//======================================================================
//
//======================================================================
HRESULT xduAddText(IXMLDOMDocument* dom, IXMLDOMElement* elm, const wchar_t* text, SimpleErrorInfo* ei) throw (SimpleErrorInfo)
{
	try {
		const HRESULT hr = xduAddTextIntrinsic(dom, elm, (const void*)text, -1, true, ei);
		if (FAILED(hr) && ei != xduDontSetSimpleErrorInfo)
			ei->add(L"xduAddText(<const wchar_t>)");
		return hr;
	} catch (SimpleErrorInfo& ei) {
		ei.add(L"xduAddText(<const wchar_t>)");
		throw;
	}
	return E_UNEXPECTED;
}

//======================================================================
//
//======================================================================
HRESULT xduAddText(IXMLDOMDocument* dom, IXMLDOMElement* elm, const char* text, SimpleErrorInfo* ei) throw (SimpleErrorInfo)
{
	try {
		const HRESULT hr = xduAddTextIntrinsic(dom, elm, (const void*)text, -1, false, ei);
		if (FAILED(hr) && ei != xduDontSetSimpleErrorInfo)
			ei->add(L"xduAddText(<const char*>)");
		return hr;
	} catch (SimpleErrorInfo& ei) {
		ei.add(L"xduAddText(<const char*>)");
		throw;
	}
	return E_UNEXPECTED;
}
//======================================================================
//
//======================================================================
HRESULT xduAddText(IXMLDOMDocument* dom, IXMLDOMElement* elm, const wchar_t* text, int textLength, SimpleErrorInfo* ei) throw (SimpleErrorInfo)
{
	try {
		const HRESULT hr = xduAddTextIntrinsic(dom, elm, (const void*)text, textLength, true, ei);
		if (FAILED(hr) && ei != xduDontSetSimpleErrorInfo)
			ei->add(L"xduAddText(<const wchar_t>)");
		return hr;
	} catch (SimpleErrorInfo& ei) {
		ei.add(L"xduAddText(<const wchar_t>)");
		throw;
	}
	return E_UNEXPECTED;
}

//======================================================================
//
//======================================================================
HRESULT xduAddText(IXMLDOMDocument* dom, IXMLDOMElement* elm, const char* text, int textLength, SimpleErrorInfo* ei) throw (SimpleErrorInfo)
{
	try {
		const HRESULT hr = xduAddTextIntrinsic(dom, elm, (const void*)text, textLength, false, ei);
		if (FAILED(hr) && ei != xduDontSetSimpleErrorInfo)
			ei->add(L"xduAddText(<const char*>)");
		return hr;
	} catch (SimpleErrorInfo& ei) {
		ei.add(L"xduAddText(<const char*>)");
		throw;
	}
	return E_UNEXPECTED;
}

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static HRESULT xduMakeElementIntrinsic(IXMLDOMDocument* dom, IXMLDOMElement* parent, const wchar_t* name, const void* value, int length, bool isWC, IXMLDOMElement** newElement, SimpleErrorInfo* ei) throw (SimpleErrorInfo)
{
	HRESULT hr;

	AutoComPtr<IXMLDOMElement> tmpElm;
	if (parent == NULL) {
		hr = dom->get_documentElement(tmpElm.r());
		if (FAILED(hr)) {
			if (ei == xduThrowSimpleErrorInfo)
				throw SimpleErrorInfo(hr, L"IXMLDOMDocument::get_documentElement()");
			else if (ei != xduDontSetSimpleErrorInfo)
				ei->set(hr, L"IXMLDOMDocument::get_documentElement()");
			return hr;
		}
		parent = tmpElm.p();
	}

	AutoComPtr<IXMLDOMElement> elm;
	hr = xduMakeElement(dom, parent, name, elm.r(), ei);
	if (FAILED(hr))
		return hr;

	hr = xduAddTextIntrinsic(dom, elm.p(), value, length, isWC, ei);
	if (FAILED(hr))
		return hr;

	if (newElement != NULL) {
		elm->AddRef();
		*newElement = elm.p();
	}

	return S_OK;
}

//======================================================================
//
//======================================================================
HRESULT xduMakeElement(IXMLDOMDocument* dom, IXMLDOMElement* parent, const wchar_t* name, const wchar_t* value, IXMLDOMElement** newElement, SimpleErrorInfo* ei) throw (SimpleErrorInfo)
{
	try {
		const HRESULT hr = xduMakeElementIntrinsic(dom, parent, name, (const void*)value, -1, true, newElement, ei);
		if (FAILED(hr) && ei != xduDontSetSimpleErrorInfo)
			ei->add(L"xduMakeElement(%1, <const wchar_t*>)", name);
		return hr;
	} catch (SimpleErrorInfo& ei) {
		ei.add(L"xduMakeElement(%1, <const wchar_t*>)", name);
		throw;
	}
	return E_UNEXPECTED;
}
//======================================================================
//
//======================================================================
HRESULT xduMakeElement(IXMLDOMDocument* dom, IXMLDOMElement* parent, const wchar_t* name, const char* value, IXMLDOMElement** newElement, SimpleErrorInfo* ei) throw (SimpleErrorInfo)
{
	try {
		const HRESULT hr = xduMakeElementIntrinsic(dom, parent, name, (const void*)value, -1, false ,newElement, ei);
		if (FAILED(hr) && ei != xduDontSetSimpleErrorInfo)
			ei->add(L"xduMakeElement(%1, <const char*>)", name);
		return hr;
	} catch (SimpleErrorInfo& ei) {
		ei.add(L"xduMakeElement(%1, <const char*>)", name);
		throw;
	}
	return E_UNEXPECTED;
}
//======================================================================
//
//======================================================================
HRESULT xduMakeElement(IXMLDOMDocument* dom, IXMLDOMElement* parent, const wchar_t* name, const wchar_t* value, int valueLength, IXMLDOMElement** newElement, SimpleErrorInfo* ei) throw (SimpleErrorInfo)
{
	try {
		const HRESULT hr = xduMakeElementIntrinsic(dom, parent, name, (const void*)value, valueLength, true, newElement, ei);
		if (FAILED(hr) && ei != xduDontSetSimpleErrorInfo)
			ei->add(L"xduMakeElement(%1, <const wchar_t*>)", name);
		return hr;
	} catch (SimpleErrorInfo& ei) {
		ei.add(L"xduMakeElement(%1, <const wchar_t*>)", name);
		throw;
	}
	return E_UNEXPECTED;
}
//======================================================================
//
//======================================================================
HRESULT xduMakeElement(IXMLDOMDocument* dom, IXMLDOMElement* parent, const wchar_t* name, const char* value, int valueLength, IXMLDOMElement** newElement, SimpleErrorInfo* ei) throw (SimpleErrorInfo)
{
	try {
		const HRESULT hr = xduMakeElementIntrinsic(dom, parent, name, (const void*)value, valueLength, false ,newElement, ei);
		if (FAILED(hr) && ei != xduDontSetSimpleErrorInfo)
			ei->add(L"xduMakeElement(%1, <const char*>)", name);
		return hr;
	} catch (SimpleErrorInfo& ei) {
		ei.add(L"xduMakeElement(%1, <const char*>)", name);
		throw;
	}
	return E_UNEXPECTED;
}
//======================================================================
//
//======================================================================
HRESULT xduMakeElement(IXMLDOMDocument* dom, IXMLDOMElement* parent, const wchar_t* name, unsigned int value, IXMLDOMElement** newElement, SimpleErrorInfo* ei) throw (SimpleErrorInfo)
{
	try {
		VARIANT var;
		VariantInit(&var);
		AutoVariantClear ac_var(var);

		var.vt = VT_UI4;
		var.ulVal = value;

		HRESULT hr = VariantChangeType(&var, &var, 0, VT_BSTR);
		if (FAILED(hr)) {
			if (ei == xduThrowSimpleErrorInfo)
				throw SimpleErrorInfo(hr, L"VariantChangeType(VT_UI4 -> VT_BSTR)");
			else if (ei != xduDontSetSimpleErrorInfo)
				ei->set(hr, L"VariantChangeType(VT_UI4 -> VT_BSTR)");
			return hr;
		}

		hr = xduMakeElementIntrinsic(dom, parent, name, (const void*)var.bstrVal, SysStringLen(var.bstrVal), true ,newElement, ei);
		if (FAILED(hr) && ei != xduDontSetSimpleErrorInfo)
			ei->add(L"xduMakeElement(%1, <unsigned int>)", name);
		return hr;
	} catch (SimpleErrorInfo& ei) {
		ei.add(L"xduMakeElement(%1, <unsigned int>)", name);
		throw;
	}
	return E_UNEXPECTED;
}
//======================================================================
//
//======================================================================
HRESULT xduMakeElement(IXMLDOMDocument* dom, IXMLDOMElement* parent, const wchar_t* name, double value, IXMLDOMElement** newElement, SimpleErrorInfo* ei) throw (SimpleErrorInfo)
{
	try {
		VARIANT var;
		VariantInit(&var);
		AutoVariantClear ac_var(var);

		var.vt = VT_R8;
		var.dblVal = value;

		HRESULT hr = VariantChangeType(&var, &var, 0, VT_BSTR);
		if (FAILED(hr)) {
			if (ei == xduThrowSimpleErrorInfo)
				throw SimpleErrorInfo(hr, L"VariantChangeType(VT_R8 -> VT_BSTR)");
			else if (ei != xduDontSetSimpleErrorInfo)
				ei->set(hr, L"VariantChangeType(VT_R8 -> VT_BSTR)");
			return hr;
		}

		hr = xduMakeElementIntrinsic(dom, parent, name, (const void*)var.bstrVal, SysStringLen(var.bstrVal), true ,newElement, ei);
		if (FAILED(hr) && ei != xduDontSetSimpleErrorInfo)
			ei->add(L"xduMakeElement(%1, <double>)", name);
		return hr;
	} catch (SimpleErrorInfo& ei) {
		ei.add(L"xduMakeElement(%1, <double>)", name);
		throw;
	}
	return E_UNEXPECTED;
}

//======================================================================
//
//======================================================================
HRESULT xduMakeErrorInfoElement(IXMLDOMDocument* dom, IXMLDOMElement* parent, const SimpleErrorInfo& errorInfo, SimpleErrorInfo* ei) throw (SimpleErrorInfo)
{
	try {
		HRESULT hr;

		AutoComPtr<IXMLDOMElement> tmpElm;
		if (parent == NULL) {
			hr = dom->get_documentElement(tmpElm.r());
			if (FAILED(hr)) {
				if (ei == xduThrowSimpleErrorInfo) {
					throw SimpleErrorInfo(hr, L"IXMLDOMDocument::get_documentElement()");
				} else if (ei != xduDontSetSimpleErrorInfo) {
					ei->set(hr, L"IXMLDOMDocument::get_documentElement()");
					ei->add(L"xduMakeErrorInfoElement()");
				}
				return hr;
			}
			parent = tmpElm.p();
		}

		const wchar_t* const tag = L"ErrorInfo";

		AutoComPtr<IXMLDOMElement> errorInfoElm;
		hr = xduMakeElement(dom, parent, tag, errorInfoElm.r(), ei);
		if (FAILED(hr))
			return hr;

		const wchar_t* cp = errorInfo.m_message;
		if (FAILED(errorInfo.m_hr)) {
			AutoComPtr<IXMLDOMElement> sysErr;
			hr = xduMakeElement(dom, errorInfoElm.p(), L"SystemError", cp, sysErr.r(), ei);
			if (FAILED(hr))
				return hr;
			hr = xduMakeAttribute(dom, sysErr.p(), L"HRESULT", static_cast<unsigned int>(errorInfo.m_hr), ei);
			if (FAILED(hr))
				return hr;
			cp += lstrlen(cp) + 1;
		}

		while (cp < errorInfo.m_tail) {
			hr = xduMakeElement(dom, errorInfoElm.p(), L"ApplicationError", cp, NULL, ei);
			if (FAILED(hr))
				return hr;
			cp += lstrlen(cp) + 1;
		}
	} catch (SimpleErrorInfo& ei) {
		ei.add(L"xduMakeErrorInfoElement()");
		throw;
	}

	return S_OK;
}

//======================================================================
//
//======================================================================
HRESULT xduGetNodeListTextIntrinsic(HRESULT hr, IXMLDOMNodeList* nodeList, BSTR& bstr, SimpleErrorInfo* ei) throw (SimpleErrorInfo)
{
	if (FAILED(hr)) {
		if (ei == xduThrowSimpleErrorInfo)
			throw SimpleErrorInfo(hr, L"selectNodes()");
		else if (ei != xduDontSetSimpleErrorInfo)
			ei->set(hr, L"selectNodes()");
		return hr;
	} else if (hr == S_FALSE) {
		if (bstr == NULL)
			bstr = SysAllocStringLen(NULL, 0);
		else
			SysReAllocStringLen(&bstr, NULL, 0);
		return hr;
	}

	BSTR retText = NULL;
	AutoSysFreeString af_retText(retText);

	while (true) {
		AutoComPtr<IXMLDOMNode> node;
		hr = nodeList->nextNode(node.r());
		if (FAILED(hr)) {
			if (ei == xduThrowSimpleErrorInfo)
				throw SimpleErrorInfo(hr, L"IXMLDOMNodeList::nextNode()");
			else if (ei != xduDontSetSimpleErrorInfo)
				ei->set(hr, L"IXMLDOMNodeList::nextNode()");
			return hr;
		} else if (hr == S_FALSE) {
			if (bstr == NULL) {
				if (retText == NULL) {
					bstr = SysAllocString(L"");
				} else {
					bstr = retText;
					retText = NULL;
				}
			} else {
				if (retText == NULL) {
					SysReAllocString(&bstr, L"");
				} else {
					SysReAllocString(&bstr, retText);
				}
			}
			return S_OK;
		}

		BSTR nodeText = NULL;
		hr = node->get_text(&nodeText);
		if (FAILED(hr)) {
			if (ei == xduThrowSimpleErrorInfo)
				throw SimpleErrorInfo(hr, L"IXMLDOMNode::get_text()");
			else if (ei != xduDontSetSimpleErrorInfo)
				ei->set(hr, L"IXMLDOMNode::get_text()");
			return hr;
		} else if (hr == S_FALSE) {
			continue;
		}

		if (retText == NULL) {
			retText = nodeText;
			continue;
		}

		BSTR tmpStr;
		VarBstrCat(retText, nodeText, &tmpStr);
		SysFreeString(retText);
		SysFreeString(nodeText);
		retText = tmpStr;
	}
}

//======================================================================
//
//======================================================================
HRESULT xduGetNodeTextIntrinsic(HRESULT hr, IXMLDOMNode* node, BSTR& bstr, SimpleErrorInfo* ei) throw (SimpleErrorInfo)
{
	if (FAILED(hr)) {
		if (ei == xduThrowSimpleErrorInfo)
			throw SimpleErrorInfo(hr, L"selectSingleNode()");
		else if (ei != xduDontSetSimpleErrorInfo)
			ei->set(hr, L"selectSingleNode()");
		return hr;
	} else if (hr == S_FALSE) {
		if (bstr == NULL)
			bstr = SysAllocStringLen(NULL, 0);
		else
			SysReAllocStringLen(&bstr, NULL, 0);
		return hr;
	}

	BSTR nodeText = NULL;
	hr = node->get_text(&nodeText);
	if (FAILED(hr)) {
		if (ei == xduThrowSimpleErrorInfo)
			throw SimpleErrorInfo(hr, L"IXMLDOMNode::get_text()");
		else if (ei != xduDontSetSimpleErrorInfo)
			ei->set(hr, L"IXMLDOMNode::get_text()");
		return hr;
	} else if (hr == S_FALSE) {
		if (bstr == NULL)
			bstr = SysAllocStringLen(NULL, 0);
		else
			SysReAllocStringLen(&bstr, NULL, 0);
		return hr;
	}

	if (bstr == NULL) {
		bstr = nodeText;
	} else {
		SysReAllocString(&bstr, nodeText);
		SysFreeString(nodeText);
	}

	return S_OK;
}
