/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "UtilClasses.h"
#include "dw-config.h"
#include "cstudydb.h"

#ifdef DEFINE_GUID
#undef DEFINE_GUID
#endif
#define DEFINE_GUID(name, l, w1, w2, b1, b2, b3, b4, b5, b6, b7, b8) \
        const GUID DECLSPEC_SELECTANY name \
                = { l, w1, w2, { b1, b2,  b3,  b4,  b5,  b6,  b7,  b8 } }

#include <ADOID.H>

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
const unsigned long CStudyDB::study_tags[] = {
	0x00080020, 0x00080030, 0x00080060, 0x00100010, 0x00100020, 0x00200010, 0xFFFFFFFF
};

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
BSTR CStudyDB::sm_table_name = NULL;
//========================================================================================
//
//========================================================================================
void CStudyDB::Init(const tDWConfig* config)
{
	sm_table_name = config->odbc.table;
}

//========================================================================================
//
//========================================================================================
CStudyDB::CStudyDB()
	: m_conn(NULL), m_connected(false), m_hr(S_OK), m_message(NULL)
{
	m_hr = CoCreateInstance(CLSID_CADOConnection, NULL, CLSCTX_INPROC_SERVER, IID_IADOConnection, (void**)&m_conn);
	if (FAILED(m_hr)) {
		CWin32ErrorText et(m_hr);
		m_message = SysAllocString(et.str());
	}	
}
//========================================================================================
//
//========================================================================================
CStudyDB::~CStudyDB()
{
	close();

	if (m_conn != NULL)
		m_conn->Release();

	if (m_message != NULL)
		SysFreeString(m_message);

}

//========================================================================================
//
//========================================================================================
bool CStudyDB::open(const BSTR connection_string)
{
	if (m_conn == NULL)
		return false;

	if (!close())
		return false;

	BSTR nullBstr = SysAllocString(L"");
	AutoSysFreeString af_nullBstr(nullBstr);
	
	m_hr = m_conn->Open(connection_string, nullBstr, nullBstr, 0);
	if (FAILED(m_hr)) {
		getErrorInfo(m_conn, m_message);
		return false;
	}

	m_connected = true;

	return true;
}

//========================================================================================
//
//========================================================================================
bool CStudyDB::close()
{
	if (m_conn == NULL)
		return true;

	if (!m_connected)
		return true;

	m_hr = m_conn->Close();
	if (FAILED(m_hr)) {
		getErrorInfo(m_conn, m_message);
		return false;
	}

	m_connected = false;

	return true;
}

//========================================================================================
//
//========================================================================================
BSTR CStudyDB::make_check_sql(const char* volume_label, const char* path)
{
	TOWC(volume_label)
	TOWC(path)

	tMultiStringConcatEntryW str_tbl[] = {
		{ L"SELECT SEQ_NO FROM ", 0 },
		{ sm_table_name, 0 },
		{ L" WHERE PATH='", 0 },
		{ _path_WC, 0 },
		{ L"' AND VOLUME='", 0 },
		{ _volume_label_WC, 0 },
		{ L"'", 0 },
		{ NULL, 0 }
	};

	return str_tbl->NewBSTR();
}
//========================================================================================
//
//========================================================================================
BSTR CStudyDB::make_select_oldest_sql(const wchar_t* vol_label, int auto_delete_flag, const wchar_t* extra_cond)
{
	const wchar_t* const order_key = auto_delete_flag == adLastRecieve ? L"INSERT_DT" : L"STUDY_DT";
	tMultiStringConcatEntryW str_tbl[] = {
		{ L"SELECT SEQ_NO, PATH FROM ", 0 },
		{ sm_table_name, 0 },
		{ L" WHERE ", 0 },
		{ order_key, 0 },
		{ L" = ( SELECT MIN ( ", 0 },
		{ order_key, 0 },
		{ L" ) FROM ", 0 },
		{ sm_table_name, 0 },
		{ L" WHERE VOLUME='", 0 },
		{ vol_label, 0 },
		{ L"'", 0 },
		{ extra_cond, 0 },
		{ L" )", 0 },
		{ NULL, 0 }
	};
	return str_tbl->NewBSTR();
}

//========================================================================================
//
//========================================================================================
bool CStudyDB::StudyExists(bool& exists, const char* const volume_label, const char* const path)
{
	exists = false;

	if (!connected())
		return false;

	BSTR sqlText = make_check_sql(volume_label, path);
	AutoSysFreeString af_sqlText(sqlText);

	return study_exists(exists, sqlText);
}
//========================================================================================
//
//========================================================================================
bool CStudyDB::study_exists(bool& exists, const BSTR sqlText)
{
	exists = false;

	if (!connected())
		return false;

	AutoComPtr<ADORecordset> rs;

	if (!Execute(sqlText, rs.r()))
		return false;

	VARIANT_BOOL vb_eof;
	m_hr = rs->get_EOF(&vb_eof);
	if (FAILED(m_hr)) {
		getErrorInfo(m_conn, m_message);
		return false;
	}

	exists = vb_eof == VARIANT_FALSE;

	rs->Close();

	return true;
}
//========================================================================================
//
//========================================================================================
BSTR CStudyDB::make_insert_sql(
	const char* const volume_label,
	const char* const scp_title,
	const char* const scu_title,
	const char* const path,
	const char* const ref_file,
	const char* const pat_id,
	const char* const pat_name,
	const char* const std_date,
	const char* const std_time,
	const char* const std_id,
	const char* const modality)
{

	char study_dt[16];
	char insert_dt[16];
	char access_dt[16];

	DateTimeToDT(std_date, std_time, study_dt);

	FILETIME insert_ft;
	SYSTEMTIME local_time;
	GetLocalTime(&local_time);
	SystemTimeToFileTime(&local_time, &insert_ft);
	FileTimeToDT(&insert_ft, insert_dt);

	DateTimeToDT("", "", access_dt);

	const wchar_t* table = sm_table_name;
	TOAC(table)

	tMultiStringConcatEntry str_tbl[] = {
		{ "INSERT INTO ", 0 },
		{ _table_AC, 0 },
		{ " ( VOLUME, PATH, SCP_TITLE, SCU_TITLE, REF_FILE, STUDY_DT, INSERT_DT, ACCESS_DT, PatientID, PatientName, StudyDate, StudyTime, StudyID, Modality ) VALUES ", 0 },
		{ " ( '", 0 },
		{ volume_label, 0 },
		{ "', '", 0 },
		{ path, 0 },
		{ "', '", 0 },
		{ scp_title, 0 },
		{ "', '", 0 },
		{ scu_title, 0 },
		{ "', '", 0 },
		{ ref_file, 0 },
		{ "', '", 0 },
		{ study_dt, 0 },
		{ "', '", 0 },
		{ insert_dt, 0 },
		{ "', '", 0 },
		{ access_dt, 0 },
		{ "', '", 0 },
		{ pat_id, 0 },
		{ "', '", 0 },
		{ pat_name, 0 },
		{ "', '", 0 },
		{ std_date, 0 },
		{ "', '", 0 },
		{ std_time, 0 },
		{ "', '", 0 },
		{ std_id, 0 },
		{ "', '", 0 },
		{ modality, 0 },
		{ "' )", 0 },
		{ NULL, 0 }
	};

	return str_tbl->NewBSTR();
}
//========================================================================================
//
//========================================================================================
bool CStudyDB::InsertStudy(
	bool& already_exists,
	const char* const volume_label,
	const char* const scp_title,
	const char* const scu_title,
	const char* const path,
	const char* const ref_file,
	const char* const pat_id,
	const char* const pat_name,
	const char* const std_date,
	const char* const std_time,
	const char* const std_id,
	const char* const modality)
{
	if (!connected())
		return false;

	if (!StudyExists(already_exists, volume_label, path))
		return false;

	if (already_exists)
		return true;

	BSTR sqlText = make_insert_sql(volume_label, scp_title, scu_title, path, ref_file, pat_id, pat_name, std_date, std_time, std_id, modality);
	AutoSysFreeString af_sqlText(sqlText);

	return Execute(sqlText);
}

//========================================================================================
//
//========================================================================================
bool CStudyDB::SelectStudies(const BSTR sqlText, CStudyRecordset& srs)
{
	if (!connected())
		return false;

	AutoComPtr<ADORecordset> rs;

	if (!Execute(sqlText, rs.r()))
		return false;

	srs.SetReocrdset(rs.p());

	return true;
}

//========================================================================================
//
//========================================================================================
bool CStudyDB::ClearStudy(DWORD seq_no)
{
	if (!connected())
		return false;

	wchar_t seq_no_str[32];
	wsprintf(seq_no_str, L"%u", seq_no);

	tMultiStringConcatEntryW str_tbl[] = {
		{ L"DELETE FROM ", 0 },
		{ sm_table_name, 0 },
		{ L" WHERE SEQ_NO = ", 0 },
		{ seq_no_str, 0 },
		{ NULL, 0 }
	};
	BSTR sqlText = str_tbl->NewBSTR();
	AutoSysFreeString af_sqlText(sqlText);

	return Execute(sqlText);
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
bool CStudyDB::Execute(const BSTR sqlText, ADORecordset** rsp)
{
	if (!connected())
		return false;

	VARIANT RecordsAffected;
	AutoVariantClear avc_RecordsAffected(RecordsAffected, true);

	AutoComPtr<ADORecordset> rs;

	m_hr = m_conn->Execute(sqlText, &RecordsAffected, rsp == NULL ? adExecuteNoRecords : 0, rsp == NULL ? NULL : rs.r());
	if (FAILED(m_hr)) {
		getErrorInfo(m_conn, m_message);
		return false;
	}

	if (rs.p() != NULL) {
		if (rsp != NULL) {
			rs->AddRef();
			*rsp = rs.p();
		} else {
			rs->Close();
		}
	}

	return true;
}

//========================================================================================
//
//
//
//========================================================================================
//========================================================================================
//
//========================================================================================
CStudyDBConnector::CStudyDBConnector(CStudyDB& studydb, const BSTR conenction_string)
	: m_studydb(studydb)
{
	m_studydb.open(conenction_string);
}
//========================================================================================
//
//========================================================================================
CStudyDBConnector::~CStudyDBConnector()
{
	m_studydb.close();
}


//========================================================================================
//
//
//
//========================================================================================
//========================================================================================
//
//========================================================================================
CStudyRecordset::CStudyRecordset()
	: m_rs(NULL), m_fields(NULL), m_val_valid(false), m_hr(S_OK)
{
	VariantInit(&m_val);
}
//========================================================================================
//
//========================================================================================
CStudyRecordset::~CStudyRecordset()
{
	if (m_val_valid)
		VariantClear(&m_val);

	if (m_fields != NULL)
		m_fields->Release();

	if (m_rs != NULL) {
		m_rs->Release();
	}
}

//========================================================================================
//
//========================================================================================
void CStudyRecordset::SetReocrdset(ADORecordset* rs)
{
	if (m_val_valid) {
		VariantClear(&m_val);
		m_val_valid = false;
	}

	if (m_fields != NULL) {
		m_fields->Release();
		m_fields = NULL;
	}

	if (m_rs != NULL) {
		m_rs->Release();
		m_rs = NULL;
	}

	if (rs == NULL)
		return;

	rs->AddRef();
	m_rs = rs;

	m_rs->MoveFirst();
}

//========================================================================================
//
//========================================================================================
bool CStudyRecordset::End()
{
	if (m_rs == NULL)
		return true;

	VARIANT_BOOL vb_eof;
	m_hr = m_rs->get_EOF(&vb_eof);
	if (FAILED(m_hr))
		return true;
	
	return vb_eof != VARIANT_FALSE;
}
//========================================================================================
//
//========================================================================================
bool CStudyRecordset::MoveNext()
{
	if (m_fields != NULL) {
		m_fields->Release();
		m_fields = NULL;
	}

	m_hr = m_rs->MoveNext();
	return !FAILED(m_hr);
}

//========================================================================================
//
//========================================================================================
VARIANT* CStudyRecordset::Field(const char* colName)
{
	if (m_fields == NULL) {
		m_hr = m_rs->get_Fields(&m_fields);
		if (FAILED(m_hr))
			return NULL;
	}

	if (m_val_valid) {
		VariantClear(&m_val);
		m_val_valid = false;
	}

	int len = multi_byte_to_wide_char(colName, -1, NULL, 0);
	BSTR bstr = SysAllocStringLen(NULL, len - 1);
	AutoSysFreeString af_bstr(bstr);
	multi_byte_to_wide_char(colName, -1, bstr, len);
	
	VARIANT index;
	VariantInit(&index);
	index.vt = VT_BSTR;
	index.bstrVal = bstr;

	AutoComPtr<ADOField> field;
	m_hr = m_fields->get_Item(index, field.r());
	if (FAILED(m_hr))
		return NULL;

	m_hr = field->get_Value(&m_val);
	if (FAILED(m_hr))
		return NULL;
	m_val_valid = true;

	return &m_val;
}
