/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "UtilClasses.h"
#include "dw-config.h"
#include "dcmdict.h"
#include "CDcmScanner.h"

#include "zip.h"
#include "mbuf.h"

#include "codec.h"
#include "libjpeg-12.h"
#include "libjpeg-8.h"
#include "libpng.h"

//----------------------------------------------------------------------------------------
// Error Logging
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
static void err_log(const char* msg)
{
	HANDLE hFile = CreateFile(L"C:\\DIOWave\\Logs\\dio-compress.log", GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile != INVALID_HANDLE_VALUE) {
		SetFilePointer(hFile, 0, NULL, FILE_END);
		DWORD bytesWritten;
		WriteFile(hFile, msg, strlen(msg), &bytesWritten, NULL);
		WriteFile(hFile, "\x0D\x0A", 2, &bytesWritten, NULL);
		CloseHandle(hFile);
	}
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
expCODEC::expCODEC(const char* message)
{
	const int len = strlen(message);
	if (len > expCODEC_msg_size - 1) {
		memcpy(msg, message, expCODEC_msg_size - 1);
		msg[expCODEC_msg_size - 1] = '\0';
	} else {
		strcpy(msg, message);
	}
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
struct bufctx_t {
	char* m_buf;
	char* m_cp;
	char* m_ep;
};
//----------------------------------------------------------------------------------------
// Main
//----------------------------------------------------------------------------------------
static bool DoCompress(const char* filePath, const char* pngFilePath, const int compresLevel)
{

	CStackFileMapping fileMapping(filePath);
	if (fileMapping.lastError() != NO_ERROR) {
		char msg[1024];
		sprintf(msg, "%s: CStackFileMapping() failed: %d", filePath, fileMapping.lastError());
		err_log(msg);
		return false;
	}
	const char* map = (const char*)fileMapping.mapPointer();
	size_t fileSize = fileMapping.mapSize();

	const char* block_begin = map;
	const char* const block_end = block_begin + fileSize;

	if (memcmp(block_begin + DCM_PREAMBLELENGTH, "DICM", 4) != 0) {
		char msg[1024];
		sprintf(msg, "%s: Preamble not found", filePath);
		err_log(msg);
		return false;
	}
	block_begin += DCM_PREAMBLELENGTH + 4;

	CDcmScannerF scanner(block_begin, block_end);
	tDcmElementHdr hdr;

	const char* cp = block_begin;
	bool explicitVR;
	while (true) {
		int sr = scanner.NextElement(cp, hdr, true);
		if (sr != CDcmScannerF::S_CONTINUE) {
			char msg[1024];
			sprintf(msg, "%s: CDcmScannerF::NextElement() Failed = %d", filePath, sr);
			err_log(msg);
			return false;
		}
		if (hdr.m_tag[0] != DCM_GROUPFILEMETA)
			break;
		if (hdr.m_tag[1] == DCM_TAG_ELEMENT(DCM_METATRANSFERSYNTAX)) {
			char transfer_syntax_uid[DICOM_UI_LENGTH + 1];
			memcpy(transfer_syntax_uid, cp + hdr.m_hsize, hdr.m_length);
			transfer_syntax_uid[hdr.m_length] = '\0';

			if (strcmp(transfer_syntax_uid, DICOM_TRANSFERLITTLEENDIAN) == 0) {
				explicitVR = false;
			} else if (strcmp(transfer_syntax_uid, DICOM_TRANSFERLITTLEENDIANEXPLICIT) == 0) {
				explicitVR = true;
			} else {
				char msg[1024];
				sprintf(msg, "%s: Unsupported Transfer Syntax = %s", filePath, transfer_syntax_uid);
				err_log(msg);
				return false;
			}
		}
		cp = hdr.next_ptr(cp);
	}
	const char* const dataset_begin = cp;

	for (;; cp = hdr.next_ptr(cp)) {
		int sr = scanner.NextElement(cp, hdr, explicitVR);
		if (sr < CDcmScannerF::S_CONTINUE) {
			char msg[1024];
			sprintf(msg, "%s: CDcmScannerF::NextElement() Failed = %d", filePath, sr);
			err_log(msg);
			return false;
		}
		const DCM_TAG ltag = hdr.ltag();
		if (DCM_TAG_GROUP(ltag) == DCM_GROUPPIXEL)
			break;
	}

	DCM_OBJECT* dcmObject;
	CONDITION cond = DCM_ImportStream((unsigned char*)dataset_begin, cp - dataset_begin, explicitVR ? DCM_EXPLICITLITTLEENDIAN : DCM_ORDERLITTLEENDIAN, &dcmObject);
	if (!CTN_SUCCESS(cond)) {
		char msg[1024];
		strcpy(msg, filePath);
		strcat(msg, ": ");
		int len = strlen(msg);
		COND_CopyText(msg + len, sizeof(msg) - len);
		err_log(msg);
		return false;
	}
	Auto_DCM_CloseObject ac_dcmObject(&dcmObject);

	unsigned short bitsAllocated, pixelRepresentation, samplesPerPixel, rows, columns;
	char photometricInterpretation[DICOM_CS_LENGTH + 1];
    DCM_ELEMENT list[] = {
		{DCM_IMGBITSALLOCATED, DCM_US, "", 1, sizeof(bitsAllocated), (char *)&bitsAllocated},
		{DCM_IMGPIXELREPRESENTATION, DCM_US, "", 1, sizeof(pixelRepresentation), (char*)&pixelRepresentation},
		{DCM_IMGSAMPLESPERPIXEL, DCM_US, "", 1, sizeof(samplesPerPixel), (char*)&samplesPerPixel},
		{DCM_IMGROWS, DCM_US, "", 1, sizeof(rows), (char*)&rows},
		{DCM_IMGCOLUMNS, DCM_US, "", 1, sizeof(columns), (char*)&columns},
		{DCM_IMGPHOTOMETRICINTERP, DCM_CS, "", 1, sizeof(photometricInterpretation), photometricInterpretation},
    };
    cond = DCM_ParseObject(&dcmObject, list, (int) DIM_OF(list), NULL, 0, NULL);
    if (!CTN_SUCCESS(cond)) {
		char msg[1024];
		strcpy(msg, filePath);
		strcat(msg, ": ");
		int len = strlen(msg);
		COND_CopyText(msg + len, sizeof(msg) - len);
		err_log(msg);
		return false;
	}

	if (strcmp(photometricInterpretation, "MONOCHROME2") != 0
	 && strcmp(photometricInterpretation, "MONOCHROME1") != 0) {
		char msg[1024];
		sprintf(msg, "%s: Bad Photometric Interpretation = %s", filePath, photometricInterpretation);
		err_log(msg);
		return false;
	}
	if (samplesPerPixel != 1) {
		char msg[1024];
		sprintf(msg, "%s: Bad Samples Per Pixel = %d, should be 1", filePath, samplesPerPixel);
		err_log(msg);
		return false;
	}	
	if (bitsAllocated != 16) {
		char msg[1024];
		sprintf(msg, "%s: Bad Bits Allocated = %d, should be 16", filePath, bitsAllocated);
		err_log(msg);
		return false;
	}

	DCM_FILE_META* fileMeta;
	cond = DCM_DefaultFileMeta(&dcmObject, &fileMeta);
    if (!CTN_SUCCESS(cond)) {
		char msg[1024];
		strcpy(msg, filePath);
		strcat(msg, ": ");
		int len = strlen(msg);
		COND_CopyText(msg + len, sizeof(msg) - len);
		err_log(msg);
		return false;
	}
	strcpy(fileMeta->transferSyntaxUID, "PNG");
	cond = DCM_SetFileMeta(&dcmObject, fileMeta);
    if (!CTN_SUCCESS(cond)) {
		char msg[1024];
		strcpy(msg, filePath);
		strcat(msg, ": ");
		int len = strlen(msg);
		COND_CopyText(msg + len, sizeof(msg) - len);
		err_log(msg);
		DCM_FreeFileMeta(&fileMeta);
		return false;
	}
	DCM_FreeFileMeta(&fileMeta);

	for (;; cp = hdr.next_ptr(cp)) {
		hdr.ReadHdr(cp, explicitVR);
		if (hdr.ltag() == DCM_PXLPIXELDATA)
			break;
	}

	unsigned short* rawPixel = (unsigned short*)hdr.value_ptr(cp);
	unsigned long pixelLength = hdr.value_length();

	if (pixelLength != sizeof(unsigned short) * columns * rows) {
		char msg[1024];
		sprintf(msg, "%s: Columns (%d) * Rows (%d) * 2 != Pixel Size (%d)", filePath, columns, rows, pixelLength);
		err_log(msg);
		return false;
	}

	void* pngCtx = _alloca(cpng_ctx_size());
	unsigned long encodedPixelLength;
	bufctx_t bufctx = { 0, 0, 0};
	try {
		init_cpng(pngCtx);
		setup_cpng(pngCtx, columns, rows, 2, compresLevel);
		unsigned short* pp = (unsigned short*)rawPixel;
		for (int i = 0; i < rows; i++) {
			proc_cpng(pngCtx, (void*)&pp, 1);
			pp += columns;
		}
		size_t size = finish_cpng(pngCtx);
		if (size & 1)
			size++;
		
		unsigned short enc[8] = { 0xFFFE, 0xE000, 0x0000, 0x0000, 0xFFFE, 0xE000, 0xAAAA, 0xAAAA };

		encodedPixelLength = size + sizeof(enc);
		*(unsigned long*)(enc + 6) = size;

		bufctx.m_buf = (char*)codec_mem_alloc(encodedPixelLength);
		if (bufctx.m_buf == NULL) {
			char msg[1024];
			sprintf(msg, "%s: HeapAlloc() Failed = %d", filePath, GetLastError());
			err_log(msg);
			return false;
		}
		bufctx.m_cp = bufctx.m_buf;
		bufctx.m_ep = bufctx.m_buf + encodedPixelLength;

		CopyMemory(bufctx.m_cp, enc, sizeof(enc));
		bufctx.m_cp += sizeof(enc);

		flush_cpng(pngCtx, (void*)&bufctx);
		destroy_cpng(pngCtx);

	} catch (expCODEC ec) {
		char msg[1024];
		sprintf(msg, "%s: CODEC Error\n %s", filePath, ec.msg);
		err_log(msg);
		if (bufctx.m_buf != NULL)
			codec_mem_free(bufctx.m_buf);
		return false;
	}

	DCM_ELEMENT pixelElm = { DCM_PXLPIXELDATA, DCM_OB, "", 1, encodedPixelLength, bufctx.m_buf };
	cond = DCM_AddElement(&dcmObject, &pixelElm);
	codec_mem_free(bufctx.m_buf);
    if (!CTN_SUCCESS(cond)) {
		char msg[1024];
		strcpy(msg, filePath);
		strcat(msg, ": ");
		int len = strlen(msg);
		COND_CopyText(msg + len, sizeof(msg) - len);
		err_log(msg);
		return false;
	}
	
	cond = DCM_WriteFile(&dcmObject, DCM_EXPLICITLITTLEENDIAN | DCM_PART10FILE, pngFilePath);
    if (!CTN_SUCCESS(cond)) {
		char msg[1024];
		strcpy(msg, filePath);
		strcat(msg, ": ");
		int len = strlen(msg);
		COND_CopyText(msg + len, sizeof(msg) - len);
		err_log(msg);
		if (GetFileAttributesA(pngFilePath) != 0xFFFFFFFF)
			DeleteFileA(pngFilePath);
		return false;
	}

	return true;
}

//========================================================================================
// Export to CODECs
//========================================================================================
int codec_write_client(void* ecb, void* buf, size_t len)
{
	bufctx_t* bufctx = (bufctx_t*)ecb;
	CopyMemory(bufctx->m_cp, buf, len);
	bufctx->m_cp += len;
	return len;
}
//========================================================================================
void codec_log_message(const char* message)
{
	err_log(message);
}
//========================================================================================
static HANDLE hCodecHeap = NULL;
//========================================================================================
void* codec_mem_alloc(size_t size)
{
	return HeapAlloc(hCodecHeap, 0, size);
}
//========================================================================================
void codec_mem_free(void* ptr)
{
	HeapFree(hCodecHeap, 0, ptr);
}

//========================================================================================
// Main
//========================================================================================
int main(int argc, const char** argv)
{
	if (argc < 3) {
		printf("usage: dio-compress <comp-level> files ...\n");
	}

	hCodecHeap = HeapCreate(0, 16 * 1024 * 1024, 0);
	if (hCodecHeap == NULL)
		return 1;

	init_mbuf();

	int complevel = atoi(argv[1]);
	if (complevel < 0)
		complevel = 0;
	else if (complevel > 9)
		complevel = 9;

	for (int i = 2; i < argc; i++) {
		char pngFilePath[MAX_PATH * 4];
		strcpy(pngFilePath, argv[i]);
		strcat(pngFilePath, ".png");
		if (DoCompress(argv[i], pngFilePath, complevel)) {
			MoveFileExA(pngFilePath, argv[i], MOVEFILE_REPLACE_EXISTING);
		}
	}

	HeapDestroy(hCodecHeap);

}
