
/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

using System;
using System.Runtime.InteropServices;
using System.Collections;
using System.Text;
using System.IO;
using System.Xml;

namespace DIOWave {

	class DIOCompress {

		//======================================================================
		//
		//======================================================================
		static void Main(String[] args) {

			if (args.Length < 3) {
				Console.WriteLine("usage: <transfer syntax> <option> <src path> [<dst dir>]");
				Console.WriteLine("    transfer syntax:"); 
				Console.WriteLine("        0: Implicit VR Little Endian");
				Console.WriteLine("        1: Explicit VR Little Endian");
				Console.WriteLine("        2: JPEG Baseline (8-bit)");
				Console.WriteLine("        3: JPEG Extended (12-bit)");
				Console.WriteLine("        4: Lossless JPEG");
				Console.WriteLine("        5: Lossless JPEG (First order Predict)");
				Console.WriteLine("        6: PNG");
				Console.WriteLine("    option:");
				Console.WriteLine("  bit 0-7: Codec Option");
				Console.WriteLine("        8: allow re-compression same method");
				Console.WriteLine("        9: allow re-compression different lossless method");
				Console.WriteLine("       10: allow re-compression from lossy to lossless");
				Console.WriteLine("       11: allow re-compression from lossy to lossy");
#if false
				Console.WriteLine("        8: ł̍Ĉk");
				Console.WriteLine("        9: tkǂł̍Ĉk");
				Console.WriteLine("       10: tk̉tkł̍Ĉk");
				Console.WriteLine("       11: tk̔tł̍Ĉk");
#endif
				return;
			}

			int ts = int.Parse(args[0]);
			int opt = int.Parse(args[1]);

			String src_dir = args[2];
			String dst_dir = src_dir;
			if (args.Length > 3)
				dst_dir = args[3];

			String cwd = Directory.GetCurrentDirectory();

			src_dir = Path.Combine(cwd, src_dir);
			dst_dir = Path.Combine(cwd, dst_dir);

			DateTime now = DateTime.Now;
			String log_file_name = "dio-compress-" + now.ToString("yyyy.MM.dd.HH.mm.ss") + ".log";
			FileStream stream = new FileStream(log_file_name, FileMode.CreateNew, FileAccess.ReadWrite, FileShare.Read);
			XmlTextWriter sax = new XmlTextWriter(stream, Encoding.GetEncoding("Shift_JIS"));
			sax.Formatting = Formatting.Indented; 
			sax.WriteStartDocument(true);
			sax.WriteStartElement("log");
			sax.WriteAttributeString("start", now.ToString("yyyy.MM.dd.HH.mm.ss"));
			StringBuilder arg_sb= new StringBuilder();
			foreach (String arg in args) {
				if (arg_sb.Length != 0)
					arg_sb.Append(' ');
				arg_sb.Append(arg);
			}
			sax.WriteAttributeString("arg", arg_sb.ToString());
			sax.WriteAttributeString("cwd", cwd);
			try {
				InitCompress();
				proc(src_dir, dst_dir, ts, opt, sax);
			} finally {
				sax.WriteEndElement();
				sax.WriteEndDocument();
				sax.Close();
				stream.Close();
				UninitCompress();
			}
		}

		//----------------------------------------------------------------------
		//
		//----------------------------------------------------------------------
		static void error_xml(String point, Exception ex, XmlTextWriter sax) {
			sax.WriteStartElement("error");
			sax.WriteAttributeString("point", point);
				sax.WriteStartElement("message");
					sax.WriteString(ex.Message);
				sax.WriteEndElement();
				sax.WriteStartElement("source");
					sax.WriteString(ex.Source);
				sax.WriteEndElement();
			sax.WriteEndElement();
		}
		//----------------------------------------------------------------------
		//
		//----------------------------------------------------------------------
		static void proc(String src_dir, String dst_dir, int ts, int opt, XmlTextWriter sax) {

			if (!Directory.Exists(dst_dir)) {
				try {
					Directory.CreateDirectory(dst_dir);
				} catch (Exception ex) {
					Console.Write("Directory.CreateDirectory(" + dst_dir + ") : ");
					Console.WriteLine(ex.Message);
					error_xml("Directory.CreateDirectory(" + dst_dir + ")", ex, sax);
				}
			}

			foreach (String srcFile in Directory.GetFiles(src_dir)) {
				String fileName = Path.GetFileName(srcFile);
				String dstFile = Path.Combine(dst_dir, fileName);
				if (!Char.IsLetterOrDigit(fileName[0])) {
					try {
						File.Copy(srcFile, dstFile, true);
					} catch (Exception ex) {
						Console.Write("File.Copy(" + srcFile + ", " + dstFile + ") : ");
						Console.WriteLine(ex.Message);
						error_xml("File.Copy(" + srcFile + ", " + dstFile + ")", ex, sax);
					}
					continue;
				}

				FileInfo src_fi = new FileInfo(srcFile);
				Console.WriteLine(srcFile);

				if (src_dir == dst_dir)
					dstFile = Path.Combine(dst_dir, dstFile + ".dio-comperss.tmp");

				StringBuilder error_msg = new StringBuilder(512);
				DateTime start = DateTime.Now;
				UInt32 hr = Compress(srcFile, dstFile, ts, opt, error_msg.Capacity, error_msg);
				DateTime finish = DateTime.Now;
				sax.WriteStartElement("compress");
				sax.WriteAttributeString("hr", hr.ToString("X8"));
				sax.WriteAttributeString("file", srcFile);
				sax.WriteAttributeString("file-size", src_fi.Length.ToString());
				sax.WriteAttributeString("start", start.ToString("yyyy.MM.dd.HH.mm.ss.ff"));
				sax.WriteAttributeString("elapsed", finish.Subtract(start).TotalSeconds.ToString());
				if (hr == 0) {
					if (src_dir == dst_dir) {
						String tmpFile = dstFile;
						dstFile = Path.Combine(dst_dir, fileName);
						try {
							File.Delete(dstFile);
						} catch (Exception ex) {
							Console.Write("File.Delete(" + dstFile + ") : ");
							Console.WriteLine(ex.Message);
							error_xml("File.Delete(" + dstFile + ")", ex, sax);
							sax.WriteEndElement();
							sax.Flush();
							continue;
						}
						try {
							File.Move(tmpFile, dstFile);
						} catch (Exception ex) {
							Console.Write("File.Move(" + tmpFile + ", " + dstFile + ") : ");
							Console.WriteLine(ex.Message);
							error_xml("File.Move(" + tmpFile + ", " + dstFile + ")", ex, sax);
							sax.WriteEndElement();
							sax.Flush();
							continue;
						}
					}
					FileInfo dst_fi = new FileInfo(dstFile);
					sax.WriteStartElement("compressed");
					sax.WriteAttributeString("file-size", dst_fi.Length.ToString());
					sax.WriteAttributeString("file-ratio", ((Double)src_fi.Length / (Double)dst_fi.Length).ToString());
					sax.WriteString(error_msg.ToString());
					sax.WriteEndElement();
				} else {
					String err_str = error_msg.ToString();
					Console.WriteLine(err_str);
					sax.WriteStartElement("error");
					sax.WriteString(err_str);
					sax.WriteEndElement();
					if (src_dir != dst_dir)	{
						try {
							File.Copy(srcFile, dstFile, true);
						} catch (Exception ex) {
							Console.Write("File.Copy(" + srcFile + ", " + dstFile + ") : ");
							Console.WriteLine(ex.Message);
							error_xml("File.Copy(" + srcFile + ", " + dstFile + ")", ex, sax);
							sax.WriteEndElement();
						}
					}
				}
				sax.WriteEndElement();
				sax.Flush();
			}

			foreach (String dir in Directory.GetDirectories(src_dir)) {
				proc(dir, Path.Combine(dst_dir, Path.GetFileName(dir)), ts, opt, sax);
			}
		}

		//----------------------------------------------------------------------
		//
		//----------------------------------------------------------------------
		[DllImport("dio-compress.dll", CharSet=CharSet.Unicode, PreserveSig=true)]
		static extern UInt32 Compress(String filePath, String outFilePath, int requested_transfer_syntax_id, int codec_option, int messageBufCapacity, StringBuilder messageBuf);
		//----------------------------------------------------------------------
		//
		//----------------------------------------------------------------------
		[DllImport("dio-compress.dll", CharSet=CharSet.Unicode, PreserveSig=true)]
		static extern UInt32 InitCompress();
		//----------------------------------------------------------------------
		//
		//----------------------------------------------------------------------
		[DllImport("dio-compress.dll", CharSet=CharSet.Unicode, PreserveSig=true)]
		static extern UInt32 UninitCompress();
}
}
