/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "UtilClasses.h"
#include "dw-config.h"
#include "dcmdict.h"
#include "CDcmScanner.h"

#include "mbuf.h"

#include "codec.h"
#include "libpng.h"
#include "libjpeg-8.h"
#include "libjpeg-12.h"
#include "libjpeg-ll.h"

extern "C" {
	__declspec(dllexport) HRESULT InitCompress();
	__declspec(dllexport) HRESULT UninitCompress();
	__declspec(dllexport) HRESULT Compress(const wchar_t* filePath, const wchar_t* outFilePath, const int requested_transfer_syntax_id, const int codec_option, int messageBufCapacity, wchar_t* messageBuf);
}

//========================================================================================
//
// Export to CODECs
//
//========================================================================================
//========================================================================================
// 
//========================================================================================
int codec_write_client(void* h, void* buf, size_t len)
{
	DWORD bytesWritten;
	if (!WriteFile((HANDLE)h, buf, len, &bytesWritten, NULL)) {
		DWORD err = GetLastError();
	}
	return TRUE;
}
//========================================================================================
// 
//========================================================================================
void codec_log_message(const char* message)
{
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static HANDLE hCodecHeap = NULL;
//========================================================================================
// 
//========================================================================================
void* codec_mem_alloc(codec_mem_t* cm, size_t size)
{
	if (cm != NULL && size < sizeof(cm->m_buf) - cm->m_allocated_size) {
		char* ptr = cm->m_buf + cm->m_allocated_size;
		cm->m_allocated_size += size;
		return (void*)ptr;
	} else {
		return HeapAlloc(hCodecHeap, 0, size);
	}
}
//========================================================================================
// 
//========================================================================================
void codec_mem_free(codec_mem_t* cm, void* ptr)
{
	if (cm != NULL && ptr < cm->m_buf + sizeof(cm->m_buf))
		return;
	else
		HeapFree(hCodecHeap, 0, ptr);
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
struct dummy_codec_t {
	HANDLE hFile;
	unsigned short width;
	unsigned long size;
};
//----------------------------------------------------------------------------------------
static void proc_dummy_codec(void* vctx, void* row_pointer, unsigned int rows) throw (expCODEC)
{
	dummy_codec_t* codec = (dummy_codec_t*)vctx;

	DWORD bytesWritten;
	WriteFile(codec->hFile, *((void**)row_pointer), codec->width * sizeof(short), &bytesWritten, NULL);

	codec->size += bytesWritten;
}

//----------------------------------------------------------------------------------------
//
// CODEC Wrapper Class
//
//----------------------------------------------------------------------------------------
class CStackCodec {
public:
	typedef enum { eNONE, eJPEG8, eJPEG12, eJPEGLL, ePNG } eCodec;
private:
	void* ctx;	
	void (*proc_ptr)(void*, void*, unsigned int);
	eCodec codec;
	HANDLE hFile;

public:
	static size_t ctx_size(const eCodec codec) {
		if (codec == eNONE)
			return sizeof(dummy_codec_t);
		else if (codec == eJPEG8)
			return cjpeg_8_ctx_size();
		else if (codec == eJPEG12)
			return cjpeg_12_ctx_size();
		else if (codec == ePNG)
			return cpng_ctx_size();
		else
			return 0;
	}

	CStackCodec(HANDLE h, const eCodec cd, void* ct, const unsigned short width, const unsigned short height, const bool isColor, const int quality)
		: ctx(ct), codec(cd), hFile(h)
	{
		if (codec == eNONE) {
			proc_ptr = proc_dummy_codec;
			dummy_codec_t* codec = (dummy_codec_t*)ctx;
			codec->hFile = h;
			codec->width = width;
			codec->size = 0;
		} else if (codec == eJPEG8) {
			proc_ptr = proc_cjpeg_8;
			init_cjpeg_8(ctx);
			setup_cjpeg_8(ctx, width, height, 0, quality, 1);
		} else if (codec == eJPEG12) {
			proc_ptr = proc_cjpeg_12;
			init_cjpeg_12(ctx);
			setup_cjpeg_12(ctx, width, height, 0, quality, 1);
		} else if (codec == ePNG) {
			proc_ptr = proc_cpng;
			init_cpng(ctx);
			setup_cpng(ctx, width, height, 2, quality);
		}
	}

	void proc(void* rowBuf) {
		proc_ptr(ctx, (void*)&rowBuf, 1);
	}

	size_t finish(void) {
		if (codec == eNONE)
			return ((dummy_codec_t*)ctx)->size;
		if (codec == eJPEG8)
			return finish_cjpeg_8(ctx);
		else if (codec == eJPEG12)
			return finish_cjpeg_12(ctx);
		else if (codec == ePNG)
			return finish_cpng(ctx);
		else
			return 0;
	}

	void flush() {
		if (codec == eJPEG8)
			flush_cjpeg_8(ctx, hFile);
		else if (codec == eJPEG12)
			flush_cjpeg_12(ctx, hFile);
		else if (codec == ePNG)
			flush_cpng(ctx, hFile);
	}

	~CStackCodec() {
		if (codec == eJPEG8)
			destroy_cjpeg_8(ctx);
		else if (codec == eJPEG12)
			destroy_cjpeg_12(ctx);
		else if (codec == ePNG)
			destroy_cpng(ctx);
	}

};


//----------------------------------------------------------------------------------------
//
// JPEG Decoder
//
//----------------------------------------------------------------------------------------
class CStackDecoder {
protected:
	void* ctx;
	const void* const data;
	const size_t length;
	const char* cur_ptr;
	const char* end_ptr;

	void init_item()
	{
		cur_ptr = (const char*)data;
		tDcmElementHdr hdr;
		hdr.ReadHdr(cur_ptr, false);
		cur_ptr = hdr.next_ptr(cur_ptr);
		if (hdr.value_length() >= 4)
			end_ptr = cur_ptr + ((unsigned long*)hdr.value_ptr((const char*)data))[1];
		else
			end_ptr = (char*)data + (length - 8);
	}


	void next_item(const char** next_input_byte, size_t* bytes_in_buffer)
	{
		if (cur_ptr < end_ptr) {
			tDcmElementHdr hdr;
			hdr.ReadHdr(cur_ptr, false);
			*next_input_byte = hdr.value_ptr(cur_ptr);
			*bytes_in_buffer = hdr.value_length();
			cur_ptr = hdr.next_ptr(cur_ptr);
		} else {
			*next_input_byte = NULL;
			*bytes_in_buffer = 0;
		}
	}

	void skip(const char** next_input_byte, size_t* bytes_in_buffer, long offset)
	{
		if (*bytes_in_buffer > offset) {
			*next_input_byte += offset;
			*bytes_in_buffer -= offset;
		} else {
			offset -= *bytes_in_buffer;
			while (offset > 0 && cur_ptr < end_ptr) {
				tDcmElementHdr hdr;
				hdr.ReadHdr(cur_ptr, false);
				if (offset > hdr.value_length()) {
					offset -= hdr.value_length();
				} else {
					*next_input_byte = hdr.value_ptr(cur_ptr) + offset;
					*bytes_in_buffer = hdr.value_length() - offset;
					offset = 0;
				}
				cur_ptr = hdr.next_ptr(cur_ptr);
			}
		}
	}

	static void seq_input(void* input_ctx, int command, const void** next_input_byte, unsigned int* bytes_in_buffer, int arg)
	{
		CStackDecoder* decorder = (CStackDecoder*)input_ctx;
		switch (command) {
		case 0:
			decorder->init_item();
		case 1:
			decorder->next_item((const char**)next_input_byte, bytes_in_buffer);
			break;
		case 2:
			decorder->skip((const char**)next_input_byte, bytes_in_buffer, arg);
			break;
		}
	}

public:
	CStackDecoder(void* c, void* jpegSeq, size_t jpegSeqLength) : ctx(c), data(jpegSeq), length(jpegSeqLength), cur_ptr(NULL), end_ptr(NULL) {}
};
//----------------------------------------------------------------------------------------
class CStackJPEGDecoder12 : public CStackDecoder {
public:

	static size_t ctx_size(void) {
		return djpeg_12_ctx_size();
	}

	CStackJPEGDecoder12(void* c, void* jpegSeq, size_t jpegSeqLength) : CStackDecoder(c, jpegSeq, jpegSeqLength) {
		init_djpeg_12(ctx, seq_input, (void*)this);
		unsigned short buf_width, buf_height;
		setup_djpeg_12(ctx, 1, &buf_width, &buf_height);
	}

	~CStackJPEGDecoder12() {
		finish_djpeg_12(ctx);
	}

	void proc(unsigned short* rowBuf) {
		proc_djpeg_12(ctx, &rowBuf, 1);
	}

	void flush(unsigned short* rowBuf) {
		while (continue_djpeg_12(ctx))
			proc_djpeg_12(ctx, &rowBuf, 1);
	}
};
//----------------------------------------------------------------------------------------
class CStackJPEGDecoder8 : public CStackDecoder {
public:

	static size_t ctx_size(void) {
		return djpeg_8_ctx_size();
	}

	CStackJPEGDecoder8(void* c, void* jpegSeq, size_t jpegSeqLength) : CStackDecoder(c, jpegSeq, jpegSeqLength) {
		init_djpeg_8(ctx, seq_input, (void*)this);
		unsigned short buf_width, buf_height;
		setup_djpeg_8(ctx, 1, &buf_width, &buf_height);
	}

	~CStackJPEGDecoder8() {
		finish_djpeg_8(ctx);
	}

	void proc(unsigned char* rowBuf) {
		proc_djpeg_8(ctx, &rowBuf, 1);
	}

	void flush(unsigned char* rowBuf) {
		while (continue_djpeg_8(ctx))
			proc_djpeg_8(ctx, &rowBuf, 1);
	}
};

//----------------------------------------------------------------------------------------
class CStackJPEGLLDecoder : public CStackDecoder {
public:

	static size_t ctx_size(void) {
		return djpeg_ll_ctx_size();
	}

	CStackJPEGLLDecoder(void* c, void* jpegSeq, size_t jpegSeqLength) : CStackDecoder(c, jpegSeq, jpegSeqLength) {
		init_djpeg_ll(ctx, seq_input, (void*)this);
		setup_djpeg_ll(ctx);
	}

	~CStackJPEGLLDecoder() {
	}

	void proc(unsigned short* rowBuf) {
		proc_djpeg_ll(ctx, (void*)rowBuf);
	}

	void flush(unsigned short* rowBuf) {
	}
};

//----------------------------------------------------------------------------------------
class CStackPNGDecoder : public CStackDecoder {
public:

	static size_t ctx_size(void) {
		return dpng_ctx_size();
	}

	CStackPNGDecoder(void* c, void* jpegSeq, size_t jpegSeqLength) : CStackDecoder(c, jpegSeq, jpegSeqLength) {
		init_dpng(ctx, seq_input, (void*)this);
		unsigned short buf_width, buf_height;
		setup_dpng(ctx, 1, &buf_width, &buf_height);
	}

	~CStackPNGDecoder() {
		finish_dpng(ctx);
	}

	void proc(unsigned short* rowBuf) {
		if (continue_dpng(ctx))
			proc_dpng(ctx, &rowBuf, 1);
	}

	void flush(unsigned short* rowBuf) {
		while (continue_dpng(ctx))
			proc_dpng(ctx, &rowBuf, 1);
	}
};

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
expCODEC::expCODEC(const char* message)
{
	const int len = strlen(message);
	if (len > expCODEC_msg_size - 1) {
		memcpy(msg, message, expCODEC_msg_size - 1);
		msg[expCODEC_msg_size - 1] = '\0';
	} else {
		strcpy(msg, message);
	}
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static unsigned long encodedPixelLength(const void* data)
{
	const char* cp = (const char*)data;
	unsigned long length = 0;

	tDcmElementHdr hdr;
	while (true) {
		hdr.ReadHdr(cp, false);
		length += hdr.element_length();
		if (DCM_DLMSEQUENCEDELIMITATIONITEM == hdr.ltag())
			break;
		cp = hdr.next_ptr(cp);
	}

	return length;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
enum {
	eDICOMError = MAKE_HRESULT(SEVERITY_ERROR, 0x0200, 1),
	ePixelError = MAKE_HRESULT(SEVERITY_ERROR, 0x0200, 2),
	eCOMPRESSError = MAKE_HRESULT(SEVERITY_ERROR, 0x0200, 3),
	eNOP = MAKE_HRESULT(SEVERITY_ERROR, 0x0200, 4)
};

//----------------------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------------------
class exError {
public:
	HRESULT hr;
	exError(const wchar_t* message, int messageBufCapacity, wchar_t* messageBuf)
	{
		const int len = lstrlen(message);
		if (len > messageBufCapacity - 1) {
			memcpy(messageBuf, message, sizeof(wchar_t) * (messageBufCapacity - 1));
			messageBuf[messageBufCapacity - 1] = L'\0';
		} else {
			lstrcpy(messageBuf, message);
		}
	}
};
//----------------------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------------------
class exDICOMError : public exError {
public:
	exDICOMError(const wchar_t* message, int messageBufCapacity, wchar_t* messageBuf)
		: exError(message, messageBufCapacity, messageBuf)
	{
		hr = eDICOMError;
	}
};
//----------------------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------------------
class exDICOMScanError : public exDICOMError {
public:
	exDICOMScanError(const wchar_t* message, DWORD offset, const char* ptr, int messageBufCapacity, wchar_t* messageBuf)
		: exDICOMError(message, messageBufCapacity, messageBuf)
	{
		hr = eDICOMError;

		wchar_t buf[64];		
		wsprintf(buf, L" : offset=%d dump=%02X %02X %02X %02X", offset, ptr[0], ptr[1], ptr[2], ptr[3]);
		const int len1 = lstrlen(messageBuf);
		const int len2 = lstrlen(buf);

		if (len1 + len2 <= messageBufCapacity - 1)
			lstrcat(messageBuf, buf);
	}
};
//----------------------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------------------
class exCTNError : public exDICOMError {
public:
	exCTNError(const wchar_t* message, int messageBufCapacity, wchar_t* messageBuf)
		: exDICOMError(message, messageBufCapacity, messageBuf)
	{
		hr = eDICOMError;

		char buf[512];
		COND_CopyText(buf, 512);

		int len = lstrlen(messageBuf);
		wchar_t* mp = messageBuf + len;
		wchar_t* ep = messageBuf + messageBufCapacity - 1;
		for (const char* cp = buf; mp < ep && *cp != '\0'; cp++, mp++)
			*mp = *cp;
		*mp = L'\0';
	}
};

//----------------------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------------------
class exWin32Error : public exError {
public:
	exWin32Error(const wchar_t* message, const DWORD lastError, int messageBufCapacity, wchar_t* messageBuf)
		: exError(message, messageBufCapacity, messageBuf)
	{
		hr = HRESULT_FROM_WIN32(lastError);

		const DWORD len = lstrlen(messageBuf);
		messageBufCapacity -= len;
		messageBuf += len;
		FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL, lastError, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), messageBuf, messageBufCapacity, NULL);
	}
};
//----------------------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------------------
class exPixelError : public exError {
public:
	exPixelError(const wchar_t* message, int messageBufCapacity, wchar_t* messageBuf)
		: exError(message, messageBufCapacity, messageBuf)
	{
		hr = ePixelError;
	}
};

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
#define _L(x) L#x

static struct {
	const char* uid;
	const wchar_t* uid_wc;
	const bool explicitVR;
	const bool can_compress;
	const bool lossy_jpeg;
	const int ctn_opt;
	const CStackCodec::eCodec codec_type;
} transfer_syntax_table[] = {
	{ DICOM_TRANSFERLITTLEENDIAN, _L(DICOM_TRANSFERLITTLEENDIAN), false, true, false, DCM_ORDERLITTLEENDIAN, CStackCodec::eNONE },
	{ DICOM_TRANSFERLITTLEENDIANEXPLICIT, _L(DICOM_TRANSFERLITTLEENDIANEXPLICIT), true, true, false, DCM_EXPLICITLITTLEENDIAN, CStackCodec::eNONE },
	{ DICOM_TRANSFERJPEGBASELINEPROCESS1, _L(DICOM_TRANSFERJPEGBASELINEPROCESS1), true, false, true, DCM_EXPLICITLITTLEENDIAN, CStackCodec::eJPEG8 },
	{ DICOM_TRANSFERJPEGEXTENDEDPROC2AND4, _L(DICOM_TRANSFERJPEGEXTENDEDPROC2AND4), true, false, true, DCM_EXPLICITLITTLEENDIAN, CStackCodec::eJPEG12 },
	{ DICOM_TRANSFERJPEGLOSSLESSPROC14, _L(DICOM_TRANSFERJPEGLOSSLESSPROC14), true, true, false, DCM_EXPLICITLITTLEENDIAN, CStackCodec::eJPEGLL },
	{ DICOM_TRANSFERJPEGLOSSLESSPROCFIRSTORDERREDICT, _L(DICOM_TRANSFERJPEGLOSSLESSPROCFIRSTORDERREDICT), true, true, false, DCM_EXPLICITLITTLEENDIAN, CStackCodec::eJPEGLL },
	{ "PNG", _L("PNG"), true, true, false, DCM_EXPLICITLITTLEENDIAN, CStackCodec::ePNG }
};

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static const char* scanMetaHeader(const char* const begin, const char* const end, int& file_transfer_syntax_id,
                                  int messageBufCapacity, wchar_t* messageBuf)
{
	const char* cp = begin;

	if (end - cp < DCM_PREAMBLELENGTH + 4)
		throw exDICOMError(L"file size too short", messageBufCapacity, messageBuf);
	if (memcmp(cp + DCM_PREAMBLELENGTH, "DICM", 4) != 0) {
/*
		if (memcmp(cp, "\x08\x00\05\0x00", 4) == 0
		 || memcmp(cp, "\x08\x00\x08\x00", 4) == 0) {
			file_transfer_syntax_id = 0;
			return begin;
		}
*/
		throw exDICOMError(L"\"DICM\" signature not found", messageBufCapacity, messageBuf);
	}

	cp += DCM_PREAMBLELENGTH + 4;

	CDcmScannerF scanner(cp, end);
	tDcmElementHdr hdr;

	while (true) {
		int sr = scanner.NextElement(cp, hdr, true);
		if (sr != CDcmScannerF::S_CONTINUE)
			throw exDICOMScanError(L"scanning META Header", cp - begin, cp, messageBufCapacity, messageBuf);
			
		if (hdr.gtag() != DCM_GROUPFILEMETA)
			break;

		if (hdr.etag() == DCM_TAG_ELEMENT(DCM_METATRANSFERSYNTAX)) {

			char transfer_syntax_uid[DICOM_UI_LENGTH + 1];
			memcpy(transfer_syntax_uid, hdr.value_ptr(cp), hdr.value_length());
			transfer_syntax_uid[hdr.value_length()] = '\0';

			for (file_transfer_syntax_id = 0; file_transfer_syntax_id < 7; file_transfer_syntax_id++)
				if (strcmp(transfer_syntax_uid, transfer_syntax_table[file_transfer_syntax_id].uid) == 0)
					break;
			if (file_transfer_syntax_id == 7) {
				char msg[128];
				strcpy(msg, "unable to handle this transfer syntax : ");
				strcat(msg, transfer_syntax_uid);
				TOWC(msg)
				throw exDICOMError(_msg_WC, messageBufCapacity, messageBuf);
			}
		}

		cp = hdr.next_ptr(cp);
	}

	return cp;
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static const char* scanDataSet(const char* const begin, const char* const end, bool explictVR, int messageBufCapacity, wchar_t* messageBuf)
{

	CDcmScannerF scanner(begin, end);
	tDcmElementHdr hdr;

	const char* cp = begin;

	while (true) {
		int sr = scanner.NextElement(cp, hdr, explictVR);
		if (sr < CDcmScannerF::S_CONTINUE)
			throw exDICOMScanError(L"scanning Data Element", cp - begin, cp, messageBufCapacity, messageBuf);

		if (hdr.gtag() == 0x7FE0) {
			return cp;
		}

		cp = hdr.next_ptr(cp);
	}

	return NULL;
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void set_lossyflag(DCM_OBJECT** dcmObject, int messageBufCapacity, wchar_t* messageBuf)
{
	char image_type_str[(DICOM_CS_LENGTH + 1) * 4];
	char derive_desc_str[DICOM_ST_LENGTH + 1];
	DCM_ELEMENT list[] = {
		{DCM_MAKETAG(0x0008, 0x0008), DCM_CS, "", 1, sizeof(image_type_str), image_type_str},
		{DCM_MAKETAG(0x0008, 0x2111), DCM_ST, "", 1, sizeof(derive_desc_str), derive_desc_str},
		{DCM_MAKETAG(0x0028, 0x2110), DCM_CS, "", 1, 2, "01"},
	};
    CONDITION cond = DCM_ParseObject(dcmObject, list, 1, NULL, 0, NULL);
    if (!CTN_SUCCESS(cond))
		throw exCTNError(L"DCM_ParseObject()", messageBufCapacity, messageBuf);

	char* cp = image_type_str;
	char* ep = image_type_str + list[0].length;
	while (cp < ep) {
		if (*cp == '\\')
			break;
		cp++;
	}
	const char* rest_str = cp;

	const char* const DERIVED = "DERIVED";
	int new_image_type_str_len = strlen(DERIVED) + strlen(rest_str);
	char* new_image_type_str = (char*)_alloca(new_image_type_str_len + 2);
	strcpy(new_image_type_str, DERIVED);
	strcat(new_image_type_str, rest_str);
	if (new_image_type_str_len & 1) {
		new_image_type_str_len++;
		strcat(new_image_type_str, " ");
	}
	list[0].length = new_image_type_str_len;
	list[0].d.string = new_image_type_str;

    cond = DCM_ParseObject(dcmObject, &list[1], 1, NULL, 0, NULL);
    if (CTN_SUCCESS(cond)) {
    	strcat(derive_desc_str, "\\Lossy JPEG by DIOWave ");
    } else {
    	strcat(derive_desc_str, "Lossy JPEG by DIOWave ");
    }
    int derive_desc_str_len = strlen(derive_desc_str);
	if (derive_desc_str_len & 1) {
		derive_desc_str[--derive_desc_str_len] = '\0';
	}
	list[1].length = derive_desc_str_len;

	cond = DCM_ModifyElements(dcmObject, list, 3, NULL, 0, NULL);
    if (!CTN_SUCCESS(cond))
		throw exCTNError(L"DCM_ModifyElements()", messageBufCapacity, messageBuf);
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static CONDITION export_callback(void *buf, U32 bytesExported, int lastFlag, void *ctx)
{
	HANDLE hFile = *(HANDLE*)ctx;

	DWORD bytesWritten;
	if (WriteFile(hFile, buf, bytesExported, &bytesWritten, NULL))
		return DCM_NORMAL;
	else
		return GetLastError();
}

//========================================================================================
// Main
//========================================================================================
HRESULT Compress(const wchar_t* filePath, const wchar_t* outFilePath, const int requested_transfer_syntax_id, const int option, int messageBufCapacity, wchar_t* messageBuf)
{
	try {

		const int codec_option = option & 0xFF;
		const bool allow_recompress_same_method = (option & 0x100) != 0;
		const bool allow_recompress_diffent_method = (option & 0x200) != 0;
		const bool allow_recompress_lossy_to_lossless = (option & 0x400) != 0;
		const bool allow_recompress_lossy_to_lossy = (option & 0x0800) != 0;
	
		CStackFileMapping fileMapping(filePath);
		if (fileMapping.lastError() != NO_ERROR)
			throw exWin32Error(L"CStackFileMapping() : ", fileMapping.lastError(), messageBufCapacity, messageBuf);
		const char* const map = (const char*)fileMapping.mapPointer();
		const size_t fileSize = fileMapping.mapSize();
		const char* const map_end = map + fileSize;

		int file_transfer_syntax_id = 7;
		const char* dataset_begin = scanMetaHeader(map, map_end, file_transfer_syntax_id, messageBufCapacity, messageBuf);

		if (file_transfer_syntax_id == requested_transfer_syntax_id) {
			if (!allow_recompress_same_method) {
				lstrcpy(messageBuf, transfer_syntax_table[file_transfer_syntax_id].uid_wc);
				lstrcat(messageBuf, L" , ");
				lstrcat(messageBuf, transfer_syntax_table[requested_transfer_syntax_id].uid_wc);
				return eNOP;
			}
		}
		if (file_transfer_syntax_id > 3 && requested_transfer_syntax_id > 3) {
			if (!allow_recompress_diffent_method) {
				lstrcpy(messageBuf, transfer_syntax_table[file_transfer_syntax_id].uid_wc);
				lstrcat(messageBuf, L" , ");
				lstrcat(messageBuf, transfer_syntax_table[requested_transfer_syntax_id].uid_wc);
				return eNOP;
			}
		}
		if (file_transfer_syntax_id == 2 || file_transfer_syntax_id == 3) {
			if (requested_transfer_syntax_id == 2 || requested_transfer_syntax_id == 3) {
				if (!allow_recompress_lossy_to_lossy) {
					lstrcpy(messageBuf, transfer_syntax_table[file_transfer_syntax_id].uid_wc);
					lstrcat(messageBuf, L" , ");
					lstrcat(messageBuf, transfer_syntax_table[requested_transfer_syntax_id].uid_wc);
					return eNOP;
				}
			} else {
				if (!allow_recompress_lossy_to_lossless) {
					lstrcpy(messageBuf, transfer_syntax_table[file_transfer_syntax_id].uid_wc);
					lstrcat(messageBuf, L" , ");
					lstrcat(messageBuf, transfer_syntax_table[requested_transfer_syntax_id].uid_wc);
					return eNOP;
				}
			}
		}

		const char* pixelgroup_begin = scanDataSet(dataset_begin, map_end, transfer_syntax_table[file_transfer_syntax_id].explicitVR, messageBufCapacity, messageBuf);

		const DWORD dataset_length = pixelgroup_begin - dataset_begin;

		//
		// Convert Data Set except Pixel Data
		//
		DCM_OBJECT* dcmObject;
		CONDITION cond = DCM_ImportStreamEx((unsigned char*)dataset_begin, dataset_length, transfer_syntax_table[file_transfer_syntax_id].ctn_opt, &dcmObject);
		if (!CTN_SUCCESS(cond))
			throw exCTNError(L"DCM_ImportStream()", messageBufCapacity, messageBuf);
		Auto_DCM_CloseObject ac_dcmObject(&dcmObject);

		unsigned short bitsAllocated, bitsStored, pixelRepresentation, samplesPerPixel, rows, columns;
		char photometricInterpretation[DICOM_CS_LENGTH + 1];
	    DCM_ELEMENT list[] = {
			{DCM_IMGBITSALLOCATED, DCM_US, "", 1, sizeof(bitsAllocated), (char *)&bitsAllocated},
			{DCM_IMGBITSSTORED, DCM_US, "", 1, sizeof(bitsStored), (char *)&bitsStored},
			{DCM_IMGPIXELREPRESENTATION, DCM_US, "", 1, sizeof(pixelRepresentation), (char*)&pixelRepresentation},
			{DCM_IMGSAMPLESPERPIXEL, DCM_US, "", 1, sizeof(samplesPerPixel), (char*)&samplesPerPixel},
			{DCM_IMGROWS, DCM_US, "", 1, sizeof(rows), (char*)&rows},
			{DCM_IMGCOLUMNS, DCM_US, "", 1, sizeof(columns), (char*)&columns},
			{DCM_IMGPHOTOMETRICINTERP, DCM_CS, "", 1, sizeof(photometricInterpretation), photometricInterpretation},
	    };
	    cond = DCM_ParseObject(&dcmObject, list, (int) DIM_OF(list), NULL, 0, NULL);
	    if (!CTN_SUCCESS(cond))
   			throw exCTNError(L"DCM_ParseObject()", messageBufCapacity, messageBuf);

		if (strcmp(photometricInterpretation, "MONOCHROME2") != 0
		 && strcmp(photometricInterpretation, "MONOCHROME1") != 0)
		 	throw exPixelError(L"Photometric Interpretation != MONOCHROME", messageBufCapacity, messageBuf);

		if (samplesPerPixel != 1)
		 	throw exPixelError(L"Samples Per Pixel != 1", messageBufCapacity, messageBuf);

		if (bitsAllocated != 16)
			throw exPixelError(L"Bits Allocated != 16", messageBufCapacity, messageBuf);

		if (transfer_syntax_table[requested_transfer_syntax_id].lossy_jpeg) {
			if (pixelRepresentation == 1)
				throw exPixelError(L"Pixel Representation == 1", messageBufCapacity, messageBuf);
			if (bitsStored > 12)
				throw exPixelError(L"Bits Stored > 12", messageBufCapacity, messageBuf);
			set_lossyflag(&dcmObject, messageBufCapacity, messageBuf);
		}

		DCM_FILE_META* fileMeta;
		cond = DCM_DefaultFileMeta(&dcmObject, &fileMeta);
	    if (!CTN_SUCCESS(cond))
   			throw exCTNError(L"DCM_DefaultFileMeta()", messageBufCapacity, messageBuf);
		strcpy(fileMeta->transferSyntaxUID, transfer_syntax_table[requested_transfer_syntax_id].uid);
		cond = DCM_SetFileMeta(&dcmObject, fileMeta);
	    if (!CTN_SUCCESS(cond))
   			throw exCTNError(L"DCM_SetFileMeta()", messageBufCapacity, messageBuf);
		DCM_FreeFileMeta(&fileMeta);

		HANDLE hOutFile = CreateFile(outFilePath, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if (hOutFile == INVALID_HANDLE_VALUE)
			throw exWin32Error(L"CreateFile()", GetLastError(), messageBufCapacity, messageBuf);
		AutoCloseAndDeleteFile acf(hOutFile, outFilePath);

		char export_work_buffer[4096];
		cond = DCM_ExportStream(&dcmObject, transfer_syntax_table[requested_transfer_syntax_id].ctn_opt | DCM_PART10FILE, export_work_buffer, sizeof(export_work_buffer), export_callback, &hOutFile);
		if (!CTN_SUCCESS(cond))
			throw exCTNError(L"DCM_ExportStream()", messageBufCapacity, messageBuf);

		const DWORD end_of_dataset_offset = SetFilePointer(hOutFile, 0, NULL, FILE_CURRENT);
		if (end_of_dataset_offset == INVALID_SET_FILE_POINTER)
			throw exWin32Error(L"SetFilePointer()", GetLastError(), messageBufCapacity, messageBuf);

		//----------------------
		// Write Pixel Data Tag
		//----------------------
		void* pixel_data_tag;
		int pixel_data_tag_length;
		if (requested_transfer_syntax_id == 0) {
			const unsigned short tag[] = { 0x7FE0, 0x0010, 0xFFFF, 0xFFFF };
			pixel_data_tag = (void*)tag;
			pixel_data_tag_length = 8;
		} else if (requested_transfer_syntax_id == 1) {
			const unsigned short tag[] = { 0x7FE0, 0x0010, 'W' << 8 | 'O', 0x0000, 0xFFFF, 0xFFFF };
			pixel_data_tag = (void*)tag;
			pixel_data_tag_length = 12;
		} else {
			const unsigned short tag[] = { 0x7FE0, 0x0010, 'B' << 8 | 'O', 0x0000, 0xFFFF, 0xFFFF, 0xFFFE, 0xE000, 0x0000, 0x0000, 0xFFFE, 0xE000, 0xAAAA, 0xAAAA };
			pixel_data_tag = (void*)tag;
			pixel_data_tag_length = 28;
		}
		DWORD bytesWritten;
		if (!WriteFile(hOutFile, pixel_data_tag, pixel_data_tag_length, &bytesWritten, NULL))
			throw exWin32Error(L"WriteFile()", GetLastError(), messageBufCapacity, messageBuf);

		//---------------------
		// Setup Encoder Codec
		//---------------------
		void* codec_ctx = (void*)_alloca(CStackCodec::ctx_size(transfer_syntax_table[requested_transfer_syntax_id].codec_type));
		CStackCodec ccodec(hOutFile, transfer_syntax_table[requested_transfer_syntax_id].codec_type, codec_ctx, columns, rows, false, codec_option);

		//---------------------
		//
		//---------------------
		const char* cp = pixelgroup_begin;
		tDcmElementHdr hdr;
		hdr.ReadHdr(cp, transfer_syntax_table[file_transfer_syntax_id].explicitVR);
		while (hdr.etag() != 0x0010)
			cp = hdr.next_ptr(cp);
		const char* const pixeldata_begin = cp;
		const char* pixel_data_ptr = hdr.value_ptr(cp);
		unsigned long pixel_data_length;
		if (file_transfer_syntax_id == 0 || file_transfer_syntax_id == 1)
			pixel_data_length = hdr.value_length();
		else
			pixel_data_length = encodedPixelLength(pixel_data_ptr);
		
		unsigned long compressedPixelLength;
		if (file_transfer_syntax_id == 0 || file_transfer_syntax_id == 1) {
			const unsigned short* sp = (unsigned short*)pixel_data_ptr;
			for (int r = 0; r < rows; r++) {
				ccodec.proc((void*)sp);
				sp += columns;
			}
		} else if (file_transfer_syntax_id == 2) {

			void* djpeg_ctx = (void*)_alloca(CStackJPEGDecoder8::ctx_size());
			CStackJPEGDecoder8 dcodec(djpeg_ctx, (void*)pixel_data_ptr, pixel_data_length);
			unsigned char* rowBuf = (unsigned char*)_alloca(sizeof(char) * columns);
			for (int r = 0; r < rows; r++) {
				dcodec.proc(rowBuf);
				ccodec.proc(rowBuf);
			}
	
		} else if (file_transfer_syntax_id == 3) {
	
			void* djpeg_ctx = (void*)_alloca(CStackJPEGDecoder12::ctx_size());
			CStackJPEGDecoder12 dcodec(djpeg_ctx, (void*)pixel_data_ptr, pixel_data_length);
			unsigned short* rowBuf = (unsigned short*)_alloca(sizeof(unsigned short) * columns);
			for (int r = 0; r < rows; r++) {
				dcodec.proc(rowBuf);
				ccodec.proc(rowBuf);
			}

		} else if (file_transfer_syntax_id == 4 || file_transfer_syntax_id == 5) {

			void* djpeg_ctx = (void*)_alloca(CStackJPEGLLDecoder::ctx_size());
			CStackJPEGLLDecoder dcodec(djpeg_ctx, (void*)pixel_data_ptr, pixel_data_length);
			unsigned short* rowBuf = (unsigned short*)_alloca(sizeof(unsigned short) * columns);
			for (int r = 0; r < rows; r++) {
				dcodec.proc(rowBuf);
				ccodec.proc(rowBuf);
			}

		} else if (file_transfer_syntax_id == 6) {

			void* dpng_ctx = (void*)_alloca(CStackPNGDecoder::ctx_size());
			CStackPNGDecoder dcodec(dpng_ctx, (void*)pixel_data_ptr, pixel_data_length);
			unsigned short* rowBuf = (unsigned short*)_alloca(sizeof(unsigned short) * columns);
			for (int r = 0; r < rows; r++) {
				dcodec.proc(rowBuf);
				ccodec.proc(rowBuf);
			}

		}

		compressedPixelLength = ccodec.finish();

		ccodec.flush();

		if (requested_transfer_syntax_id > 1) {
			unsigned short delim[] = { 0xFFFE, 0xE0DD, 0x0000, 0x0000 };
			WriteFile(hOutFile, delim, 8, &bytesWritten, NULL);
		}

		//----------------------
		// Fixup Pixel Data Data Tag
		//----------------------
		if (requested_transfer_syntax_id == 0) {
			SetFilePointer(hOutFile, end_of_dataset_offset + 4, NULL, FILE_BEGIN);
			WriteFile(hOutFile, &compressedPixelLength, 4, &bytesWritten, NULL);
		} else if (requested_transfer_syntax_id == 1) {
			SetFilePointer(hOutFile, end_of_dataset_offset + 8, NULL, FILE_BEGIN);
			WriteFile(hOutFile, &compressedPixelLength, 4, &bytesWritten, NULL);
		} else {
			SetFilePointer(hOutFile, end_of_dataset_offset + 24, NULL, FILE_BEGIN);
			WriteFile(hOutFile, &compressedPixelLength, 4, &bytesWritten, NULL);
		}

		acf.dontDelete();

		const int raster_size = rows * columns * 2;
		const double ratio = (double)raster_size / (double)compressedPixelLength;
		const int ratio_i = (int)ratio;
		const int ratio_f = (int)(ratio * 100) - ratio_i * 100;
		wnsprintf(messageBuf, messageBufCapacity - 1, L"rows=%d; columns=%d; raster=%d; raw=%d; compressed=%d; ratio=%d.%d;", rows, columns, raster_size, pixel_data_length, compressedPixelLength, ratio_i, ratio_f);

		return S_OK;

	} catch (expCODEC& ex) {
		
		lstrcpy(messageBuf, L"CODEC Error :");
		int len = lstrlen(messageBuf);
		wchar_t* mp = messageBuf + len;
		wchar_t* ep = messageBuf + messageBufCapacity - 1;
		for (const char* cp = ex.msg; mp < ep && *cp != '\0'; cp++, mp++)
			*mp = *cp;
		*mp = L'\0';

		return eCOMPRESSError;

	} catch (exError& ex) {
		return ex.hr;
	}

	return E_FAIL;
}

//========================================================================================
//
//========================================================================================
HRESULT InitCompress()
{
	hCodecHeap = HeapCreate(0, 16 * 1024 * 1024, 0);
	mbuf_init(hCodecHeap);
	return S_OK;
}
//========================================================================================
//
//========================================================================================
HRESULT UninitCompress()
{
	mbuf_uninit(hCodecHeap);
	HeapDestroy(hCodecHeap);
	return S_OK;
}


