/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "UtilClasses.h"
#include "dw-config.h"
#include "dcmdict.h"
#include "CDcmScanner.h"

#include "zip.h"
#include "mbuf.h"

#include "codec.h"
#include "libjpeg-12.h"
#include "libjpeg-8.h"
#include "libjpeg-ll.h"
#include "libpng.h"

#include "voi_proc.h"
#include "psf.h"

#if 0

#include "etag-cache.h"

#else

#define ETAG_VALUE_LIMIT 256
static void InitETag(HANDLE hHeap) {}
static void FinishETag(HANDLE hHeap) {}
static bool CheckETagEntry(const char* etag) { return true; }

#endif

#define DEBUG_TEXT_OUT 0

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static DWORD SuccessCode = HSE_STATUS_SUCCESS;
static tDWConfig* gDWConfig = NULL;

//----------------------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------------------
class expInternalError {
public:
	char msg[256];
	bool write_err_log;

	expInternalError(const char* message, bool log)
		: write_err_log(log) {
		const int len = strlen(message);
		if (len > 256 - 1) {
			memcpy(msg, message, 256 - 1);
			msg[256 - 1] = '\0';
		} else {
			strcpy(msg, message);
		}
	}
};

//----------------------------------------------------------------------------------------
//
// Logging
//
//----------------------------------------------------------------------------------------
#define _DIO_IMAGE_LOG_ 0
#if _DIO_IMAGE_LOG_
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
#define KLogBufferSize 1024 * 64
static CRITICAL_SECTION logging_crit;
static HANDLE hLogFile = INVALID_HANDLE_VALUE;
static char logBuf[KLogBufferSize];
static DWORD logLen;
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void init_logging(void)
{
	InitializeCriticalSectionAndSpinCount(&logging_crit, 0x80000000 + 4096);

	if (!gDWConfig->dio_image_log)
		return;

	logLen = 0;
	*logBuf = '\0';

	int dir_len = lstrlen(kDefaultLogDirectory);
	int name_len = lstrlen(CFP_DIO_IMAGE_LOG);
	wchar_t* logFile = (wchar_t*)_alloca(sizeof(wchar_t) * (dir_len + 1 + name_len + 1));
	strncat3(logFile, kDefaultLogDirectory, dir_len, L'\\', CFP_DIO_IMAGE_LOG, name_len);
	HANDLE hFile = CreateFile(logFile, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
		return;

	SetFilePointer(hFile, 0, 0, FILE_END);

	hLogFile = hFile;
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void flush_log_buffer(void)
{
	DWORD bytesWritten;
	if (!WriteFile(hLogFile, logBuf, logLen, &bytesWritten, NULL)) {
		CloseHandle(hLogFile);
		hLogFile = INVALID_HANDLE_VALUE;
		return;
	}

	logLen = 0;
	*logBuf = '\0';
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void logging(char* logKey, EXTENSION_CONTROL_BLOCK* pECB)
{
	if (!gDWConfig->dio_image_log)
		return;

	if (hLogFile == INVALID_HANDLE_VALUE)
		return;

	DWORD len;

	SYSTEMTIME tm;
	GetLocalTime(&tm);

	char remote_addr[32];
	*remote_addr = '\0';
	len = sizeof(remote_addr);
	pECB->GetServerVariable(pECB->ConnID, "REMOTE_ADDR", remote_addr, &len);

	char buf[MAX_PATH * 2];
	sprintf(buf, gDWConfig->dio_image_log_format,
			tm.wYear, tm.wMonth, tm.wDay, tm.wHour, tm.wMinute, tm.wSecond, tm.wMilliseconds,
			logKey,
			remote_addr,
			pECB->lpszPathInfo,
			pECB->lpszQueryString);
	strcat(buf, "\r\n");
	len = strlen(buf);	

	CStackCriticalSection cs(&logging_crit);

	size_t remain = KLogBufferSize - logLen;
	if (len >= remain)
		flush_log_buffer();
	strcpy(logBuf + logLen, buf);
	logLen +=len;
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void finish_logging(void)
{
	if (!gDWConfig->dio_image_log)
		return;

	if (hLogFile == INVALID_HANDLE_VALUE)
		return;

	flush_log_buffer();
	CloseHandle(hLogFile);
	DeleteCriticalSection(&logging_crit);

}
#endif
//========================================================================================
// Floating Point
//========================================================================================
//----------------------------------------------------------------------------------------
// FPU Control Word
//----------------------------------------------------------------------------------------
unsigned short ceil_cw = 0x0B7F;
unsigned short floor_cw = 0x077F;
//----------------------------------------------------------------------------------------
// Ceil
//----------------------------------------------------------------------------------------
static asm int xceil(float x)
{
	asm {
		fld dword ptr 4[esp]
		sub esp, 4
		fnstcw word ptr [esp]
		fldcw ceil_cw
		sub esp, 4
		fistp dword ptr [esp]
		pop eax
		fldcw word ptr [esp]
		add esp, 4
		ret
	}
/*
000002BC: DD 44 24 04             fld       qword ptr [esp]+04
000002C0: 9B                      wait      
000002C1: D9 3D 00 00 00 00       fnstcw     TEMP2
000002C7: D9 2D 00 00 00 00       fldcw      TO_INFINITY
000002CD: D9 FC                   frndint   
000002CF: D9 2D 00 00 00 00       fldcw      TEMP2
000002D5: C3                      ret       near
*/
}
//----------------------------------------------------------------------------------------
// Floor
//----------------------------------------------------------------------------------------
static asm int xfloor(float x)
{
	asm {
		fld dword ptr 4[esp]
		sub esp, 4
		fnstcw word ptr [esp]
		fldcw floor_cw
		sub esp, 4
		fistp dword ptr [esp]
		pop eax
		fldcw word ptr [esp]
		add esp, 4
		ret
	}
/*
000002D8: DD 44 24 04             fld       qword ptr [esp]+04
000002DC: 9B                      wait      
000002DD: D9 3D 00 00 00 00       fnstcw     TEMP2
000002E3: D9 2D 00 00 00 00       fldcw      TO_M_INFINITY
000002E9: D9 FC                   frndint   
000002EB: D9 2D 00 00 00 00       fldcw      TEMP2
000002F1: C3                      ret       near
*/
}

//----------------------------------------------------------------------------------------
// Write To Client
//----------------------------------------------------------------------------------------
inline int write_client(void* ecb, void* buf, size_t len)
{
#if DEBUG_TEXT_OUT
	return 0;
#endif

	EXTENSION_CONTROL_BLOCK* pECB = (EXTENSION_CONTROL_BLOCK*)ecb;

	DWORD dwLen = len;
	return pECB->WriteClient(pECB->ConnID, buf, &dwLen, HSE_IO_SYNC);
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void hexstr(unsigned int val, char* str, unsigned int len)
{
	char* cp = str + len - 1;
	do {
		unsigned int ch = val & 0xF;
		if (ch < 10)
			ch += '0';
		else
			ch += 'A' - 10;
		*cp-- = ch;
		val = val >> 4;
	} while (cp >= str);

	*(str + len) = '\0';
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void dump_attr(void* ecb, const tDcmElementHdr& hdr, const char* cp)
{
	char str[512];
	sprintf(str, "(%04X,%04X) %c%c [%4d] ", hdr.gtag(), hdr.etag(), hdr.m_vr[0], hdr.m_vr[1], hdr.value_length());
	int len = strlen(str);
	char* dp = str + len;
	if (hdr.isString()) {
		*dp++ = '\"'; len++;
		int buf_remain = sizeof(str) - 2 - len;
		int write_len = hdr.value_length() > buf_remain ? buf_remain : hdr.value_length();
		CopyMemory(dp, hdr.value_ptr(cp), write_len); dp += write_len; len += write_len;
		*dp++ = '\"'; len++;
	} else {
		int size;
		if (hdr.m_vr[0] == 'A' && hdr.m_vr[1] == 'T')
			size = 4;
		else if (hdr.m_vr[1] == 'S' && (hdr.m_vr[0] == 'S' || hdr.m_vr[0] == 'U'))
			size = 2;
		else if (hdr.m_vr[1] == 'L' && (hdr.m_vr[0] == 'S' || hdr.m_vr[0] == 'U'))
			size = 4;
		else if (hdr.m_vr[0] == 'F' && hdr.m_vr[1] == 'L')
			size = 4;
		else if (hdr.m_vr[0] == 'F' && hdr.m_vr[1] == 'D')
			size = 8;
		else
			size = 0;
		if (size != 0) {
			const int count = hdr.value_length() / size;
			const unsigned char* sp = (const unsigned char*)hdr.value_ptr(cp);
			for (int i = 0; i < count; i++, sp += size) {
				if (i != 0) {
					*dp++ = ','; len++;
					*dp++ = ' '; len++;
					
				}
				*dp++ = '0'; len++;
				*dp++ = 'x'; len++;
				for (int j = size - 1; j >= 0; j--) {
					unsigned int ch = sp[j] >> 4;
					if (ch < 10)
						ch += '0';
					else
						ch += 'A' - 10;
					*dp++ = ch;	len++;
					ch = sp[j] & 0x0F;
					if (ch < 10)
						ch += '0';
					else
						ch += 'A' - 10;
					*dp++ = ch;	len++;
				}
			}
		} else {
			const char* not_imp = "not implemented";
			CopyMemory(dp, not_imp, 15); dp += 15; len += 15;
		}
	}
	*dp++ = '\n'; len++;
	write_client(ecb, str, len);
}

//----------------------------------------------------------------------------------------
//
// CODEC
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
// BMP codec
//----------------------------------------------------------------------------------------
struct tBMPCodecContext {
	int width;
	int height;
	bool isColor;

	EXTENSION_CONTROL_BLOCK *ecb;
	int write_count;
	int row_width;
	size_t stream_length;
};

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void bmp_write_header(tBMPCodecContext* ctx)
{
	unsigned char header[sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + sizeof(RGBQUAD) * 256];

	BITMAPFILEHEADER* fh = (BITMAPFILEHEADER*)header;
	fh->bfType = 'MB';
	fh->bfSize = ctx->stream_length;
	fh->bfReserved1 = fh->bfReserved2 = 0;
	fh->bfOffBits = sizeof(BITMAPFILEHEADER);

	BITMAPINFOHEADER* ih = (BITMAPINFOHEADER*)(header + sizeof(BITMAPFILEHEADER));
	ih->biSize = sizeof(BITMAPINFOHEADER);
	ih->biWidth = ctx->width;
	ih->biHeight = ctx->height;
	ih->biPlanes = 1;
	ih->biBitCount = ctx->isColor ? 24 : 8;
	ih->biCompression = BI_RGB;
	ih->biSizeImage = 0;
	ih->biXPelsPerMeter = 1000;
	ih->biYPelsPerMeter = 1000;
	ih->biClrUsed = 0;
	ih->biClrImportant = 0;

	if (ctx->isColor) {
		write_client(ctx->ecb, header, sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER));	
	} else {
		RGBQUAD* palp = (RGBQUAD*)(header + sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER));
		for (int i = 0; i < 256; i++, palp++) {
			palp->rgbBlue = palp->rgbGreen = palp->rgbRed = i;
			palp->rgbReserved = 0;
		}
		write_client(ctx->ecb, header, sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + sizeof(RGBQUAD) * 256);	
	}
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void bmp_write_rows(void* vctx, void* buf, unsigned int row_count)
{
	tBMPCodecContext* ctx = (tBMPCodecContext*)vctx;
	if (ctx->write_count == 0) {
		bmp_write_header(ctx);
	} else {
		unsigned char* row_ptr = *(unsigned char**)buf;
		write_client((void*)ctx->ecb, row_ptr, ctx->row_width);
	}
	ctx->write_count++;
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static size_t bmp_stream_length(void* vctx)
{
	tBMPCodecContext* ctx = (tBMPCodecContext*)vctx;
	return ctx->stream_length;
}

//----------------------------------------------------------------------------------------
//
// CODEC Class
//
//----------------------------------------------------------------------------------------
class CStackCodec {
public:
	typedef enum { JPEG_CODEC, PNG_CODEC, BMP_CODEC } eCodec;
private:
	void* ctx;	
	void (*proc_ptr)(void*, void*, unsigned int);
	eCodec codec;
	EXTENSION_CONTROL_BLOCK *ecb;
public:
	static size_t ctx_size(const eCodec codec) {
		if (codec == JPEG_CODEC)
			return cjpeg_8_ctx_size();
		else if (codec == PNG_CODEC)
			return cpng_ctx_size();
		else if (codec == BMP_CODEC)
			return sizeof(tBMPCodecContext);
		else
			return 0;
	}

	CStackCodec(EXTENSION_CONTROL_BLOCK *pECB, const eCodec cd, void* ct, const unsigned short width, const unsigned short height, const bool isColor, const int quality, const int precision) : ctx(ct), codec(cd), ecb(pECB) {
		if (codec == JPEG_CODEC) {
			proc_ptr = proc_cjpeg_8;
			init_cjpeg_8(ctx);
			setup_cjpeg_8(ctx, width, height, isColor, quality, precision);
		} else if (codec == PNG_CODEC) {
			proc_ptr = proc_cpng;
			init_cpng(ctx);
			setup_cpng(ctx, width, height, isColor, quality);
		} else if (codec == BMP_CODEC) {
			tBMPCodecContext* bctx = (tBMPCodecContext*)ctx;
			bctx->width = width;
			bctx->height = height;
			bctx->isColor = isColor;
			bctx->write_count = 0;
			bctx->ecb = ecb;
			bctx->row_width = width * (isColor ? 3 : 1);
			bctx->row_width += (bctx->row_width % 4) != 0 ? 4 - (bctx->row_width % 4) : 0;
			bctx->stream_length = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + (isColor ? 0 : sizeof(RGBQUAD) * 256 ) + bctx->row_width * height;
			proc_ptr = bmp_write_rows;
		}
	}

	void proc(unsigned char* rowBuf) {
		proc_ptr(ctx, (void*)&rowBuf, 1);
	}

	size_t finish(void) {
		if (codec == JPEG_CODEC)
			return finish_cjpeg_8(ctx);
		else if (codec == PNG_CODEC)
			return finish_cpng(ctx);
		else
			return 0;
	}

	void flush() {
		if (codec == JPEG_CODEC)
			flush_cjpeg_8(ctx, ecb);
		else if (codec == PNG_CODEC)
			flush_cpng(ctx, ecb);
	}

	~CStackCodec() {
		if (codec == JPEG_CODEC) {
			destroy_cjpeg_8(ctx);
		} else if (codec == PNG_CODEC) {
			destroy_cpng(ctx);
		}
	}

	bool is_buffered(void) { return codec != BMP_CODEC; }
	size_t stream_length(void) { return codec == BMP_CODEC ? bmp_stream_length(ctx) : -1; }

	char* content_type(void) { return content_type(codec); }

	static char* content_type(const eCodec codec) {
		if (codec == JPEG_CODEC)
			return "image/jpeg";
		else if (codec == PNG_CODEC)
			return "image/png";
		else if (codec == BMP_CODEC)
			return "image/x-MS-bmp";
		else
			return "binary/octet-stream";
	}
};

//----------------------------------------------------------------------------------------
//
// JPEG Decoder
//
//----------------------------------------------------------------------------------------
class CStackDecoder {
protected:
	void* ctx;
	const void* const data;
	const size_t length;
	const char* cur_ptr;
	const char* end_ptr;

	void init_item()
	{
		cur_ptr = (const char*)data;
		tDcmElementHdr hdr;
		hdr.ReadHdr(cur_ptr, false);
		cur_ptr = hdr.next_ptr(cur_ptr);
		if (hdr.value_length() >= 4)
			end_ptr = cur_ptr + ((unsigned long*)hdr.value_ptr((const char*)data))[1];
		else
			end_ptr = (char*)data + (length - 8);
	}


	void next_item(const char** next_input_byte, size_t* bytes_in_buffer)
	{
		if (cur_ptr < end_ptr) {
			tDcmElementHdr hdr;
			hdr.ReadHdr(cur_ptr, false);
			*next_input_byte = hdr.value_ptr(cur_ptr);
			*bytes_in_buffer = hdr.value_length();
			cur_ptr = hdr.next_ptr(cur_ptr);
		} else {
			*next_input_byte = NULL;
			*bytes_in_buffer = 0;
		}
	}

	void skip(const char** next_input_byte, size_t* bytes_in_buffer, long offset)
	{
		if (*bytes_in_buffer > offset) {
			*next_input_byte += offset;
			*bytes_in_buffer -= offset;
		} else {
			offset -= *bytes_in_buffer;
			while (offset > 0 && cur_ptr < end_ptr) {
				tDcmElementHdr hdr;
				hdr.ReadHdr(cur_ptr, false);
				if (offset > hdr.value_length()) {
					offset -= hdr.value_length();
				} else {
					*next_input_byte = hdr.value_ptr(cur_ptr) + offset;
					*bytes_in_buffer = hdr.value_length() - offset;
					offset = 0;
				}
				cur_ptr = hdr.next_ptr(cur_ptr);
			}
		}
	}

	static void seq_input(void* input_ctx, int command, const void** next_input_byte, unsigned int* bytes_in_buffer, int arg)
	{
		CStackDecoder* decorder = (CStackDecoder*)input_ctx;
		switch (command) {
		case 0:
			decorder->init_item();
		case 1:
			decorder->next_item((const char**)next_input_byte, bytes_in_buffer);
			break;
		case 2:
			decorder->skip((const char**)next_input_byte, bytes_in_buffer, arg);
			break;
		}
	}

public:
	CStackDecoder(void* c, void* jpegSeq, size_t jpegSeqLength) : ctx(c), data(jpegSeq), length(jpegSeqLength), cur_ptr(NULL), end_ptr(NULL) {}
};
//----------------------------------------------------------------------------------------
class CStackJPEGDecoder12 : public CStackDecoder {
public:

	static size_t ctx_size(void) {
		return djpeg_12_ctx_size();
	}

	CStackJPEGDecoder12(void* c, void* jpegSeq, size_t jpegSeqLength) : CStackDecoder(c, jpegSeq, jpegSeqLength) {
		init_djpeg_12(ctx, seq_input, (void*)this);
		unsigned short buf_width, buf_height;
		setup_djpeg_12(ctx, 1, &buf_width, &buf_height);
	}

	~CStackJPEGDecoder12() {
		finish_djpeg_12(ctx);
	}

	void proc(unsigned short* rowBuf) {
		proc_djpeg_12(ctx, &rowBuf, 1);
	}

	void flush(unsigned short* rowBuf) {
		while (continue_djpeg_12(ctx))
			proc_djpeg_12(ctx, &rowBuf, 1);
	}
};
//----------------------------------------------------------------------------------------
class CStackJPEGDecoder8 : public CStackDecoder {
public:

	static size_t ctx_size(void) {
		return djpeg_8_ctx_size();
	}

	CStackJPEGDecoder8(void* c, void* jpegSeq, size_t jpegSeqLength) : CStackDecoder(c, jpegSeq, jpegSeqLength) {
		init_djpeg_8(ctx, seq_input, (void*)this);
		unsigned short buf_width, buf_height;
		setup_djpeg_8(ctx, 1, &buf_width, &buf_height);
	}

	~CStackJPEGDecoder8() {
		finish_djpeg_8(ctx);
	}

	void proc(unsigned char* rowBuf) {
		proc_djpeg_8(ctx, &rowBuf, 1);
	}

	void flush(unsigned char* rowBuf) {
		while (continue_djpeg_8(ctx))
			proc_djpeg_8(ctx, &rowBuf, 1);
	}
};

//----------------------------------------------------------------------------------------
class CStackJPEGLLDecoder : public CStackDecoder {
public:

	static size_t ctx_size(void) {
		return djpeg_ll_ctx_size();
	}

	CStackJPEGLLDecoder(void* c, void* jpegSeq, size_t jpegSeqLength) : CStackDecoder(c, jpegSeq, jpegSeqLength) {
		init_djpeg_ll(ctx, seq_input, (void*)this);
		setup_djpeg_ll(ctx);
	}

	~CStackJPEGLLDecoder() {
	}

	void proc(unsigned short* rowBuf) {
		proc_djpeg_ll(ctx, (void*)rowBuf);
	}

	void flush(unsigned short* rowBuf) {
	}
};

//----------------------------------------------------------------------------------------
class CStackPNGDecoder : public CStackDecoder {
public:

	static size_t ctx_size(void) {
		return dpng_ctx_size();
	}

	CStackPNGDecoder(void* c, void* jpegSeq, size_t jpegSeqLength) : CStackDecoder(c, jpegSeq, jpegSeqLength) {
		init_dpng(ctx, seq_input, (void*)this);
		unsigned short buf_width, buf_height;
		setup_dpng(ctx, 1, &buf_width, &buf_height);
	}

	~CStackPNGDecoder() {
		finish_dpng(ctx);
	}

	void proc(unsigned short* rowBuf) {
		if (continue_dpng(ctx))
			proc_dpng(ctx, &rowBuf, 1);
	}

	void flush(unsigned short* rowBuf) {
		while (continue_dpng(ctx))
			proc_dpng(ctx, &rowBuf, 1);
	}
};

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
expCODEC::expCODEC(const char* message)
{
	const int len = strlen(message);
	if (len > expCODEC_msg_size - 1) {
		memcpy(msg, message, expCODEC_msg_size - 1);
		msg[expCODEC_msg_size - 1] = '\0';
	} else {
		strcpy(msg, message);
	}
}

//----------------------------------------------------------------------------------------
//
// Error Logging
//
//----------------------------------------------------------------------------------------
static CRITICAL_SECTION err_log_crit;
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void init_err_log()
{
	InitializeCriticalSectionAndSpinCount(&err_log_crit, 0x80000000 + 4096);
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void finish_err_log()
{
	DeleteCriticalSection(&err_log_crit);
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void err_log(const char* msg, const char* msg1 = NULL, const char* msg2 = NULL, const char* msg3 = NULL)
{
	CStackCriticalSection cs(&err_log_crit);

	static wchar_t logfile[MAX_PATH] = { '\x0' };
	if (logfile[0] == '\x0') {
		lstrcpy(logfile, gDWConfig->logDir);
		lstrcat(logfile, L"\\dio-image-error.log");
	}

	HANDLE hFile = CreateFile(logfile, GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
		return;
	AutoCloseFileHandle ac(hFile);
	
	SetFilePointer(hFile, 0, NULL, FILE_END);
	DWORD bytesWritten;

	SYSTEMTIME tm;
	GetLocalTime(&tm);
	char buf[80];
	sprintf(buf, "%04d.%02d.%02d.%02d.%02d.%02d.%03d: ",
			tm.wYear, tm.wMonth, tm.wDay, tm.wHour, tm.wMinute, tm.wSecond, tm.wMilliseconds);
	WriteFile(hFile, buf, strlen(buf), &bytesWritten, NULL);

	WriteFile(hFile, msg, strlen(msg), &bytesWritten, NULL);
	if (msg1 != NULL)
		WriteFile(hFile, msg1, strlen(msg1), &bytesWritten, NULL);
	if (msg2 != NULL)
		WriteFile(hFile, msg2, strlen(msg2), &bytesWritten, NULL);
	if (msg3 != NULL)
		WriteFile(hFile, msg3, strlen(msg3), &bytesWritten, NULL);

	WriteFile(hFile, "\x0D\x0A", 2, &bytesWritten, NULL);
}

//----------------------------------------------------------------------------------------
//
// QUERY_STRING proc
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
enum {
	prmWindowCenter,
	prmWindowWidth,
	prmScale,
	prmImageWidth,
	prmImageHeight,
	prmZoomX,
	prmZoomY,
	prmCodec,
	prmCompLevel,
	prmPrecision,
	prmVOIPrecision,
	prmDCTPrecision,
	prmPSFPrecision,
	prmCachingVOILUT,
	prmCacheControl,
	prmDump,
	kNumParams
};
//----------------------------------------------------------------------------------------
const tQueryParam QueryParamTable[kNumParams] = {
	{ "WC", eInteger, false, false, 0, NULL },
	{ "WW", eInteger, false, false, 0, NULL },
	{ "SCALE", eReal, false, false, 0, NULL },
	{ "WIDTH", eInteger, false, false, 0, NULL },
	{ "HEIGHT", eInteger, false, false, 0, NULL },
	{ "ZX", eInteger, false, false, 0, NULL },
	{ "ZY", eInteger, false, false, 0, NULL },
	{ "CODEC", eString, false, false, 0, NULL },
	{ "CMPLEV", eInteger, false, false, 0, NULL },
	{ "PRC", eInteger, false, false, 0, NULL },
	{ "VOIPRC", eInteger, false, false, 0, NULL },
	{ "DCTPRC", eInteger, false, false, 0, NULL },
	{ "PSFPRC", eInteger, false, false, 0, NULL },
	{ "CVL", eInteger, false, false, 0, NULL },
	{ "CACHE", eInteger, false, false, 0, NULL },
	{ "DUMP", eInteger, false, false, 0, NULL }
};

//----------------------------------------------------------------------------------------
//
// Attribute Table
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
struct tAttr {
	const unsigned long ltag;
	const tValueType type;
	union {
		int integer;
		float real;
		void* pix;
		char string[256];
	} val;
	unsigned long len;
	bool hasValue;
};
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
enum {
	atrSOPInstanceUID,
	atrModality,
	atrStudyInstanceUID,
	atrSeriesInstanceUID,
	atrPhotometricInterpretation,
	atrNumOfFrames,
	atrRows,
	atrColumns,
	atrBitsAllocated,
	atrBitsStored,
	atrHighBit,
	atrPixelRepresentation,
	atrWindowCenter,
	atrWindowWidth,
	atrRescaleIntercept,
	atrRescaleSlope,
	atrPixelData,
	kNumAttributes
};
//----------------------------------------------------------------------------------------
const tAttr attrTbl[kNumAttributes + 1] = {
	{ DCM_IDSOPINSTANCEUID, eString, 0, 0, false },
	{ DCM_IDMODALITY, eString, 0, 0, false },
	{ DCM_RELSTUDYINSTANCEUID, eString, 0, 0, false },
	{ DCM_RELSERIESINSTANCEUID, eString, 0, 0, false },
	{ DCM_IMGPHOTOMETRICINTERP, eString, 0, 0, false },
	{ DCM_IMGNUMBEROFFRAMES, eString, 0, 0, false },
	{ DCM_IMGROWS, eInteger, 0, 0, false },
	{ DCM_IMGCOLUMNS, eInteger, 0, 0, false },
	{ DCM_IMGBITSALLOCATED, eInteger, 0, 0, false },
	{ DCM_IMGBITSSTORED, eInteger, 0, 0, false },
	{ DCM_IMGHIGHBIT, eInteger, 0, 0, false },
	{ DCM_IMGPIXELREPRESENTATION, eInteger, 0, 0, false },
	{ DCM_IMGWINDOWCENTER, eReal, 0, 0, false },
	{ DCM_IMGWINDOWWIDTH, eReal, 0, 0, false },
	{ DCM_IMGRESCALEINTERCEPT, eReal, 0, 0, false },
	{ DCM_IMGRESCALESLOPE, eReal, 0, 0, false },
	{ DCM_PXLPIXELDATA, ePixel, 0, 0, false },
	{ 0xFFFFFFFF, eUnknown, 0, 0, false }
};

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static unsigned long encodedPixelLength(const void* data)
{
	const char* cp = (const char*)data;
	unsigned long length = 0;

	tDcmElementHdr hdr;
	while (true) {
		hdr.ReadHdr(cp, false);
		length += hdr.element_length();
		if (DCM_DLMSEQUENCEDELIMITATIONITEM == hdr.ltag())
			break;
		cp = hdr.next_ptr(cp);
	}

	return length;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
enum input_codec_t {
	eNoCompress,
	eLossyJPEG,
	eLosslessJPEG,
	ePNG,
};
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static bool scanAttributes(tAttr* attr, const char* map, const size_t fileSize, const bool doDump, void* pECB, input_codec_t& input_codec)
{
	const char* block_begin = map;
	const char* const block_end = block_begin + fileSize;

	if (memcmp(block_begin + DCM_PREAMBLELENGTH, "DICM", 4) != 0) {
		return false;
	}
	block_begin += DCM_PREAMBLELENGTH + 4;

	CDcmScannerF scanner(block_begin, block_end);
	tDcmElementHdr hdr;

	const char* cp = block_begin;
	bool explicitVR;
	while (true) {
		int sr = scanner.NextElement(cp, hdr, true);
		if (sr != CDcmScannerF::S_CONTINUE) {
			return false;
		}
		if (hdr.m_tag[0] != DCM_GROUPFILEMETA)
			break;
		if (hdr.m_tag[1] == DCM_TAG_ELEMENT(DCM_METATRANSFERSYNTAX)) {
			char transfer_syntax_uid[DICOM_UI_LENGTH + 1];
			memcpy(transfer_syntax_uid, cp + hdr.m_hsize, hdr.m_length);
			transfer_syntax_uid[hdr.m_length] = '\0';

			if (strcmp(transfer_syntax_uid, DICOM_TRANSFERLITTLEENDIAN) == 0) {
				explicitVR = false;
				input_codec = eNoCompress;
			} else if (strcmp(transfer_syntax_uid, DICOM_TRANSFERLITTLEENDIANEXPLICIT) == 0) {
				explicitVR = true;
				input_codec = eNoCompress;
			} else if (strcmp(transfer_syntax_uid, DICOM_TRANSFERJPEGBASELINEPROCESS1) == 0) {
				explicitVR = true;
				input_codec = eLossyJPEG;
			} else if (strcmp(transfer_syntax_uid, DICOM_TRANSFERJPEGEXTENDEDPROC2AND4) == 0) {
				explicitVR = true;
				input_codec = eLossyJPEG;
			} else if (strcmp(transfer_syntax_uid, DICOM_TRANSFERJPEGLOSSLESSPROC14) == 0) {
				explicitVR = true;
				input_codec = eLosslessJPEG;
			} else if (strcmp(transfer_syntax_uid, DICOM_TRANSFERJPEGLOSSLESSPROCFIRSTORDERREDICT) == 0) {
				explicitVR = true;
				input_codec = eLosslessJPEG;
			} else if (strcmp(transfer_syntax_uid, "PNG") == 0) {
				explicitVR = true;
				input_codec = ePNG;
			} else {
				return false;
			}
		}
		cp = hdr.next_ptr(cp);
	}
	const char* const dataset_begin = cp;

	CDcmDict dict;

	for (tAttr* tp = attr; tp->ltag != 0xFFFFFFFF; cp = hdr.next_ptr(cp)) {
		int sr = scanner.NextElement(cp, hdr, explicitVR);
		if (sr < CDcmScannerF::S_CONTINUE) {
			return false;
		}
		const DCM_TAG ltag = hdr.ltag();
		if (!explicitVR) {
			const DCMDICT* d = dict.Lookup(ltag);
			if (d == NULL) {
				continue;
			}
			hdr.m_vr = DcmGetVRStr(d->representation);
		}
		if (hdr.isSQ()) {
			continue;
		}

		if (ltag > tp->ltag)
			while (ltag > tp->ltag)
				tp++;

		if (ltag == tp->ltag) {
			switch (tp->type) {
			case eInteger:
				tp->val.integer = hdr.get_integer(cp);
				tp->len = hdr.value_length();
				break;
			case eReal:
				tp->val.real = hdr.get_real(cp);
				tp->len = hdr.value_length();
				break;
			case eString:
				hdr.get_string(cp, tp->val.string, sizeof(tp->val.string));
				tp->len = hdr.value_length();
				break;
			case ePixel:
				tp->val.pix = (void*)hdr.value_ptr(cp);
				if (hdr.value_length() == 0xFFFFFFFF)
					tp->len = encodedPixelLength(hdr.value_ptr(cp));
				else
					tp->len = hdr.value_length();
				break;
			}
			tp->hasValue = true;
			tp++;
		}

		if (doDump) {
			dump_attr(pECB, hdr, cp);
		}
	}

	return true;
}

//----------------------------------------------------------------------------------------
//
// VOI table and caching
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
struct tVOIentry {
	char* key;
	int center;
	int width;
};
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static tVOIentry* voiTable = NULL;
static char* voiFileBuf = NULL;
static CRITICAL_SECTION voiTable_crit;
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void loadVOItable(HANDLE hHeap)
{
	CStackFileMapping map("C:\\DIOWave\\etc\\VOITable.txt");
	if (map.lastError() != NO_ERROR)
		return;

	char* fileBuf = (char*)HeapAlloc(hHeap, 0, map.mapSize() + 1);
	if (fileBuf == NULL)
		return;

	memcpy(fileBuf, map.mapPointer(), map.mapSize());
	char* ep = fileBuf + map.mapSize();
	*ep = '\0';

	int entries = 0;

	char* eol;
	char* nl;
	for (char* cp = fileBuf; cp < ep; cp = nl) {
		char* cr = strchr(cp, '\r');
		char* lf = strchr(cp, '\n');
		if (cr != NULL && lf != NULL) {
			eol = cr < lf ? cr : lf;
			nl = cr > lf ? cr + 1 : lf + 1;
		} else if (cr != NULL) {
			eol = cr;
			nl = cr + 1;
		} else if (lf != NULL) {
			eol = lf;
			nl = lf + 1;
		} else {
			eol = nl = ep;
		}
		*eol = '\0';

		if (*cp == '#')
			continue;
		if (strchr(cp, ':') != NULL)
			entries++;
	}

	tVOIentry* vt = (tVOIentry*)HeapAlloc(hHeap, 0, sizeof(tVOIentry) * (entries + 1));
	if (vt == NULL) {
		HeapFree(hHeap, 0, fileBuf);
		return;
	}

	tVOIentry* voi = vt;
	for (char* cp = fileBuf; cp < ep; cp = nl) {
		int llen = strlen(cp);
		nl = cp + strlen(cp) + 1;
		if (*nl == '\r' || *nl == '\n')
			nl++;

		if (*cp == '#')
			continue;
		if (strchr(cp, ':') == NULL)
			continue;

		int center, width;
		if (sscanf(cp, "%*s %d %d", &center, &width) != 2)
			continue;
		voi->center = center;
		voi->width = width;
		voi->key = cp;
		char* tp = strchr(cp, ':');
		*tp = '\0';
		voi++;
	}

	voi->key = NULL;

	if (voiTable != NULL)
		HeapFree(hHeap, 0, voiTable);
	voiTable = vt;

	if (voiFileBuf != NULL)
		HeapFree(hHeap, 0, voiFileBuf);
	voiFileBuf = fileBuf;
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void initVOItable(HANDLE hHeap)
{
	loadVOItable(hHeap);

	if (voiTable != NULL)
		InitializeCriticalSectionAndSpinCount(&voiTable_crit, 0x80000000 + 4096);
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void finishVOItable(HANDLE hHeap)
{
	if (voiTable != NULL) {
		DeleteCriticalSection(&voiTable_crit);
		HeapFree(hHeap, 0, voiTable);
		HeapFree(hHeap, 0, voiFileBuf);
		voiTable = NULL;
		voiFileBuf = NULL;
	}
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void regVOItable(const int wc, const int ww, tAttr* attr)
{
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void checkVOItable(EXTENSION_CONTROL_BLOCK *pECB, tAttr* attr, int& wc, int& ww)
{
	if (voiTable == NULL)
		return;

	CStackCriticalSection cs(&voiTable_crit);

	for (tVOIentry* voi = voiTable; voi->key != NULL; voi++) {
		int klen = strlen(voi->key);
		if (memcmp(pECB->lpszPathInfo, voi->key, klen) == 0
		 && pECB->lpszPathInfo[klen] == '/') {
			wc = voi->center;
			ww = voi->width;
			return;
		}
	}

	for (tVOIentry* voi = voiTable; voi->key != NULL; voi++) {
		if (attr[atrModality].hasValue
		 && strcmp(attr[atrModality].val.string, voi->key) == 0) {
			wc = voi->center;
			ww = voi->width;
			return;
		}
	}
}

//----------------------------------------------------------------------------------------
//
// "ETag" Caching
//
//----------------------------------------------------------------------------------------
//
// ETag = <SOP Instance UID>_<Last Write Time>_<WC>_<WW>_<SCALE>_<WIDTH>_<HEIGHT>_<ZX>_<ZY>_<CODEC>_<COMPLEVEL>
//
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static bool genETag(char* uid, const FILETIME* ft1, const FILETIME* ft2, tQueryParam* prm, char* etag)
{
	static const int tagKey[] = {
		prmWindowCenter,
		prmWindowWidth,
		prmScale,
		prmImageWidth,
		prmImageHeight,
		prmZoomX,
		prmZoomY,
		prmCodec,
		prmCompLevel,
		-1,
	};

	char* ep = etag;
	const char* const etag_limit = ep + ETAG_VALUE_LIMIT - 1;

	for (const char* cp = uid; *cp != '\0'; cp++) {
		if (ep == etag_limit)
			return false;
		const char ch = *cp;
		*ep++ = ch;
	}

	if (ep + 17 >= etag_limit)
		return false;
	sprintf(ep, "_%08X%08X", ft1->dwHighDateTime, ft1->dwLowDateTime);
	ep += 17;

	if (ep + 17 >= etag_limit)
		return false;
	sprintf(ep, "_%08X%08X", ft2->dwHighDateTime, ft2->dwLowDateTime);
	ep += 17;

	for (const int* key = tagKey; *key != -1; key++) {
		if (ep == etag_limit)
			return false;
		*ep++ = '_';
		
		if (prm[*key].hasValue) {
			const char* cp = prm[*key].val_ptr;
			const char* const val_end = cp + prm[*key].val_len;
			for (; cp < val_end; cp++) {
				if (ep == etag_limit)
					return false;
				const char ch = *cp;
				*ep++ = ch;
			}
		}
	}
	*ep = '\0';

	return true;
}

//----------------------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void invert_row_inplace(unsigned char* outRowPtr, const unsigned int output_width)
{
	for (unsigned int i = 0; i < output_width; i++, outRowPtr++)
		*outRowPtr = 256 - *outRowPtr;
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void invert_row_copy(unsigned char* rowPtr, unsigned char* outRowPtr, const unsigned int output_width)
{
	for (unsigned int i = 0; i < output_width; i++)
		*outRowPtr++ = 256 - *rowPtr++;
}

//========================================================================================
//
// Export to CODECs
//
//========================================================================================
//========================================================================================
// 
//========================================================================================
int codec_write_client(void* ecb, void* buf, size_t len)
{
	return write_client(ecb, buf, len);
}
//========================================================================================
// 
//========================================================================================
void codec_log_message(const char* message)
{
	err_log("codec_log_message(", message, ")");
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static HANDLE hCodecHeap = NULL;
//========================================================================================
// 
//========================================================================================
void* codec_mem_alloc(codec_mem_t* cm, size_t size)
{
	if (cm != NULL && size < sizeof(cm->m_buf) - cm->m_allocated_size) {
		char* ptr = cm->m_buf + cm->m_allocated_size;
		cm->m_allocated_size += size;
		return (void*)ptr;
	} else {
		return HeapAlloc(hCodecHeap, 0, size);
	}
}
//========================================================================================
// 
//========================================================================================
void codec_mem_free(codec_mem_t* cm, void* ptr)
{
	if (cm != NULL && ptr < cm->m_buf + sizeof(cm->m_buf))
		return;
	else
		HeapFree(hCodecHeap, 0, ptr);
}

//----------------------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------------------
static long HttpExtensionProc_running_count = 0;
static long total_request_count = 0;
static long total_response_count = 0;
static long total_etag_hit_count = 0;
static long total_error_count = 0;
static FILETIME startFT;
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static DWORD dump_status(EXTENSION_CONTROL_BLOCK *pECB)
{
	const char* text_content_type = "Content-Type: text/plain\r\n\r\n";
	DWORD len = strlen(text_content_type);
	if (!pECB->ServerSupportFunction(pECB->ConnID, HSE_REQ_SEND_RESPONSE_HEADER, (LPVOID)"200 OK", &len, (LPDWORD)text_content_type)) {
		sprintf((char*)pECB->lpszLogData, " ServerSupportFunction(HSE_REQ_SEND_RESPONSE_HEADER) Error as %d", GetLastError());
		pECB->dwHttpStatusCode = HSE_STATUS_ERROR;
		return HSE_STATUS_ERROR;
	}

	char msg[256];

	SYSTEMTIME st;
	FileTimeToSystemTime(&startFT, &st);
	sprintf(msg, "Extension started (GMT): %04d-%02d-%02dT%02d:%02d:%02d.%03d\r\n", st.wYear, st.wMonth, st.wDay, st.wHour, st.wMinute, st.wSecond, st.wMilliseconds);
	write_client((void*)pECB, (void*)msg, strlen(msg));

	sprintf(msg, "Total requested count: %d\r\nSuccessfull response count: %d\r\nETag hit count: %d\r\nError Count: %d\r\n", total_request_count, total_response_count, total_etag_hit_count, total_error_count);
	write_client((void*)pECB, (void*)msg, strlen(msg));

	sprintf(msg, "Current HttpExtensionProc() running count: %d\r\n", HttpExtensionProc_running_count);
	write_client((void*)pECB, (void*)msg, strlen(msg));

	if (HeapValidate(hCodecHeap, 0, NULL)) {
		const char* msg = "hCodecHeap is valid\r\n";
		write_client((void*)pECB, (void*)msg, strlen(msg));
	} else {
		const char* msg = "hCodecHeap is invalid\r\n";
		write_client((void*)pECB, (void*)msg, strlen(msg));
	}

	sprintf(msg, "Current mbuf_free_count(): %d\r\n", mbuf_free_count());
	write_client((void*)pECB, (void*)msg, strlen(msg));

	pECB->dwHttpStatusCode = HSE_STATUS_SUCCESS;
	return HSE_STATUS_SUCCESS;
}

//========================================================================================
//
// ISAPI Functions
//
//========================================================================================
//========================================================================================
// GetExtensionVersion
//========================================================================================
BOOL WINAPI GetExtensionVersion(HSE_VERSION_INFO  *pVer)
{
    pVer->dwExtensionVersion = MAKELONG( HSE_VERSION_MINOR,
                                         HSE_VERSION_MAJOR );
    strncpy(pVer->lpszExtensionDesc, "DIOWave image processor <dio-image.isa> Ver. 0.9", HSE_MAX_EXT_DLL_NAME_LEN);

	HRESULT hr = ReadConfig(gDWConfig, NULL, false);
	if (FAILED(hr))
		return FALSE;

	if (gDWConfig->iis.version >= 400)
		SuccessCode = HSE_STATUS_SUCCESS_AND_KEEP_CONN;

	init_err_log();

	hCodecHeap = HeapCreate(0, 32 * 1024 * 1024, 0);
	if (hCodecHeap == NULL) {
		finish_err_log();
		return FALSE;
	}

	InitETag(hCodecHeap);
	mbuf_init(hCodecHeap);
	initVOItable(hCodecHeap);
#if _DIO_IMAGE_LOG_
	init_logging();
#endif

	for (int i = 0; i < gDWConfig->num_vol; i++) {
		tVOLprm* vol = gDWConfig->vol_tbl + i;
		if (vol->user != NULL) {
			DWORD err = vol->logon();
			if (err != NO_ERROR) {
				char msg[256];
				sprintf(msg, "logon failed for %s as %d", vol->label, err);
				err_log(msg);
				return FALSE;
			}
		}
	}

	GetSystemTimeAsFileTime(&startFT);
	HttpExtensionProc_running_count = 0;
	total_request_count = 0;
	total_response_count = 0;
	total_etag_hit_count = 0;
	total_error_count = 0;

	return TRUE;
}

//========================================================================================
// TerminateExtension
//========================================================================================
BOOL WINAPI TerminateExtension(DWORD dwFlags)
{
#if _DIO_IMAGE_LOG_
	finish_logging();
#endif
	finishVOItable(hCodecHeap);
	mbuf_uninit(hCodecHeap);
	FinishETag(hCodecHeap);

	HeapDestroy(hCodecHeap);

	finish_err_log();

	ReleaseConfigData(gDWConfig);

	return TRUE;
}
//========================================================================================
// HttpExtensionProc
//========================================================================================
DWORD WINAPI HttpExtensionProc(EXTENSION_CONTROL_BLOCK *pECB)
{
try {
	*pECB->lpszLogData = '\0';
	char httpHeader[1024];

	//--------------------
	// Parse QUERY_STRING
	//--------------------
	tQueryParam prm[kNumParams];
	memcpy(prm, QueryParamTable, sizeof(prm));
	GetMultiQueryValue(pECB->lpszQueryString, prm, kNumParams);

	//--------------------
	// check dump request
	//--------------------
	bool doDump = prm[prmDump].hasParam;
	if (doDump) {
		const int dump_type = prm[prmDump].atoi();
		if (dump_type == 2) {
			return dump_status(pECB);
		} else if (dump_type != 1) {
			doDump = false;
		}
	}

	AutoInterlockedCounter counter(&HttpExtensionProc_running_count);
	InterlockedIncrement(&total_request_count);

	//------------------------
	// Establish File Mapping
	//------------------------
	const tVOLprm* vol;
	DWORD len = gDWConfig->mapStoragePath(pECB->lpszPathInfo, NULL, &vol);
	char* const filePath = (char*)_alloca(len + 1);
	gDWConfig->mapStoragePath(pECB->lpszPathInfo, filePath);

	if (vol->hToken) {
		if (!ImpersonateLoggedOnUser(vol->hToken)) {
			char msg[256];
			sprintf(msg, "ImpersonateLoggedOnUser() Failed = %d", GetLastError());
			throw expInternalError(msg, true);
		}
	}

	CStackFileMapping fileMapping(filePath);

	if (vol->hToken)
		RevertToSelf();

	if (fileMapping.lastError() != NO_ERROR) {
		char msg[256];
		sprintf(msg, "CStackFileMapping() Failed = %d", fileMapping.lastError());
		throw expInternalError(msg, fileMapping.lastError() != ERROR_FILE_NOT_FOUND);
	}

	const char* map = (const char*)fileMapping.mapPointer();
	size_t fileSize = fileMapping.mapSize();

	//--------------
	// Select Codec
	//--------------
	const CStackCodec::eCodec codec_type = (prm[prmCodec].hasValue && prm[prmCodec].strcmp("PNG") == 0)
											? CStackCodec::PNG_CODEC
											: (prm[prmCodec].hasValue && prm[prmCodec].strcmp("BMP") == 0) 
	                                     		? CStackCodec::BMP_CODEC
	                                    		: CStackCodec::JPEG_CODEC;

	//-------------------------
	// Quick Precision Setting
	//------------------------
	if (prm[prmPrecision].hasValue) {
		prm[prmVOIPrecision].hasValue = true;
		prm[prmVOIPrecision].val_ptr = prm[prmPrecision].val_ptr;
		prm[prmVOIPrecision].val_len = prm[prmPrecision].val_len;

		prm[prmDCTPrecision].hasValue = true;
		prm[prmDCTPrecision].val_ptr = prm[prmPrecision].val_ptr;
		prm[prmDCTPrecision].val_len = prm[prmPrecision].val_len;

		prm[prmPSFPrecision].hasValue = true;
		prm[prmPSFPrecision].val_ptr = prm[prmPrecision].val_ptr;
		prm[prmPSFPrecision].val_len = prm[prmPrecision].val_len;
	}

	//-----------------
	// Scan Attributes
	//-----------------
	if (doDump) {
		const char* text_content_type = "Content-Type: text/plain\r\n\r\n";
		len = strlen(text_content_type);
		if (!pECB->ServerSupportFunction(pECB->ConnID, HSE_REQ_SEND_RESPONSE_HEADER, (LPVOID)"200 OK", &len, (LPDWORD)text_content_type)) {
			sprintf((char*)pECB->lpszLogData, " ServerSupportFunction(HSE_REQ_SEND_RESPONSE_HEADER) Error as %d", GetLastError());
			InterlockedIncrement(&total_error_count);
			pECB->dwHttpStatusCode = HSE_STATUS_ERROR;
			return HSE_STATUS_ERROR;
		}
	}

	tAttr attr[kNumAttributes + 1];
	memcpy(attr, attrTbl, sizeof(attr));
	input_codec_t input_codec;
	const bool supportedSyntax = scanAttributes(attr, map, fileSize, doDump, pECB, input_codec);
	if (!supportedSyntax) {
		throw expInternalError( "scanAttributes() Failed", false);
	}

	if (doDump) {
		InterlockedIncrement(&total_response_count);
		pECB->dwHttpStatusCode = HSE_STATUS_SUCCESS;
		return HSE_STATUS_SUCCESS;
	}

	//
	// Check Attributes existance
	//
	if (!(attr[atrPhotometricInterpretation].hasValue
	   && attr[atrRows].hasValue
	   && attr[atrColumns].hasValue
	   && attr[atrBitsAllocated].hasValue
	   && attr[atrBitsStored].hasValue
	   && attr[atrHighBit].hasValue
	   && attr[atrPixelRepresentation].hasValue
	   && attr[atrPixelData].hasValue)) {

		throw expInternalError("missing mandatory attributes", false);
	}

	//
	// cache params
	//
	const bool isColor = strcmp(attr[atrPhotometricInterpretation].val.string, "RGB") == 0;
	const bool isMonochrome1 = strcmp(attr[atrPhotometricInterpretation].val.string, "MONOCHROME1") == 0;
	const bool isMonochrome2 = strcmp(attr[atrPhotometricInterpretation].val.string, "MONOCHROME2") == 0;
	const bool isMonochrome = isMonochrome1 || isMonochrome2;          
	const int bitsAlloc = attr[atrBitsAllocated].val.integer;

	//
	//
	//
	if (bitsAlloc == 8) {
		if (input_codec == eLosslessJPEG)
			throw expInternalError("8-bit gray lossless jpeg decording is not supported", false);
		if (input_codec == ePNG)
			throw expInternalError("8-bit gray PNG decording is not supported", false);
	}
	if (isColor) {
		if (input_codec == eLosslessJPEG)
			throw expInternalError("8-bit color lossless jpeg decording is not supported", false);
		if (input_codec == ePNG)
			throw expInternalError("8-bit color PNG decording is not supported", false);
	}

	//------------
	// Check ETag
	//------------
	char etag[ETAG_VALUE_LIMIT];
	bool got_valid_etag = false;
	FILETIME lastWriteFT;
	GetFileTime(fileMapping.fileHandle(), NULL, NULL, &lastWriteFT);
	if (genETag(attr[atrSOPInstanceUID].val.string, &lastWriteFT, &startFT, prm, etag)) {
		len = sizeof(httpHeader);
		if (pECB->GetServerVariable(pECB->ConnID, "HTTP_IF_NONE_MATCH", httpHeader, &len)) {
			const char* tp = httpHeader;
			tp = strchr(tp, '"');
			if (tp != NULL) {
				tp++;

				const char* ep = etag;
				while (*ep != '\0' && *tp != '\0' && *tp != '"' && *ep == *tp) {
					ep++;
					tp++;
				}
				if (*ep == '\0' && *tp == '"') {
					got_valid_etag = true;
				}
			}
		}
	} else {
		*etag = '\0';
	}

	if (*etag != '\0' && CheckETagEntry(etag) && got_valid_etag) {
		//--------------------
		// Send HTTP Response
		//--------------------
		sprintf(httpHeader, "Content-Type: %s\r\nContent-Length: 0\r\nETag: \"%s\"\r\n\r\n", CStackCodec::content_type(codec_type), etag);
		len = strlen(httpHeader);
		if (pECB->ServerSupportFunction(pECB->ConnID, HSE_REQ_SEND_RESPONSE_HEADER, (LPVOID)"304 Not Modified", &len, (LPDWORD)httpHeader)) {
			InterlockedIncrement(&total_etag_hit_count);
			pECB->dwHttpStatusCode = SuccessCode;
			return SuccessCode;
		} else {
			sprintf((char*)pECB->lpszLogData, " ServerSupportFunction(HSE_REQ_SEND_RESPONSE_HEADER) Error as %d", GetLastError());
			InterlockedIncrement(&total_error_count);
			pECB->dwHttpStatusCode = HSE_STATUS_ERROR;
			return HSE_STATUS_ERROR;
		}
	}

	//
	// Prepare Image Scaling
	//
	const unsigned int pixel_width = attr[atrColumns].val.integer;
	const unsigned int pixel_height = attr[atrRows].val.integer;

	float scale = 1.0;
	unsigned int output_width = pixel_width;
	unsigned int output_height = pixel_height;

	unsigned int view_width = pixel_width;
	unsigned int view_height = pixel_height;
	int view_x = 0;
	int view_y = 0;

	if (prm[prmZoomX].hasValue && prm[prmZoomY].hasValue) {
		if (prm[prmScale].hasValue)
			scale = prm[prmScale].atof();

		if (prm[prmImageWidth].hasValue && prm[prmImageHeight].hasValue) {
			output_width = prm[prmImageWidth].atoi();
			output_height = prm[prmImageHeight].atoi();
		} else if (prm[prmImageWidth].hasValue) {
			output_width = prm[prmImageWidth].atoi();
			output_height = output_width;
		} else if (prm[prmImageHeight].hasValue) {
			output_height = prm[prmImageHeight].atoi();
			output_width = output_height;
		}

		view_width = xceil(output_width / scale);
		view_x = prm[prmZoomX].atoi() - view_width / 2;
		if (view_x < 0)
			view_x = 0;
		else if (view_x + view_width > pixel_width)
			view_x = pixel_width - view_width;

		view_height = xceil(output_height / scale);
		view_y = prm[prmZoomY].atoi() - view_height / 2;
		if (view_y < 0)
			view_y = 0;
		else if (view_y + view_height > pixel_height)
			view_y = pixel_height - view_height;

	} else {
		if (prm[prmScale].hasValue) {
			scale = prm[prmScale].atof();
			output_width = xceil(pixel_width * scale);
			output_height = xceil(pixel_height * scale);
		} else if (prm[prmImageWidth].hasValue && prm[prmImageHeight].hasValue) {
			output_width = prm[prmImageWidth].atoi();
			output_height = prm[prmImageHeight].atoi();
			float x_scale = (float)output_width / (float)pixel_width;
			float y_scale = (float)output_height / (float)pixel_height;
			if (x_scale < y_scale) {
				scale = x_scale;
				output_height = xceil(pixel_height * scale);
			} else {
				scale = y_scale;
				output_width = xceil(pixel_width * scale);
			}
		} else if (prm[prmImageWidth].hasValue) {
			output_width = prm[prmImageWidth].atoi();
			scale = (float)output_width / (float)pixel_width;
			output_height = xceil(pixel_height * scale);
		} else if (prm[prmImageHeight].hasValue) {
			output_height = prm[prmImageHeight].atoi();
			scale = (float)output_height / (float)pixel_height;
			output_width = xceil(pixel_width * scale);
		}
	}

	const int skip = xfloor(1.0 / scale);
	const float inv_scale = 1.0 / scale;

	const bool doResampling = (float)skip != 1.0 / scale;

	const calc_psf1_row_t calc_psf1_row = prm[prmPSFPrecision].hasValue ? (prm[prmPSFPrecision].atoi() == 1 ? calc_psf1_row_float : calc_psf1_row_int) : calc_psf1_row_int;

	//---------------------
	// Setup Encoder Codec
	//---------------------
	void* codec_ctx = (void*)_alloca(CStackCodec::ctx_size(codec_type));
	CStackCodec codec(pECB, codec_type, codec_ctx, output_width, output_height, isColor, prm[prmCompLevel].hasValue ? prm[prmCompLevel].atoi() : -1, prm[prmDCTPrecision].hasValue ? prm[prmDCTPrecision].atoi() : 0);

	if (!codec.is_buffered()) {
		//-----------------------------------------
		// Set HTTP Response for un-buffered codec
		//-----------------------------------------
		//
		// Build Response Header
		//
		size_t content_length = codec.stream_length();
		if (*etag != '\0')
			sprintf(httpHeader, "Content-Type: %s\r\nContent-Length: %d\r\nETag: \"%s\"\r\n\r\n", codec.content_type(), content_length, etag);
		else
			sprintf(httpHeader, "Content-Type: %s\r\nContent-Length: %d\r\n\r\n", codec.content_type(), content_length);
		len = strlen(httpHeader);
	
		//
		// Send Response Header
		//
		if (!pECB->ServerSupportFunction(pECB->ConnID, HSE_REQ_SEND_RESPONSE_HEADER, (LPVOID)"200 OK", &len, (LPDWORD)httpHeader)) {
			sprintf((char*)pECB->lpszLogData, " ServerSupportFunction(HSE_REQ_SEND_RESPONSE_HEADER) Error as %d", GetLastError());
			InterlockedIncrement(&total_error_count);
			pECB->dwHttpStatusCode = HSE_STATUS_ERROR;
			return HSE_STATUS_ERROR;
		}
	}

	size_t outRowBufLength = sizeof(char) * output_width * (isColor ? 3 : 1);
	outRowBufLength += (outRowBufLength % 4) != 0 ? 4 - (outRowBufLength % 4) : 0;
	unsigned char* outRowBuf = (unsigned char*)_alloca(outRowBufLength);

	if (isMonochrome && bitsAlloc == 16) {
		//-------------------------------
		// Proc 16-bits Monochrome Image
		//-------------------------------

		// Masks
		const unsigned short signMask = 0xFFFF << attr[atrBitsStored].val.integer;
		const unsigned short signBitMask = 1 << (attr[atrBitsStored].val.integer - 1);
		const unsigned int shr_count = attr[atrHighBit].val.integer - (attr[atrBitsStored].val.integer - 1); 

		const bool isSigned = attr[atrPixelRepresentation].val.integer != 0;

		// VOI
		bool hasPrmWL = false;
		int prmWC, prmWW;
		if ((prm[prmWindowWidth].hasValue && prm[prmWindowWidth].atoi() != 0)
		 && (prm[prmWindowCenter].hasValue)) {
		 	prmWC = prm[prmWindowCenter].atoi();
			prmWW = prm[prmWindowWidth].atoi();
			hasPrmWL = true;
			regVOItable(prmWC, prmWW, attr);
		}

		bool hasAtrWL = false;
		int atrWC, atrWW;
		if ((attr[atrWindowWidth].hasValue && attr[atrWindowWidth].val.integer != 0)
		 && (attr[atrWindowCenter].hasValue)){
			atrWC = attr[atrWindowCenter].val.real;
			atrWW = attr[atrWindowWidth].val.real;
			hasAtrWL = true;
		}

		int wc, ww;
		if (hasPrmWL) {
			wc = prmWC;
			ww = prmWW;
		} else if (hasAtrWL) {
			wc = atrWC;
			ww = atrWW;
		} else {
			ww = 1 << attr[atrBitsStored].val.integer;
			wc = attr[atrPixelRepresentation].val.integer ? 0 : ww / 2;
			checkVOItable(pECB, attr, wc, ww);
		}

		if (isMonochrome1)
			ww = -ww;

		// Rescale
		const bool doRescale = attr[atrRescaleSlope].hasValue && attr[atrRescaleIntercept].hasValue;

		const float rescaleSlope = doRescale ? attr[atrRescaleSlope].val.real : 1.0;
		const int shift = doRescale ? (int)attr[atrRescaleIntercept].val.real - (wc - ww / 2) : - (wc - ww / 2);
		const float slope = 256.0 / (float)ww;

		voi_proc_c voi(
				prm[prmCachingVOILUT].hasValue ? prm[prmCachingVOILUT].atoi() == 0 : false,
				prm[prmVOIPrecision].hasValue ? prm[prmVOIPrecision].atoi() : 0,
				isSigned,
				doRescale,
				shr_count,
				signMask, signBitMask,
				rescaleSlope,
				shift, slope);

		if (input_codec == eLossyJPEG) {
			//------------------------
			// Proc 12-bit Lossy JPEG
			//------------------------

			// Setup Decoder
			void* djpeg_ctx = (void*)_alloca(CStackJPEGDecoder12::ctx_size());

			unsigned short* inRowBuf = (unsigned short*)_alloca(sizeof(short) * pixel_width);

			CStackJPEGDecoder12 djpeg(djpeg_ctx, attr[atrPixelData].val.pix, attr[atrPixelData].len);

			if (doResampling) {
				//-------------------------------------------
				// Do Resampling for 12-bits Monochrome JPEG
				//-------------------------------------------

				// Allocate and Setup Resampling Pre Fetch Buffer
				unsigned char* buf_pool = (unsigned char*)_alloca(sizeof(char) * pixel_width * 2);

				unsigned char* buf_ptr[2];
				int buf_row[2];
				for (int r = 0; r < 2; r++) {
					buf_ptr[r] = buf_pool + pixel_width * r;
					buf_row[r] = -1;
				}
				int read_row = -1;

				float yi = view_y;
				for (int y = 0; y < output_height; y++, yi += inv_scale) {
					int yi0, yi1, dy0;
					calc_coord(yi, pixel_height, yi0, yi1, dy0);

					int r;
					unsigned char* buf_ptr2[2];
					if (buf_row[0] == yi0) {
						r = 0;
						buf_ptr2[0] = buf_ptr[r];
					} else if (buf_row[1] == yi0) {
						r = 1;
						buf_ptr2[0] = buf_ptr[r];
					} else {
						r = 0;
						while (read_row < yi0) {
							djpeg.proc(inRowBuf);
							read_row++;
						}	
						buf_row[r] = yi0;
						buf_ptr2[0] = buf_ptr[r];
						voi.proc(inRowBuf, buf_ptr2[0], pixel_width);
					}
					if (yi1 > yi0) {
						r = (r + 1) & 1;
						if (buf_row[r] == yi1) {
							buf_ptr2[1] = buf_ptr[r];
						} else {
							while (read_row < yi1) {
								djpeg.proc(inRowBuf);
								read_row++;
							}	
							buf_row[r] = yi1;
							buf_ptr2[1] = buf_ptr[r];
							voi.proc(inRowBuf, buf_ptr2[1], pixel_width);
						}
					}

					calc_psf1_row(buf_ptr2, pixel_width, outRowBuf, output_width, inv_scale, view_x, yi0, yi1, dy0); 
					codec.proc(outRowBuf);

				}

			} else {
				//-------------------------------------------
				// Simple Skipping for 12-bits Monochrome JPEG
				//-------------------------------------------

				for (int i = view_y; i != 0; i--)
					djpeg.proc(inRowBuf);
					
				for (int y = 0; y < output_height; y++) {
					djpeg.proc(inRowBuf);
					unsigned short* rowPtr = inRowBuf + view_x;
					voi.proc(rowPtr, outRowBuf, output_width, skip);
					codec.proc(outRowBuf);
					int i = skip;
					while (--i) {
						djpeg.proc(inRowBuf);
					}
				}
			}

			djpeg.flush(inRowBuf);
		} else if (input_codec == eLosslessJPEG) {
			//---------------------------
			// Proc Lossless JPEG
			//---------------------------

			// Setup Decoder
			void* djpeg_ctx = (void*)_alloca(CStackJPEGLLDecoder::ctx_size());

			unsigned short* inRowBuf = (unsigned short*)_alloca(sizeof(short) * pixel_width);

			CStackJPEGLLDecoder djpeg(djpeg_ctx, attr[atrPixelData].val.pix, attr[atrPixelData].len);

			if (doResampling) {
				//---------------------------------------------------
				// Do Resampling for Monochrome 16-bit Lossless JPEG
				//---------------------------------------------------

				// Allocate and Setup Resampling Pre Fetch Buffer
				unsigned char* buf_pool = (unsigned char*)_alloca(sizeof(char) * pixel_width * 2);

				unsigned char* buf_ptr[2];
				int buf_row[2];
				for (int r = 0; r < 2; r++) {
					buf_ptr[r] = buf_pool + pixel_width * r;
					buf_row[r] = -1;
				}
				int read_row = -1;

				float yi = view_y;
				for (int y = 0; y < output_height; y++, yi += inv_scale) {
					int yi0, yi1, dy0;
					calc_coord(yi, pixel_height, yi0, yi1, dy0);

					int r;
					unsigned char* buf_ptr2[2];
					if (buf_row[0] == yi0) {
						r = 0;
						buf_ptr2[0] = buf_ptr[r];
					} else if (buf_row[1] == yi0) {
						r = 1;
						buf_ptr2[0] = buf_ptr[r];
					} else {
						r = 0;
						while (read_row < yi0) {
							djpeg.proc(inRowBuf);
							read_row++;
						}	
						buf_row[r] = yi0;
						buf_ptr2[0] = buf_ptr[r];
						voi.proc(inRowBuf, buf_ptr2[0], pixel_width);
					}
					if (yi1 > yi0) {
						r = (r + 1) & 1;
						if (buf_row[r] == yi1) {
							buf_ptr2[1] = buf_ptr[r];
						} else {
							while (read_row < yi1) {
								djpeg.proc(inRowBuf);
								read_row++;
							}	
							buf_row[r] = yi1;
							buf_ptr2[1] = buf_ptr[r];
							voi.proc(inRowBuf, buf_ptr2[1], pixel_width);
						}
					}

					calc_psf1_row(buf_ptr2, pixel_width, outRowBuf, output_width, inv_scale, view_x, yi0, yi1, dy0); 
					codec.proc(outRowBuf);

				}

			} else {
				//-----------------------------------------------------
				// Simple Skipping for Monochrome 16-bit Lossless JPEG
				//-----------------------------------------------------

				for (int i = view_y; i != 0; i--)
					djpeg.proc(inRowBuf);
					
				for (int y = 0; y < output_height; y++) {
					djpeg.proc(inRowBuf);
					unsigned short* rowPtr = inRowBuf + view_x;
					voi.proc(rowPtr, outRowBuf, output_width, skip);
					codec.proc(outRowBuf);
					int i = skip;
					while (--i) {
						djpeg.proc(inRowBuf);
					}
				}
			}

			djpeg.flush(inRowBuf);
		
		} else if (input_codec == ePNG) {
			//------------------
			// Proc 16-bit PNG
			//------------------

			// Setup Decoder
			void* dpng_ctx = (void*)_alloca(CStackPNGDecoder::ctx_size());

			unsigned short* inRowBuf = (unsigned short*)_alloca(sizeof(short) * pixel_width);

			CStackPNGDecoder dpng(dpng_ctx, attr[atrPixelData].val.pix, attr[atrPixelData].len);

			if (doResampling) {
				//-------------------------------------------
				// Do Resampling for 16-bits Monochrome PNG
				//-------------------------------------------

				// Allocate and Setup Resampling Pre Fetch Buffer
				unsigned char* buf_pool = (unsigned char*)_alloca(sizeof(char) * pixel_width * 2);

				unsigned char* buf_ptr[2];
				int buf_row[2];
				for (int r = 0; r < 2; r++) {
					buf_ptr[r] = buf_pool + pixel_width * r;
					buf_row[r] = -1;
				}
				int read_row = -1;

				float yi = view_y;
				for (int y = 0; y < output_height; y++, yi += inv_scale) {
					int yi0, yi1, dy0;
					calc_coord(yi, pixel_height, yi0, yi1, dy0);

					int r;
					unsigned char* buf_ptr2[2];
					if (buf_row[0] == yi0) {
						r = 0;
						buf_ptr2[0] = buf_ptr[r];
					} else if (buf_row[1] == yi0) {
						r = 1;
						buf_ptr2[0] = buf_ptr[r];
					} else {
						r = 0;
						while (read_row < yi0) {
							dpng.proc(inRowBuf);
							read_row++;
						}	
						buf_row[r] = yi0;
						buf_ptr2[0] = buf_ptr[r];
						voi.proc(inRowBuf, buf_ptr2[0], pixel_width);
					}
					if (yi1 > yi0) {
						r = (r + 1) & 1;
						if (buf_row[r] == yi1) {
							buf_ptr2[1] = buf_ptr[r];
						} else {
							while (read_row < yi1) {
								dpng.proc(inRowBuf);
								read_row++;
							}	
							buf_row[r] = yi1;
							buf_ptr2[1] = buf_ptr[r];
							voi.proc(inRowBuf, buf_ptr2[1], pixel_width);
						}
					}

					calc_psf1_row(buf_ptr2, pixel_width, outRowBuf, output_width, inv_scale, view_x, yi0, yi1, dy0); 
					codec.proc(outRowBuf);

				}

			} else {
				//-------------------------------------------
				// Simple Skipping for 16-bits Monochrome PNG
				//-------------------------------------------

				for (int i = view_y; i != 0; i--)
					dpng.proc(inRowBuf);
					
				for (int y = 0; y < output_height; y++) {
					dpng.proc(inRowBuf);
					unsigned short* rowPtr = inRowBuf + view_x;
					voi.proc(rowPtr, outRowBuf, output_width, skip);
					codec.proc(outRowBuf);
					int i = skip;
					while (--i) {
						dpng.proc(inRowBuf);
					}
				}
			}

			dpng.flush(inRowBuf);

		} else {
			//--------------------
			// Proc 16-bit Raster
			//--------------------

			unsigned short* pixel = (unsigned short*)attr[atrPixelData].val.pix;

			if (doResampling) {
				//---------------------------------------------
				// Do Resampling for 16-bits Monochrome Raster
				//---------------------------------------------

				// Allocate and Setup Resampling Pre Fetch Buffer
				unsigned char* buf_pool = (unsigned char*)_alloca(sizeof(char) * pixel_width * 2);

				unsigned char* buf_ptr[2];
				int buf_row[2];
				for (int r = 0; r < 2; r++) {
					buf_ptr[r] = buf_pool + pixel_width * r;
					buf_row[r] = -1;
				}

				float yi = view_y;
				for (int y = 0; y < output_height; y++, yi += inv_scale) {
					int yi0, yi1, dy0;
					calc_coord(yi, pixel_height, yi0, yi1, dy0);

					int r;
					unsigned char* buf_ptr2[2];
					if (buf_row[0] == yi0) {
						r = 0;
						buf_ptr2[0] = buf_ptr[r];
					} else if (buf_row[1] == yi0) {
						r = 1;
						buf_ptr2[0] = buf_ptr[r];
					} else {
						r = 0;
						buf_row[r] = yi0;
						buf_ptr2[0] = buf_ptr[r];
						voi.proc(pixel + pixel_width * yi0, buf_ptr2[0], pixel_width);
					}
					if (yi1 > yi0) {
						r = (r + 1) & 1;
						if (buf_row[r] == yi1) {
							buf_ptr2[1] = buf_ptr[r];
						} else {
							buf_row[r] = yi1;
							buf_ptr2[1] = buf_ptr[r];
							voi.proc(pixel + pixel_width * yi1, buf_ptr2[1], pixel_width);
						}
					}

					calc_psf1_row(buf_ptr2, pixel_width, outRowBuf, output_width, inv_scale, view_x, yi0, yi1, dy0); 
					codec.proc(outRowBuf);
				}

			} else {
				//-----------------------------------------------
				// Simple Skipping for 16-bits Monochrome Raster
				//-----------------------------------------------

				pixel += pixel_width * view_y;

				for (int y = 0; y < output_height; y++) {
					unsigned short* rowPtr = pixel + view_x;
					voi.proc(rowPtr, outRowBuf, output_width, skip);
					codec.proc(outRowBuf);
					pixel += pixel_width * skip;
				}
			}
		}

	} else if (isMonochrome && bitsAlloc == 8) {
		//------------------------------
		// Proc 8-bits Monochrome Image
		//------------------------------

		if (input_codec == eLossyJPEG) {
			//------------------
			// Proc 8-bits JPEG
			//------------------

			// Setup Decoder
			void* djpeg_ctx = (void*)_alloca(CStackJPEGDecoder8::ctx_size());

			CStackJPEGDecoder8 djpeg(djpeg_ctx, attr[atrPixelData].val.pix, attr[atrPixelData].len);

			if (doResampling) {
				//------------------------------------------
				// Do Resampling for 8-bits Monochrome JPEG
				//------------------------------------------

				// Allocate and Setup Resampling Pre Fetch Buffer
				unsigned char* buf_pool = (unsigned char*)_alloca(sizeof(char) * pixel_width * 2);

				unsigned char* buf_ptr[2];
				int buf_row[2];
				for (int r = 0; r < 2; r++) {
					buf_ptr[r] = buf_pool + pixel_width * r;
					buf_row[r] = -1;
				}
				int read_row = -1;

				float yi = view_y;
				for (int y = 0; y < output_height; y++, yi += inv_scale) {
					int yi0, yi1, dy0;
					calc_coord(yi, pixel_height, yi0, yi1, dy0);

					int r;
					unsigned char* buf_ptr2[2];
					if (buf_row[0] == yi0) {
						r = 0;
						buf_ptr2[0] = buf_ptr[r];
					} else if (buf_row[1] == yi0) {
						r = 1;
						buf_ptr2[0] = buf_ptr[r];
					} else {
						r = 0;	
						buf_row[r] = yi0;
						buf_ptr2[0] = buf_ptr[r];
						while (read_row < yi0) {
							djpeg.proc(buf_ptr2[0]);
							read_row++;
						}
					}
					if (yi1 > yi0) {
						r = (r + 1) & 1;
						if (buf_row[r] == yi1) {
							buf_ptr2[1] = buf_ptr[r];
						} else {
							buf_row[r] = yi1;
							buf_ptr2[1] = buf_ptr[r];
							while (read_row < yi1) {
								djpeg.proc(buf_ptr2[1]);
								read_row++;
							}	
						}
					}
					calc_psf1_row(buf_ptr2, pixel_width, outRowBuf, output_width, inv_scale, view_x, yi0, yi1, dy0);
					if (isMonochrome1)
						invert_row_inplace(outRowBuf, output_width);
					codec.proc(outRowBuf);
				}
				djpeg.flush(buf_pool);

			} else {
				//------------------------------------------
				// Simple Skipping for 8-bits Monochrome JPEG
				//------------------------------------------

				unsigned char* inRowBuf = (unsigned char*)_alloca(sizeof(char) * pixel_width);

				for (int i = view_y; i != 0; i--)
					djpeg.proc(inRowBuf);

				for (int y = 0; y < output_height; y++) {
					djpeg.proc(inRowBuf);
					unsigned char* rowPtr = inRowBuf + view_x;
					if (skip > 1) {
						unsigned char* ip = (unsigned char*)rowPtr;
						unsigned char* op = (unsigned char*)outRowBuf;
						unsigned char* ep = (unsigned char*)outRowBuf + output_width;
						while (op < ep) {
							*op++ = *ip;
							ip += skip;
						}
						if (isMonochrome1)
							invert_row_inplace(outRowBuf, output_width);
						codec.proc(outRowBuf);
						int i = skip;
						while (--i) {
							djpeg.proc(inRowBuf);
						}
					} else {
						if (isMonochrome1)
							invert_row_inplace(rowPtr, output_width);
						codec.proc(rowPtr);
					}
				}

				djpeg.flush(inRowBuf);

			}

		} else if (input_codec == eLosslessJPEG) {
			throw expInternalError("8-bit gray lossless jpeg decording is not supported", false);
		} else if (input_codec == ePNG) {
			throw expInternalError("8-bit gray PNG decording is not supported", false);
		} else {
			//-------------------------------
			// Proc 8-bits Monochrome Raster
			//-------------------------------
			unsigned char* pixel = (unsigned char*)attr[atrPixelData].val.pix;

			if (doResampling) {
				//--------------------------------------------
				// Do Resampling for 8-bits Monochrome Raster
				//--------------------------------------------

				float yi = view_y;
				for (int y = 0; y < output_height; y++, yi += inv_scale) {
					int yi0, yi1, dy0;
					calc_coord(yi, pixel_height, yi0, yi1, dy0);

					unsigned char* buf_ptr2[2];
					buf_ptr2[0] = pixel + pixel_width * yi0;
					buf_ptr2[1] = pixel + pixel_width * yi1;

					calc_psf1_row(buf_ptr2, pixel_width, outRowBuf, output_width, inv_scale, view_x, yi0, yi1, dy0); 
					if (isMonochrome1)
						invert_row_inplace(outRowBuf, output_width);
					codec.proc(outRowBuf);
				}

			} else {
				//----------------------------------------------
				// Simple Skipping for 8-bits Monochrome Raster
				//----------------------------------------------

				pixel += view_y * pixel_width;

				for (int y = 0; y < output_height; y++) {
					unsigned char* rowPtr = pixel + view_x;
					if (skip > 1) {
						unsigned char* ip = (unsigned char*)rowPtr;
						unsigned char* op = (unsigned char*)outRowBuf;
						unsigned char* ep = (unsigned char*)outRowBuf + output_width;
						while (op < ep) {
							*op++ = *ip;
							ip += skip;
						}
						if (isMonochrome1)
							invert_row_inplace(outRowBuf, output_width);
						codec.proc(outRowBuf);
					} else {
						if (isMonochrome1) {
							invert_row_copy(rowPtr, outRowBuf, output_width);
							codec.proc(outRowBuf);
						} else {
							codec.proc(rowPtr);
						}
					}
					pixel += pixel_width * skip;
				}
			}
		}

	} else if (isColor && bitsAlloc == 8) {
		//------------------------------
		// Proc Color Image
		//------------------------------

		unsigned char* pixel = (unsigned char*)attr[atrPixelData].val.pix;
		size_t pixelLength = attr[atrPixelData].len;

		if (input_codec == eLossyJPEG) {
			//------------------
			// Proc Color JPEG
			//------------------

			// Setup Decoder
			void* djpeg_ctx = (void*)_alloca(CStackJPEGDecoder8::ctx_size());

			CStackJPEGDecoder8 djpeg(djpeg_ctx, pixel, pixelLength);

			if (doResampling) {
				//--------------------------
				// Do Resampling Color JPEG
				//--------------------------

				// Allocate and Setup Resampling Pre Fetch Buffer
				unsigned char* buf_pool = (unsigned char*)_alloca(sizeof(char) * 8 * pixel_width * 2);

				unsigned char* buf_ptr[2];
				int buf_row[2];
				for (int r = 0; r < 2; r++) {
					buf_ptr[r] = buf_pool + 3 * pixel_width * r;
					buf_row[r] = -1;
				}
				int read_row = -1;

				float yi = view_y;
				for (int y = 0; y < output_height; y++, yi += inv_scale) {
					int yi0, yi1, dy0;
					calc_coord(yi, pixel_height, yi0, yi1, dy0);

					int r;
					unsigned char* buf_ptr2[2];
					if (buf_row[0] == yi0) {
						r = 0;
						buf_ptr2[0] = buf_ptr[r];
					} else if (buf_row[1] == yi0) {
						r = 1;
						buf_ptr2[0] = buf_ptr[r];
					} else {
						r = 0;	
						buf_row[r] = yi0;
						buf_ptr2[0] = buf_ptr[r];
						while (read_row < yi0) {
							djpeg.proc(buf_ptr2[0]);
							read_row++;
						}
					}
					if (yi1 > yi0) {
						r = (r + 1) & 1;
						if (buf_row[r] == yi1) {
							buf_ptr2[1] = buf_ptr[r];
						} else {
							buf_row[r] = yi1;
							buf_ptr2[1] = buf_ptr[r];
							while (read_row < yi1) {
								djpeg.proc(buf_ptr2[1]);
								read_row++;
							}	
						}
					}

					unsigned char* op = (unsigned char*)outRowBuf;

					float xi = view_x;
					for (int x = 0; x < output_width; x++, xi += inv_scale) {
						int xi0, xi1, dx0;
						calc_coord(xi, pixel_width, xi0, xi1, dx0);

						calc_psf3(yi0, yi1, dy0, xi0, xi1, dx0, buf_ptr2, op);
						op += 3;
					}
					codec.proc(outRowBuf);
				}

				djpeg.flush(buf_pool);

			} else {
				//-------------------------------------------
				// Simple Skipping Resampling for Color JPEG
				//-------------------------------------------

				unsigned char* inRowBuf = (unsigned char*)_alloca(sizeof(char) * 3 * pixel_width);

				int hop = (skip - 1) * 3;

				for (int i = view_y; i != 0; i--)
					djpeg.proc(inRowBuf);

				for (int y = 0; y < output_height; y++) {
					djpeg.proc(inRowBuf);
					unsigned char* rowPtr = inRowBuf + view_x * 3;

					if (skip > 1) {
						unsigned char* ip = (unsigned char*)inRowBuf;
						unsigned char* op = (unsigned char*)outRowBuf;
						unsigned char* ep = (unsigned char*)outRowBuf + 3 * output_width;
						while (op < ep) {
							*op++ = *ip++;
							*op++ = *ip++;
							*op++ = *ip++;
							ip += hop;
						}
						codec.proc(outRowBuf);
	
						int i = skip;
						while (--i)
							djpeg.proc(inRowBuf);
					} else {
						codec.proc(rowPtr);
					}
				}

				djpeg.flush(inRowBuf);
			}

		} else if (input_codec == eLosslessJPEG) {
			throw expInternalError("8-bit color lossless jpeg decording is not supported", false);
		} else if (input_codec == ePNG) {
			throw expInternalError("8-bit color PNG decording is not supported", false);
		} else {
			//--------------------------
			// Proc Color Raster
			//--------------------------

			if (doResampling) {
				//--------------------------------
				// Do Resampling for Color Raster
				//--------------------------------

				float yi = view_y;
				for (int y = 0; y < output_height; y++, yi += inv_scale) {
					int yi0, yi1, dy0;
					calc_coord(yi, pixel_height, yi0, yi1, dy0);

					unsigned char* buf_ptr2[2];
					buf_ptr2[0] = pixel + 3 * pixel_width * yi0;
					buf_ptr2[1] = pixel + 3 * pixel_width * yi1;

					unsigned char* op = (unsigned char*)outRowBuf;

					float xi = view_x;
					for (int x = 0; x < output_width; x++, xi += inv_scale) {
						int xi0, xi1, dx0;
						calc_coord(xi, pixel_width, xi0, xi1, dx0);

						calc_psf3(yi0, yi1, dy0, xi0, xi1, dx0, buf_ptr2, op);
						op += 3;
					}
					codec.proc(outRowBuf);
				}

			} else {
				//--------------------------------
				// No Resampling for Color Raster
				//--------------------------------

				int hop = (skip - 1) * 3;

				pixel += view_y * pixel_width * 3;

				for (int y = 0; y < output_height; y++) {
					unsigned char* rowPtr = pixel + view_x * 3;

					if (skip > 1) {
						unsigned char* ip = (unsigned char*)rowPtr;
						unsigned char* op = (unsigned char*)outRowBuf;
						unsigned char* ep = (unsigned char*)outRowBuf + 3 * output_width;
						while (op < ep) {
							*op++ = *ip++;
							*op++ = *ip++;
							*op++ = *ip++;
							ip += hop;
						}
						codec.proc(outRowBuf);
					} else {
						codec.proc(rowPtr);
					}

					pixel += 3 * pixel_width * skip;
				}
			}
		}

	}

	if (codec.is_buffered()) {
		//--------------------------------------
		// Set HTTP Response for buffered codec
		//--------------------------------------
		//
		// Build Response Header
		//
		const char* cache_flag;
		int cache_f = 0;
		if (prm[prmCacheControl].hasValue)
			cache_f = prm[prmCacheControl].atoi();
		if (cache_f == 0) {
			cache_flag = "";
		} else if (cache_f == 1) {
			cache_flag = "Cache-Control: no-store\r\n";
		} else if (cache_f == 2) {
			cache_flag = "Cache-Control: no-cache\r\n";
		}

		char etag_flag[ETAG_VALUE_LIMIT + 16];
		if (*etag && cache_f == 0) {
			strcpy(etag_flag, "ETag: \"");
			strcat(etag_flag, etag);
			strcat(etag_flag, "\"\r\n");
		} else {
			etag_flag[0] = '\0';
		}

		const size_t content_length = codec.finish();

		sprintf(httpHeader, "Content-Type: %s\r\nContent-Length: %d\r\n%s%s\r\n", codec.content_type(), content_length, cache_flag, etag_flag);
		len = strlen(httpHeader);

		//
		// Send Response Header
		//
		if (gDWConfig->iis.version >= 400) {
			HSE_SEND_HEADER_EX_INFO header_ex_info;
			static const char* const ok = "200 OK";
			header_ex_info.pszStatus = ok;
			header_ex_info.pszHeader = httpHeader;
			header_ex_info.cchStatus = strlen(ok);
			header_ex_info.cchHeader = len;
			header_ex_info.fKeepConn = TRUE;
			if (!pECB->ServerSupportFunction(pECB->ConnID, HSE_REQ_SEND_RESPONSE_HEADER_EX, &header_ex_info, NULL, NULL)) {
				sprintf((char*)pECB->lpszLogData, " ServerSupportFunction(HSE_REQ_SEND_RESPONSE_HEADER_EX) Error as %d", GetLastError());
				InterlockedIncrement(&total_error_count);
				pECB->dwHttpStatusCode = HSE_STATUS_ERROR;
				return HSE_STATUS_ERROR;
			}
		} else {
			if (!pECB->ServerSupportFunction(pECB->ConnID, HSE_REQ_SEND_RESPONSE_HEADER, (LPVOID)"200 OK", &len, (LPDWORD)httpHeader)) {
				sprintf((char*)pECB->lpszLogData, " ServerSupportFunction(HSE_REQ_SEND_RESPONSE_HEADER) Error as %d", GetLastError());
				InterlockedIncrement(&total_error_count);
				pECB->dwHttpStatusCode = HSE_STATUS_ERROR;
				return HSE_STATUS_ERROR;
			}
		}
		//
		// Send Response Body
		//
		codec.flush();
	}


	//
	// log
	//
#ifdef _DIO_IMAGE_LOG_
//	char logKey[256];
//	if (GetSingleQueryValue(pECB->lpszQueryString, "LOG", logKey, sizeof(logKey)))
//		logging(logKey, pECB);
#endif
	InterlockedIncrement(&total_response_count);
	pECB->dwHttpStatusCode = SuccessCode;
	return SuccessCode;
	
} catch (expCODEC& e) {
	const char* header = "Content-Length: 0\r\n\r\n";
	DWORD len = strlen(header);
	pECB->ServerSupportFunction(pECB->ConnID, HSE_REQ_SEND_RESPONSE_HEADER, (LPVOID)"500 Internal Error", &len, (LPDWORD)header);
	strcpy((char*)pECB->lpszLogData, " expCODEC(");
	strcat((char*)pECB->lpszLogData, e.msg);
	strcat((char*)pECB->lpszLogData, ")");
	err_log(pECB->lpszPathInfo, "?", pECB->lpszQueryString, (char*)pECB->lpszLogData);
} catch (expInternalError& e) {
	const char* header = "Content-Length: 0\r\n\r\n";
	DWORD len = strlen(header);
	pECB->ServerSupportFunction(pECB->ConnID, HSE_REQ_SEND_RESPONSE_HEADER, (LPVOID)"500 Internal Error", &len, (LPDWORD)header);
	strcpy((char*)pECB->lpszLogData, " expInternalError(");
	strcat((char*)pECB->lpszLogData, e.msg);
	strcat((char*)pECB->lpszLogData, ")");
	if (e.write_err_log)
		err_log(pECB->lpszPathInfo, "?", pECB->lpszQueryString, (char*)pECB->lpszLogData);
} catch (...) {
	const char* header = "Content-Length: 0\r\n\r\n";
	DWORD len = strlen(header);
	pECB->ServerSupportFunction(pECB->ConnID, HSE_REQ_SEND_RESPONSE_HEADER, (LPVOID)"500 Internal Error", &len, (LPDWORD)header);
	err_log(pECB->lpszPathInfo, "?", pECB->lpszQueryString, "unexpected excpetion");
}

	InterlockedIncrement(&total_error_count);
	pECB->dwHttpStatusCode = HSE_STATUS_ERROR;
	return HSE_STATUS_ERROR;
}


#if 0
//========================================================================================
// Main
//========================================================================================
BOOL WINAPI DllMain( HINSTANCE hInst, DWORD wDataSeg, LPVOID lpReserved );
BOOL WINAPI DllMain( HINSTANCE hInst, DWORD wDataSeg, LPVOID lpReserved )
{
	switch (wDataSeg) {
    case DLL_PROCESS_ATTACH:
	case DLL_PROCESS_DETACH:
	case DLL_THREAD_ATTACH:
	case DLL_THREAD_DETACH:
		return true;
   	}

   	return false;
}

#endif
