/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "UtilClasses.h"

#include "etag-cache.h"

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
struct tETagEntry {
	tETagEntry* hash_next;
	tETagEntry* hash_prev;
	tETagEntry* free_next;
	unsigned long sum;
	char etag[ETAG_VALUE_LIMIT];
};
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
#define kNumETagEntryies (4 * 1024 * 1024 / sizeof(tETagEntry))
#define kHashSize 1024
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
struct tETagCacheTable {
	tETagEntry table[kNumETagEntryies];
	tETagEntry* hash[kHashSize];
	unsigned long crc_table[256];
	tETagEntry* free;
	tETagEntry* used_head;
	tETagEntry* used_tail;
	CRITICAL_SECTION crit;

	//------------------------------------------------------------------------------------
	//
	//------------------------------------------------------------------------------------
	void buildCRCTable()
	{
		const unsigned long CRC32_POLYNOMIAL = 0xEDB88320;

		for (unsigned long i = 0; i < 256; i++) {
			unsigned long crc = i;
			for (unsigned long j = 8; j > 0; j--) {
				if ((crc & 1) == 1) {
					crc = (crc >> 1) ^ CRC32_POLYNOMIAL;
				} else {
					crc >>= 1;
				}
			}
			crc_table[i] = crc;
		}
	}

	//------------------------------------------------------------------------------------
	//
	//------------------------------------------------------------------------------------
	void init()
	{
		free = &table[0];
		for (int i = 0; i < kNumETagEntryies - 1; i++)
			table[i].free_next = &table[i + 1];
		table[kNumETagEntryies - 1].free_next = NULL;

		buildCRCTable();

		InitializeCriticalSectionAndSpinCount(&crit, 0x80000000 + 4096);
	}

	//------------------------------------------------------------------------------------
	//
	//------------------------------------------------------------------------------------
	void finish()
	{
		DeleteCriticalSection(&crit);
	}

	//------------------------------------------------------------------------------------
	//
	//------------------------------------------------------------------------------------
	tETagEntry* get_free_entry(const char* etag, unsigned long sum)
	{
		tETagEntry* et = free;
		if (et != NULL) {
			free = et->free_next;
			if (used_tail == NULL) {
				used_head = used_tail = et;
			} else {
				used_tail->free_next = et;
				used_tail = et;
			}	
		} else {
			et = used_head;
			used_head = et->free_next;
			used_tail->free_next = et;
			used_tail = et;

			tETagEntry* prev = et->hash_prev;
			tETagEntry* next = et->hash_next;

			if (prev != NULL)
				prev->hash_next = next;
			else
				hash[calc_hash(et->sum)] = next;

			if (next != NULL)
				next->hash_prev = prev;
		}

		et->hash_next = NULL;
		et->hash_prev = NULL;
		et->free_next = NULL;
		et->sum = sum;
		strcpy(et->etag, etag);

		return et;
	}

	//------------------------------------------------------------------------------------
	//
	//------------------------------------------------------------------------------------
	unsigned long calc_sum(const char* str)
	{
		unsigned long crc = 0;

		for (unsigned char* cp = (unsigned char*)str; *cp != '\0'; cp++) {
			unsigned long temp1 = crc >> 8;
			unsigned long temp2 = crc_table[(crc ^ *cp) & 0xFF];
			crc = temp1 ^ temp2;
		}

		return crc;
	}

	//------------------------------------------------------------------------------------
	//
	//------------------------------------------------------------------------------------
	static unsigned long calc_hash(unsigned long sum)
	{
		return sum & 0x3FF;
	}

	//------------------------------------------------------------------------------------
	//
	//------------------------------------------------------------------------------------
	bool check_and_register_entry(const char* etag)
	{
		const unsigned long sum = calc_sum(etag);
		const unsigned long hash_value = calc_hash(sum);

		CStackCriticalSection cs(&crit);

		tETagEntry* et = hash[hash_value];
		if (et == NULL) {
			et = get_free_entry(etag, sum);
			hash[hash_value] = et;
			return false;
		} else {
			tETagEntry* last = NULL;
			while (et != NULL) {
				if (et->sum == sum && strcmp(et->etag, etag) == 0) {
					if (last != NULL) {
						// move hit entry to head of hash bucket list
						last->hash_next = et->hash_next;
						if (et->hash_next != NULL)
							et->hash_next->hash_prev = last;
						et->hash_next = hash[hash_value];
						hash[hash_value]->hash_prev = et;
						hash[hash_value] = et;
					}
					return true;
				}
				last = et;
				et = et->hash_next;
			}
			et = get_free_entry(etag, sum);
			if (hash[hash_value] != NULL) { // this check is prepararation for in case of acidently get_free_entry() delete first hash entry
				et->hash_next = hash[hash_value];
				hash[hash_value]->hash_prev = et;
			}
			hash[hash_value] = et;
			return false;
		}
	}
	
};

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static tETagCacheTable* ETagCacheTable;
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
void InitETag(HANDLE hHeap)
{
	ETagCacheTable = (tETagCacheTable*)HeapAlloc(hHeap, 0, sizeof(tETagCacheTable));
	memset(ETagCacheTable, 0, sizeof(tETagCacheTable));

	ETagCacheTable->init();
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
void FinishETag(HANDLE hHeap)
{
	ETagCacheTable->finish();

	HeapFree(hHeap, 0, ETagCacheTable);
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
bool CheckETagEntry(const char* etag)
{
	return ETagCacheTable->check_and_register_entry(etag);
}
