/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "UtilClasses.h"
#include "dw-config.h"
#include "cstudydb.h"

#include "dio-storage.h"

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static bool consoleMode;
static HANDLE stopEventHandle = NULL;

tDWConfig* gDWConfig;

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------

// internal variables
static SERVICE_STATUS ssStatus = { 0, 0, 0, 0, 0, 0, 0 };       // current status of the service
static SERVICE_STATUS_HANDLE sshStatusHandle;
static DWORD dwErr = 0;
static char szErr[256];

//
//  FUNCTION: GetLastErrorText
//
//  PURPOSE: copies error message text to string
//
//  PARAMETERS:
//    lpszBuf - destination buffer
//    dwSize - size of buffer
//
//  RETURN VALUE:
//    destination buffer
//
//  COMMENTS:
//
static char* GetLastErrorText(char* lpszBuf, DWORD dwSize)
{
    DWORD dwRet;
    wchar_t* lpszTemp = NULL;

    dwRet = FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM |FORMAT_MESSAGE_ARGUMENT_ARRAY,
                           NULL,
                           GetLastError(),
                           LANG_NEUTRAL,
                           (LPTSTR)&lpszTemp,
                           0,
                           NULL );
	TOAC(lpszTemp)
	LocalFree(lpszTemp);
	
    // supplied buffer is not long enough
    if ( !dwRet || ( (long)dwSize < (long)dwRet+14 ) )
        lpszBuf[0] = '\0';
    else
    {
        *(const_cast<char*>(_lpszTemp_AC) + _lpszTemp_AC_len - 2) = '\0';  //remove cr and newline character
        sprintf( lpszBuf, "%s (0x%x)", _lpszTemp_AC, GetLastError() );
    }

    return lpszBuf;
}

//
//  FUNCTION: AddToMessageLog(LPTSTR lpszMsg)
//
//  PURPOSE: Allows any thread to log an error message
//
//  PARAMETERS:
//    lpszMsg - text for message
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//
static VOID AddToMessageLog(const char* lpszMsg)
{
    char   szMsg[MAX_PATH];
    HANDLE  hEventSource;
    const wchar_t*  lpszStrings[2];


    if (!consoleMode)
    {
        dwErr = GetLastError();

        // Use event logging to log the error.
        //
        hEventSource = RegisterEventSource(NULL, TEXT(DIO_STORAGE_SERVICENAME));

        sprintf(szMsg, "%s error: %d", DIO_STORAGE_SERVICENAME, dwErr);
        TOWC(szMsg)
        TOWC(lpszMsg)
        
        lpszStrings[0] = _szMsg_WC;
        lpszStrings[1] = _lpszMsg_WC;

        if (hEventSource != NULL) {
            ReportEvent(hEventSource, // handle of event source
                EVENTLOG_ERROR_TYPE,  // event type
                0,                    // event category
                0,                    // event ID
                NULL,                 // current user's SID
                2,                    // strings in lpszStrings
                0,                    // no bytes of raw data
                lpszStrings,          // array of error strings
                NULL);                // no raw data

            (VOID) DeregisterEventSource(hEventSource);
        }
    }
}


//
//  FUNCTION: ReportStatusToSCMgr()
//
//  PURPOSE: Sets the current status of the service and
//           reports it to the Service Control Manager
//
//  PARAMETERS:
//    dwCurrentState - the state of the service
//    dwWin32ExitCode - error code to report
//    dwWaitHint - worst case estimate to next checkpoint
//
//  RETURN VALUE:
//    TRUE  - success
//    FALSE - failure
//
//  COMMENTS:
//
static BOOL ReportStatusToSCMgr(DWORD dwCurrentState,
                         DWORD dwWin32ExitCode,
                         DWORD dwWaitHint,
                         DWORD dwServiceSpecificExitCode = 0)
{
    static DWORD dwCheckPoint = 1;
    BOOL fResult = TRUE;


    if ( !consoleMode ) // when debugging we don't report to the SCM
    {
        if (dwCurrentState == SERVICE_START_PENDING)
            ssStatus.dwControlsAccepted = 0;
        else
            ssStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP;

        ssStatus.dwCurrentState = dwCurrentState;
        ssStatus.dwWin32ExitCode = dwWin32ExitCode;
        ssStatus.dwWaitHint = dwWaitHint;

        if ( ( dwCurrentState == SERVICE_RUNNING ) ||
             ( dwCurrentState == SERVICE_STOPPED ) )
            ssStatus.dwCheckPoint = 0;
        else
            ssStatus.dwCheckPoint = dwCheckPoint++;

		if (dwServiceSpecificExitCode != 0)
			ssStatus.dwServiceSpecificExitCode = dwServiceSpecificExitCode;

        // Report the status of the service to the service control manager.
        //
        if (!(fResult = SetServiceStatus( sshStatusHandle, &ssStatus))) {
            AddToMessageLog("SetServiceStatus");
        }
    }
    return fResult;
}







///////////////////////////////////////////////////////////////////
//
//  The following code handles service installation and removal
//


//
//  FUNCTION: CmdInstallService()
//
//  PURPOSE: Installs the service
//
//  PARAMETERS:
//    none
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//
static void installService(const char* cf)
{
	if (!RegisterConfigFilePath(cf))
		return;

	wchar_t fullPath[MAX_PATH];
	if (GetModuleFileName( NULL, fullPath, sizeof(fullPath) ) == 0) {
		printf("Unable to install %s - %s\n", DIO_STORAGE_SERVICENAME, GetLastErrorText(szErr, 256));
		return;
	}
	
	SC_HANDLE schSCManager = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
	if (schSCManager) {
		SC_HANDLE schService = CreateService(schSCManager, // SCManager database
		                                     TEXT(DIO_STORAGE_SERVICENAME), // name of service
		   	                                 TEXT(DIO_STORAGE_SERVICDISPLAYENAME), // name to display
		                                     SERVICE_ALL_ACCESS, // desired access
		                                     SERVICE_WIN32_OWN_PROCESS, // service type
		                                     SERVICE_AUTO_START, // start type
		                                     SERVICE_ERROR_NORMAL, // error control type
		                                     fullPath, // service's binary
		                                     NULL, // no load ordering group
		                                     NULL, // no tag identifier
		                                     TEXT(""), // dependencies
		                                     NULL, // LocalSystem account
		                                     NULL); // no password
	
		if (schService) {
			printf("%s installed.\n", DIO_STORAGE_SERVICENAME);
			CloseServiceHandle(schService);
		} else {
		printf("CreateService failed - %s\n", GetLastErrorText(szErr, 256));
		}

		CloseServiceHandle(schSCManager);
	} else {
		printf("OpenSCManager failed - %s\n", GetLastErrorText(szErr,256));
	}
}


//
//  FUNCTION: stopService()
//
//  PURPOSE: Stops and removes the service
//
//  PARAMETERS:
//    none
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//
static void stopService(bool remove)
{
	SC_HANDLE schSCManager = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
	if (schSCManager) {
		SC_HANDLE schService = OpenService(schSCManager, TEXT(DIO_STORAGE_SERVICENAME), SERVICE_ALL_ACCESS);
		if (schService) {
			// try to stop the service
			QueryServiceStatus(schService, &ssStatus);
			if (ssStatus.dwCurrentState == SERVICE_RUNNING) {
				if (ControlService(schService, SERVICE_CONTROL_STOP, &ssStatus)) {
					printf("Stopping %s.", DIO_STORAGE_SERVICENAME);
					Sleep(ssStatus.dwWaitHint);
					while(QueryServiceStatus(schService, &ssStatus)) {
						if (ssStatus.dwCurrentState == SERVICE_STOP_PENDING) {
							printf(".");
							Sleep(ssStatus.dwWaitHint);
						} else {
							break;
						}
					}
					if (ssStatus.dwCurrentState == SERVICE_STOPPED) {
						printf("\n%s stopped.\n", DIO_STORAGE_SERVICENAME);
					} else {
						printf("\n%s failed to stop. (%d)\n", DIO_STORAGE_SERVICENAME, ssStatus.dwCurrentState);
					}
				} else {
					printf("ControlService() Failed for %s - %s\n", DIO_STORAGE_SERVICENAME, GetLastErrorText(szErr, 256));
				}
			} else {
				printf("%s not running.\n", DIO_STORAGE_SERVICENAME);
			}

			if (remove) {
				// now remove the service
				if (DeleteService(schService)) {
					printf("%s removed.\n", DIO_STORAGE_SERVICENAME);
				} else {
					printf("DeleteService() failed - %s\n", GetLastErrorText(szErr,256));
				}
			}

			CloseServiceHandle(schService);

		} else {
			printf("OpenService() failed - %s\n", GetLastErrorText(szErr,256));
		}

		CloseServiceHandle(schSCManager);

	} else {
		printf("OpenSCManager() failed - %s\n", GetLastErrorText(szErr,256));
	}

	if (remove)
		RemoveRegistry();
}

//
//  FUNCTION: startService()
//
//  PURPOSE: Stops and removes the service
//
//  PARAMETERS:
//    none
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//
static void startService()
{
	SC_HANDLE schSCManager = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
	if (schSCManager) {
		SC_HANDLE schService = OpenService(schSCManager, TEXT(DIO_STORAGE_SERVICENAME), SERVICE_ALL_ACCESS);
		if (schService) {
			// try to start the service
			if (StartService(schService, 0, NULL)) {
				printf("Starting %s.", DIO_STORAGE_SERVICENAME);
				QueryServiceStatus(schService, &ssStatus);
				DWORD cp;
				while (ssStatus.dwCurrentState != SERVICE_RUNNING) {
					cp = ssStatus.dwCheckPoint;
					printf(".");
					Sleep(ssStatus.dwWaitHint);
					if (!QueryServiceStatus(schService, &ssStatus))
						break;
				}
					if (ssStatus.dwCurrentState == SERVICE_RUNNING) {
					printf("\n%s started.\n", DIO_STORAGE_SERVICENAME);
				} else {
					printf("\n%s failed to started. (%d)\n", DIO_STORAGE_SERVICENAME, ssStatus.dwCurrentState);
				}
			} else {
				printf("StartService() Failed for %s - %s\n", DIO_STORAGE_SERVICENAME, GetLastErrorText(szErr, 256));
			}
			CloseServiceHandle(schService);

		} else {
			printf("OpenService() failed - %s\n", GetLastErrorText(szErr,256));
		}

		CloseServiceHandle(schSCManager);

	} else {
		printf("OpenSCManager() failed - %s\n", GetLastErrorText(szErr,256));
	}
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static int cmp_tag(const void* va, const void* vb)
{
	const DCM_TAG* ta = (DCM_TAG*)va;
	const DCM_TAG* tb = (DCM_TAG*)vb;

	unsigned short a = DCM_TAG_GROUP(*ta);
	unsigned short b = DCM_TAG_GROUP(*tb);
	if (a != b)
		return a - b;

	return DCM_TAG_ELEMENT(*ta) - DCM_TAG_ELEMENT(*tb);
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static HRESULT make_allowable_security_attributes(SECURITY_ATTRIBUTES** sapp)
{
	HANDLE hToken;
	if (OpenProcessToken(GetCurrentProcess(), TOKEN_QUERY, &hToken) == 0)
		return HRESULT_FROM_WIN32(GetLastError());
	AutoCloseHandle ac_hToken(hToken);

	DWORD daclBufLen = 0;
	GetTokenInformation(hToken, TokenDefaultDacl, NULL, 0, &daclBufLen);
	if (daclBufLen == 0)
		return HRESULT_FROM_WIN32(GetLastError());
	PTOKEN_DEFAULT_DACL dacl = (PTOKEN_DEFAULT_DACL)_alloca(daclBufLen);
	if (GetTokenInformation(hToken, TokenDefaultDacl, dacl, daclBufLen, &daclBufLen) == 0)
		return HRESULT_FROM_WIN32(GetLastError());

	ACL_SIZE_INFORMATION acl_size;
	if (GetAclInformation(dacl->DefaultDacl, &acl_size, sizeof(acl_size), AclSizeInformation) == 0)
		return HRESULT_FROM_WIN32(GetLastError());

	DWORD newAclSize = sizeof(ACL);

	for (int i = 0; i < acl_size.AceCount; i++) {
		ACE_HEADER* ace;
		GetAce(dacl->DefaultDacl, i, (LPVOID*)&ace);
		newAclSize += ace->AceSize;
	}

	SID sid = {SID_REVISION, 1, SECURITY_WORLD_SID_AUTHORITY, DOMAIN_ALIAS_RID_GUESTS};
	newAclSize += sizeof(ACCESS_ALLOWED_ACE) + GetLengthSid(&sid);

	PACL newAcl = (PACL)_alloca(newAclSize);
	if (InitializeAcl(newAcl, newAclSize, ACL_REVISION) == 0)
		return HRESULT_FROM_WIN32(GetLastError());

	for (int i = 0; i < acl_size.AceCount; i++) {
		ACE_HEADER* ace;
		GetAce(dacl->DefaultDacl, i, (LPVOID*)&ace);
		if (AddAce(newAcl, ACL_REVISION, i, ace, ace->AceSize) == 0)
			return HRESULT_FROM_WIN32(GetLastError());
	}

	if (AddAccessAllowedAce(newAcl, ACL_REVISION, GENERIC_READ, &sid) == 0)
		return HRESULT_FROM_WIN32(GetLastError());

	static SECURITY_DESCRIPTOR sd;
	InitializeSecurityDescriptor(&sd, SECURITY_DESCRIPTOR_REVISION);
	if (SetSecurityDescriptorDacl(&sd, TRUE, newAcl, FALSE) == 0)
		return HRESULT_FROM_WIN32(GetLastError());

	static SECURITY_ATTRIBUTES sa = { sizeof(sa), NULL, FALSE };
	*sapp = &sa;

	return S_OK;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static HRESULT ServiceStart(const char* cf)
{
	HRESULT hr = CoInitializeEx(NULL, COINIT_MULTITHREADED);
	if (FAILED(hr))
		return hr;
	AutoCoUninitialize acu;

	SECURITY_ATTRIBUTES* sa = NULL;
//	hr = make_allowable_security_attributes(&sa);
//	if (FAILED(hr))
//		return hr;

	StorageLog slog;
	slog.Init(false && consoleMode, L"C:\\DIOWave\\logs\\dio-storage", sa);
	StorageLog::self = &slog;

	//
	//
	hr = ReadConfig(gDWConfig, cf, consoleMode);
	if (FAILED(hr))
		return hr;

	//
	//
	//
	CStudyDB::Init(gDWConfig);
	InitVolumeMangement();
	InitRTLog();

	//
	//
	//
	Init_CTN_HEAP();

	//
	// Initialize to prepare COND facilitiy as Multi Theaded Aware
	//
	COND_Initialize();

	//
	//
	//
	WSADATA wsaData;
	if (WSAStartup(WINSOCK_VERSION, &wsaData))
		return HRESULT_FROM_WIN32(GetLastError());

	DUL_Initialize();

	//
	//
	//
	stopEventHandle = CreateEvent(NULL, TRUE, FALSE,  NULL);	// Manual Reset Event
	if (stopEventHandle == NULL)
		return HRESULT_FROM_WIN32(GetLastError());
	AutoCloseHandle ac_stopEventHandle(stopEventHandle);

	//
	//
	//
	tStorageParam* storageParams = (tStorageParam*)_alloca(sizeof(tStorageParam) * gDWConfig->num_scp);
	ZeroMemory(storageParams, sizeof(tStorageParam) * gDWConfig->num_scp);

	tStorageParam* sp = storageParams;
	for (int scp_index = 0; scp_index < gDWConfig->num_scp; scp_index++, sp++) {
		const tSCPprm* scp = gDWConfig->scp_tbl + scp_index;

		sp->pConfig = gDWConfig;
		sp->pSCPprm = scp;

		sp->wsp.dwAcceptTimeout = INFINITE;
		sp->wsp.dwReadSocketTimeout = scp->read_timeout;
		
		sp->wsp.hAcceptEvent = CreateEvent(NULL, FALSE, FALSE, NULL); // Auto Reset Event
		sp->wsp.hCloseEvent = CreateEvent(NULL, TRUE, TRUE, NULL);  // Manual Reset Event, Signaled
		sp->wsp.hStopEvent = stopEventHandle;
		sp->hThread = (HANDLE)_beginthreadex((void*)NULL, (unsigned)0, (unsigned (__stdcall *)(void *))StorageProviderThread, (void*)sp, (unsigned)0, (unsigned int*)&sp->dwThreadId);

		if (sp->hThread == NULL || sp->wsp.hAcceptEvent == NULL || sp->wsp.hCloseEvent == NULL) {
			if (sp->hThread != NULL) {
				TerminateThread(sp->hThread, 0);
				CloseHandle(sp->hThread);
			}
			if (sp->wsp.hAcceptEvent != NULL)
				CloseHandle(sp->wsp.hAcceptEvent);
			if (sp->wsp.hCloseEvent != NULL)
				CloseHandle(sp->wsp.hCloseEvent);

			tStorageParam* sp1 = storageParams;
			while (sp != sp1) {
				TerminateThread(sp1->hThread, 0);
				CloseHandle(sp1->hThread);
				CloseHandle(sp1->wsp.hAcceptEvent);
				CloseHandle(sp1->wsp.hCloseEvent);
			}

			return 1;
		}
	}

	ReportStatusToSCMgr(SERVICE_RUNNING, NO_ERROR, 1000);
	//---------------------------------------------
	//
	WaitForSingleObject(stopEventHandle, INFINITE);
	//
	//---------------------------------------------
	ReportStatusToSCMgr(SERVICE_STOP_PENDING, NO_ERROR, 1000);

	//
	//
	//
	sp = storageParams;
	for (int scp_index = 0; scp_index < gDWConfig->num_scp; scp_index++, sp++) {
		if (WAIT_OBJECT_0 != WaitForSingleObject(sp->hThread, 30*1000))
			TerminateThread(sp->hThread, 0);
		CloseHandle(sp->hThread);
		CloseHandle(sp->wsp.hAcceptEvent);
		CloseHandle(sp->wsp.hCloseEvent);

		ReportStatusToSCMgr(SERVICE_STOP_PENDING, NO_ERROR, 1000);
	}

	//
	//
	//
	WSACleanup();

	//
	//
	//
	FinRTLog();
	FinVolumeManagement();

	return S_OK;
}


//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void ServiceStop(void)
{
    if (stopEventHandle != NULL)
        SetEvent(stopEventHandle);
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static VOID WINAPI service_ctrl(DWORD dwCtrlCode)
{
	//
	// Handle the requested control code.
	//
	switch(dwCtrlCode) {

	case SERVICE_CONTROL_STOP:
		//
		// Stop the service.
		//
		// SERVICE_STOP_PENDING should be reported before
		// setting the Stop Event - hServerStopEvent - in
		// ServiceStop().  This avoids a race condition
		// which may result in a 1053 - The Service did not respond...
		// error.
		ReportStatusToSCMgr(SERVICE_STOP_PENDING, NO_ERROR, 0);
		ServiceStop();
		return;
		
	case SERVICE_CONTROL_INTERROGATE:
		//
		// Update the service status.
		//
		break;
		
	default:
		//
		// invalid control code
		//
		break;
	}
	
	ReportStatusToSCMgr(ssStatus.dwCurrentState, NO_ERROR, 0);
}


//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void WINAPI service_main(DWORD dwArgc, LPTSTR *lpszArgv)
{
	//
	// register our service control handler:
	//
	sshStatusHandle = RegisterServiceCtrlHandler(TEXT(DIO_STORAGE_SERVICENAME), service_ctrl);
	if (sshStatusHandle == NULL)
		return ;

	//
	// SERVICE_STATUS members that don't change in example
	//
	ssStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
	ssStatus.dwServiceSpecificExitCode = 0;

	if (!ReportStatusToSCMgr(SERVICE_START_PENDING, NO_ERROR, 3000)) {
		return;
	}

	const HRESULT hr = ServiceStart(NULL);

	ReportStatusToSCMgr(SERVICE_STOPPED, hr == S_OK ? NO_ERROR : ERROR_SERVICE_SPECIFIC_ERROR, 0, hr);
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static BOOL WINAPI ControlHandler (DWORD dwCtrlType)
{
	switch (dwCtrlType) {
	case CTRL_BREAK_EVENT:  // use Ctrl+C or Ctrl+Break to simulate
	case CTRL_C_EVENT:      // SERVICE_CONTROL_STOP in debug mode
		printf("Stopping %s.\n", DIO_STORAGE_SERVICENAME);
		ServiceStop();
		return TRUE;
	
	}
	return FALSE;
}

//
//  FUNCTION: runService()
//
//  PURPOSE: Runs the service as a console application
//
//  PARAMETERS:
//    none
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//
static void runService(const char* cf)
{
	SetConsoleCtrlHandler(ControlHandler, TRUE);
	ServiceStart(cf);
}


//
//  FUNCTION: main
//
//  PURPOSE: entrypoint for service
//
//  PARAMETERS:
//    argc - number of command line arguments
//    argv - array of command line arguments
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//    main() either performs the command line task, or
//    call StartServiceCtrlDispatcher to register the
//    main service thread.  When the this call returns,
//    the service has stopped, so exit.
//
int main(int argc, const char** argv)
{
	if (argc == 1) {
		consoleMode = false;
		//
		// Service Main, SCP Dispatcher
		//
		SERVICE_TABLE_ENTRY dispatchTable[] = {
			{ TEXT(DIO_STORAGE_SERVICENAME), (LPSERVICE_MAIN_FUNCTION)service_main },
			{ NULL, NULL }
		};
		if (!StartServiceCtrlDispatcher(dispatchTable))
			AddToMessageLog("StartServiceCtrlDispatcher failed.");
	} else {
		consoleMode = true;

		const char* arg = NULL;
		if (argc > 2)
			arg = argv[2];

		if (_stricmp("-install", argv[1]) == 0) {
			installService(arg);
		} else if (_stricmp("-start", argv[1]) == 0) {
			startService();
		} else if (_stricmp("-stop", argv[1]) == 0) {
			stopService(false);
		} else if (_stricmp("-remove", argv[1]) == 0) {
			stopService(true);
		} else if (_stricmp("-config", argv[1]) == 0) {
			RegisterConfigFilePath(arg);
		} else if (_stricmp("-run", argv[1]) == 0) {
			runService(arg);
		} else {
			printf("Options\n");
			printf(" -install [<Config File>] to install the service\n");
			printf(" -remove                  to remove the service\n");
			printf(" -start                   to start the service after <delay> seconds.\n");
			printf(" -stop                    to stop the service\n");
			printf(" -config [<Config File>]  to re-configure service\n");
			printf(" -run [<Config File>]     to run as a console application\n");
		}
	}

	return 0;
}


