/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "UtilClasses.h"
#include "dw-config.h"
#include "dcmdict.h"
#include "CDcmScanner.h"
#include "DataElementCache.h"
#include "InstanceCache.h"

#include "dio-study.dll.h"

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static void scan_files(const wchar_t* studyDir, const wchar_t* location, CDataElementCache& dc, CInstanceCache& ic) throw (HRESULT)
{
	CStackFindFile find(studyDir);
	if (find.handle == INVALID_HANDLE_VALUE)
		throw static_cast<HRESULT>(HRESULT_FROM_WIN32(GetLastError()));

	const int studyDirLen = lstrlen(studyDir);
	wchar_t* const fullPath = (wchar_t*)_alloca(sizeof(wchar_t) * (studyDirLen + 1 + MAX_PATH + 1));
	CopyMemory(fullPath, studyDir, sizeof(wchar_t) * studyDirLen);
	*(fullPath + studyDirLen) = L'\\';
	wchar_t* const dir = fullPath + studyDirLen + 1;

	const int locationLen = lstrlen(location);
	wchar_t* const fullLoc = (wchar_t*)_alloca(sizeof(wchar_t) * ((locationLen != 0 ? locationLen + 1 : 0) +  MAX_PATH + 1));
	if (locationLen != 0) {
		CopyMemory(fullLoc, location, sizeof(wchar_t) * locationLen);
		*(fullLoc + locationLen) = L'/';
	}
	wchar_t* const loc = fullLoc + (locationLen != 0 ? locationLen + 1 : 0);

	do {
		if (*find.data.cFileName == L'.')
			continue;

		lstrcpy(dir, find.data.cFileName);
		lstrcpy(loc, find.data.cFileName);

		if (find.data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
			scan_files(fullPath, fullLoc, dc, ic);
		} else {
			HRESULT hr = ic.AppendInstance(fullPath, fullLoc, dc);
			if (FAILED(hr))
				throw hr;
		}
	} while (find.next());
}

//======================================================================
//
//======================================================================
HRESULT CreateStudyCache(const wchar_t* studyDir)
{
	CDataElementCache dc;
	CInstanceCache ic;

	try {
		scan_files(studyDir, L"", dc, ic);
	} catch (HRESULT hr) {
		return hr;
	}

	const int studyDirLen = lstrlen(studyDir);

	wchar_t* const dcFile = (wchar_t*)_alloca(sizeof(wchar_t) * (studyDirLen + 1 + lstrlen(CDataElementCache::filename) + 1));
	lstrcpy(dcFile, studyDir);
	lstrcat(dcFile, L"\\");
	lstrcat(dcFile, CDataElementCache::filename);

	HRESULT hr = dc.Dump(dcFile);
	if (FAILED(hr))
		return hr;

	wchar_t* const icFile = (wchar_t*)_alloca(sizeof(wchar_t) * (studyDirLen + 1 + lstrlen(CInstanceCache::filename) + 1));
	lstrcpy(icFile, studyDir);
	lstrcat(icFile, L"\\");
	lstrcat(icFile, CInstanceCache::filename);

	hr = ic.Dump(icFile);
	if (FAILED(hr))
		return hr;

	return S_OK;
}


//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
class CStudyCache {
public:
	CDataElementCacheFile m_dc;
	CInstanceCacheFile m_ic;
};

//======================================================================
//
//======================================================================
HRESULT OpenStudyCache(const wchar_t* studyDir, int* ctx)
{
	*ctx = 0;

	CStudyCache* sc = new CStudyCache;

	const int studyDirLen = lstrlen(studyDir);

	wchar_t* const dcFile = (wchar_t*)_alloca(sizeof(wchar_t) * (studyDirLen + 1 + lstrlen(CDataElementCache::filename) + 1));
	lstrcpy(dcFile, studyDir);
	lstrcat(dcFile, L"\\");
	lstrcat(dcFile, CDataElementCache::filename);

	HRESULT hr = sc->m_dc.Open(dcFile);
	if (FAILED(hr)) {
		delete sc;
		return hr;
	}

	wchar_t* const icFile = (wchar_t*)_alloca(sizeof(wchar_t) * (studyDirLen + 1 + lstrlen(CInstanceCache::filename) + 1));
	lstrcpy(icFile, studyDir);
	lstrcat(icFile, L"\\");
	lstrcat(icFile, CInstanceCache::filename);

	hr = sc->m_ic.Open(icFile);
	if (FAILED(hr)) {
		delete sc;
		return hr;
	}

	*ctx = (int)sc;
	return S_OK;
}
//======================================================================
//
//======================================================================
HRESULT AppendStudyCache(int ctx, const wchar_t* filePath, const wchar_t* location)
{
	CStudyCache* sc = (CStudyCache*)ctx;

	return sc->m_ic.AppendInstance(filePath, location, sc->m_dc);
}
//======================================================================
//
//======================================================================
HRESULT CommitStudyCache(int ctx)
{
	CStudyCache* sc = (CStudyCache*)ctx;

	HRESULT hr = sc->m_dc.Commit();
	if (FAILED(hr))
		return hr;

	hr = sc->m_ic.Commit();
	if (FAILED(hr))
		return hr;
	
	return S_OK;
}

//======================================================================
//
//======================================================================
HRESULT CloseStudyCache(int ctx)
{
	CStudyCache* sc = (CStudyCache*)ctx;

	delete sc;

	return S_OK;
}

//======================================================================
//
//======================================================================
HRESULT GetStudyCache(DWORD dwOption, const wchar_t* studyDir, IUnknown** ppUnk)
{
	CLSID clsidDomDoc;
	HRESULT hr = CLSIDFromProgID(L"Msxml2.FreeThreadedDOMDocument", &clsidDomDoc);
	if (FAILED(hr))
		return hr;

	AutoComPtr<IXMLDOMDocument> dom;
	hr = CoCreateInstance(clsidDomDoc, NULL, CLSCTX_INPROC_SERVER, IID_IXMLDOMDocument, (void**)dom.r());
	if (FAILED(hr))
		return hr;

	hr = dom->put_async(VARIANT_FALSE);
	if (FAILED(hr))
		return hr;

	BSTR dicomstudy = SysAllocString(L"DICOMSTUDY");
	if (dicomstudy == NULL)
		return E_OUTOFMEMORY;
	AutoSysFreeString af_dicomstudy(dicomstudy);

	AutoComPtr<IXMLDOMElement> documentElement;
	hr = dom->createElement(dicomstudy, documentElement.r());
	if (FAILED(hr))
		return hr;

	hr = dom->putref_documentElement(documentElement.p());
	if (FAILED(hr))
		return hr;

	if (dwOption == 0) {
		CDataElementCache dc;
		CInstanceCache ic;

		try {
			scan_files(studyDir, L"", dc, ic);
		} catch (HRESULT hr) {
			return hr;
		}

		hr = dc.GetDOM(dom.p());
		if (FAILED(hr))
			return hr;

		hr = ic.GetDOM(dom.p());
		if (FAILED(hr))
			return hr;
	
	} else {
		const int studyDirLen = lstrlen(studyDir);
		
		wchar_t* const dcFile = (wchar_t*)_alloca(sizeof(wchar_t) * (studyDirLen + 1 + lstrlen(CDataElementCache::filename) + 1));
		lstrcpy(dcFile, studyDir);
		lstrcat(dcFile, L"\\");
		lstrcat(dcFile, CDataElementCache::filename);

		hr = CDataElementCacheFile::LoadDOM(dcFile, dom.p());
		if (FAILED(hr))
			return hr;

		wchar_t* const icFile = (wchar_t*)_alloca(sizeof(wchar_t) * (studyDirLen + 1 + lstrlen(CInstanceCache::filename) + 1));
		lstrcpy(icFile, studyDir);
		lstrcat(icFile, L"\\");
		lstrcat(icFile, CInstanceCache::filename);

		hr = CInstanceCacheFile::LoadDOM(icFile, dom.p());
		if (FAILED(hr))
			return hr;
	}

	return dom->QueryInterface(IID_IUnknown, (void**)ppUnk);
}

//======================================================================
//
//======================================================================
HRESULT GetStudyInfo(DWORD dwOption, const wchar_t* filePath, wchar_t** values)
{
	*values = NULL;

	CStackFileMapping map(filePath);
	if (map.lastError() != NO_ERROR)
		return HRESULT_FROM_WIN32(map.lastError());

	const char* const block_begin = static_cast<const char*>(map.mapPointer());
	const char* const block_end = block_begin + map.mapSize();

	if (memcmp(block_begin + DCM_PREAMBLELENGTH, "DICM", 4) != 0)
		return E_UNEXPECTED;

	CDcmScannerF scanner(block_begin, block_end);
	tDcmElementHdr hdr;

	const char* cp = block_begin + DCM_PREAMBLELENGTH + 4;
	bool explicitVR;

	while (true) {
		int sr = scanner.NextElement(cp, hdr, true);
		if (sr != CDcmScannerF::S_CONTINUE) {
			return E_UNEXPECTED;
		}
		if (hdr.m_tag[0] != DCM_GROUPFILEMETA)
			break;
		if (hdr.m_tag[1] == DCM_TAG_ELEMENT(DCM_METATRANSFERSYNTAX)) {
			char transfer_syntax_uid[DICOM_UI_LENGTH + 1];
			memcpy(transfer_syntax_uid, cp + hdr.m_hsize, hdr.m_length);
			transfer_syntax_uid[hdr.m_length] = '\0';

			if (strcmp(transfer_syntax_uid, DICOM_TRANSFERLITTLEENDIAN) == 0) {
				explicitVR = false;
			} else if (strcmp(transfer_syntax_uid, DICOM_TRANSFERLITTLEENDIANEXPLICIT) == 0) {
				explicitVR = true;
			} else if (strcmp(transfer_syntax_uid, DICOM_TRANSFERJPEGBASELINEPROCESS1) == 0) {
				explicitVR = true;
			} else if (strcmp(transfer_syntax_uid, DICOM_TRANSFERJPEGEXTENDEDPROC2AND4) == 0) {
				explicitVR = true;
			} else if (strcmp(transfer_syntax_uid, DICOM_TRANSFERJPEGLOSSLESSPROC14) == 0) {
				explicitVR = true;
			} else if (strcmp(transfer_syntax_uid, DICOM_TRANSFERJPEGLOSSLESSPROCFIRSTORDERREDICT) == 0) {
				explicitVR = true;
			} else if (strcmp(transfer_syntax_uid, "PNG") == 0) {
				explicitVR = true;
			} else {
				return E_UNEXPECTED;
			}
		}
		cp = hdr.next_ptr(cp);
	}

	const char* const dataset_begin = cp;

	enum {
		StudyDate, StudyTime, Modality, PatientName, PatientID, StudyID, STUDY_DT, INSERT_DT, ACCESS_DT, kNumItems
	};
	enum {
		ascii, mbstr, date, time
	};
	struct {
		DCM_TAG tag;
		const char* data;
		unsigned int data_len;
		unsigned int str_len;
		unsigned int type;
	} element_tbl[kNumItems] = {
		{ 0x00080020, "", 0, 0, date }, // 0 std_date
		{ 0x00080030, "", 0, 0, time }, // 1 std_time
		{ 0x00080060, "", 0, 0, ascii }, // 2 modality
		{ 0x00100010, "", 0, 0, mbstr }, // 3 pat_name
		{ 0x00100020, "", 0, 0, ascii }, // 4 pat_id
		{ 0x00200010, "", 0, 0, ascii }, // 5 std_id
		{ 0xFFFFFFFF, "", 0, 0, ascii }, // 6 study_dt
		{ 0xFFFFFFFF, "", 0, 0, ascii }, // 7 insert_dt
		{ 0xFFFFFFFF, "", 0, 0, ascii }  // 8 access_dt
	};

	unsigned int index = 0;
	unsigned int total_len = 8;

	int r;
	while ((r = scanner.NextElement(cp, hdr, explicitVR)) == CDcmScannerF::S_CONTINUE) {
		DCM_TAG tag = hdr.ltag();

		for (unsigned int i = index; element_tbl[i].tag != 0xFFFFFFFF; i++) {
			if (tag == element_tbl[i].tag) {
				const size_t len = hdr.value_length();
				if (len > 0) {
					const unsigned char* sp = (const unsigned char*)hdr.value_ptr(cp);
					const unsigned char* ep = sp + len;

					while (sp < ep && *sp <= ' ')
						sp++;
					while (sp < ep && *(ep - 1) <= ' ')
						ep--;

					element_tbl[i].data = (char*)sp;
					element_tbl[i].data_len = ep - sp;

					if (element_tbl[i].type == ascii) {
						total_len += element_tbl[i].str_len = ep - sp;
					} else if (element_tbl[i].type == mbstr) {
						total_len += element_tbl[i].str_len = multi_byte_to_wide_char((const char*)sp, element_tbl[i].data_len, NULL, 0);
					} else if (element_tbl[i].type == date) {
						int len = 0;
						for (const unsigned char* tp = sp; tp < ep; tp++)
							if (isdigit(*tp)) 
								len++;
						total_len += element_tbl[i].str_len = len;
					} else if (element_tbl[i].type == time) {
						int len = 0;
						for (const unsigned char* tp = sp; tp < ep; tp++)
							if (isdigit(*tp) || *tp == '.')
								len++;
						total_len += element_tbl[i].str_len = len;
					}
				}

				index++;
				break;
			}
		}
	
		cp = hdr.next_ptr(cp);
	}

	char study_dt[16];
	char insert_dt[16];
	char access_dt[16];

	char* std_date = (char*)_alloca(element_tbl[StudyDate].data_len + 1);
	CopyMemory(std_date, element_tbl[StudyDate].data, element_tbl[StudyDate].data_len);
	std_date[element_tbl[StudyDate].data_len] = '\0';

	char* std_time = (char*)_alloca(element_tbl[StudyTime].data_len + 1);
	CopyMemory(std_time, element_tbl[StudyTime].data, element_tbl[StudyTime].data_len);
	std_time[element_tbl[StudyTime].data_len] = '\0';

	DateTimeToDT(std_date, std_time, study_dt);
	element_tbl[STUDY_DT].data = study_dt;
	element_tbl[STUDY_DT].data_len = strlen(study_dt);
	total_len += (element_tbl[STUDY_DT].str_len = element_tbl[STUDY_DT].data_len);

	FILETIME insert_ft;
	SYSTEMTIME local_time;
	GetLocalTime(&local_time);
	SystemTimeToFileTime(&local_time, &insert_ft);
	FileTimeToDT(&insert_ft, insert_dt);
	element_tbl[INSERT_DT].data = insert_dt;
	element_tbl[INSERT_DT].data_len = strlen(insert_dt);
	total_len += element_tbl[INSERT_DT].str_len = element_tbl[INSERT_DT].data_len;

	DateTimeToDT("", "", access_dt);
	element_tbl[ACCESS_DT].data = access_dt;
	element_tbl[ACCESS_DT].data_len = strlen(access_dt);
	total_len += element_tbl[ACCESS_DT].str_len = element_tbl[ACCESS_DT].data_len;

	wchar_t* const wstr = (wchar_t*)CoTaskMemAlloc(sizeof(wchar_t) * (total_len + 1));
	if (wstr == NULL)
		return E_OUTOFMEMORY;

	wchar_t* wp = wstr;

	for (int index = 0; index < kNumItems; index++) {
		if (index != 0)
			*wp++ = L'\t';

		const char* sp = element_tbl[index].data;
		const char* const ep = sp + element_tbl[index].data_len;

		if (element_tbl[index].type == ascii) {
			while (sp < ep)
				*wp++ = *sp++;
		} else if (element_tbl[index].type == mbstr) {
			multi_byte_to_wide_char(element_tbl[index].data, element_tbl[index].data_len, wp, element_tbl[index].str_len);
			wp += element_tbl[index].str_len;
		} else if (element_tbl[index].type == date) {
			for (; sp < ep; sp++)
				if (isdigit(*sp))
					*wp++ = *sp;
		} else if (element_tbl[index].type == time) {
			for (; sp < ep; sp++)
				if (isdigit(*sp) || *sp == '.')
					*wp++ = *sp;
		}
	}
	*wp = L'\0';

	*values = wstr;

	return S_OK;	
}

//======================================================================
//
//======================================================================
static tDWConfig* gDWConfig = NULL;
//======================================================================
//
//======================================================================
HRESULT LoadConfig(bool logon)
{
	HRESULT hr = ReadConfig(gDWConfig, NULL, false);
	if (FAILED(hr))
		return hr;

	if (logon) {
		for (int i = 0; i < gDWConfig->num_vol; i++) {
			tVOLprm* vol = gDWConfig->vol_tbl + i;
			if (vol->user != NULL) {
				DWORD err;

				err = vol->logon();
				if (err != NO_ERROR)
					return HRESULT_FROM_WIN32(err);
			}
		}
	}

	return S_OK;
}

//======================================================================
//
//======================================================================
HRESULT UnloadConfig()
{
	return ReleaseConfigData(gDWConfig);
}

//======================================================================
//
//======================================================================
HRESULT GetStudyCache2(const wchar_t* studyLoc, IUnknown** ppUnk)
{
	const tVOLprm* vol;
	const DWORD len = gDWConfig->mapStoragePath(studyLoc, NULL, &vol);
	if (len == 0)
		return HRESULT_FROM_WIN32(ERROR_FILE_NOT_FOUND);
	wchar_t* studyDir = (wchar_t*)_alloca(sizeof(wchar_t) * (len + 1));
	gDWConfig->mapStoragePath(studyLoc, studyDir);

	if (vol->hToken)
		if (ImpersonateLoggedOnUser(vol->hToken) == 0)
			return HRESULT_FROM_WIN32(GetLastError());

	const HRESULT hr = GetStudyCache(vol->study_cache_flag, studyDir, ppUnk);

	if (vol->hToken)
		RevertToSelf();

	return hr;
}

//======================================================================
//
//
//
//======================================================================
template < class CACHE_T >
class CCacheMap {
	HANDLE m_hFile;
	HANDLE m_hMap;
	LPVOID m_lpMap;
	OVERLAPPED m_ov;

	typedef CACHE_T::header_t file_header_t;

public:
	//==================================================================
	//
	//==================================================================
	CCacheMap()
		: m_lpMap(NULL)
	{
	}
	
	//==================================================================
	//
	//==================================================================
	HRESULT Map(const wchar_t* studyDir, LPVOID* map, unsigned long* size) {

		m_hFile = INVALID_HANDLE_VALUE;
		m_hMap = NULL;
		m_lpMap = NULL;
		ZeroMemory(&m_ov, sizeof(m_ov));

		wchar_t* const path = (wchar_t*)_alloca(sizeof(wchar_t) * (lstrlen(studyDir) + 1 + lstrlen(CACHE_T::filename) + 1));
		lstrcpy(path, studyDir);
		lstrcat(path, L"\\");
		lstrcat(path, CACHE_T::filename);

		HANDLE hFile = CreateFile(path, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if (hFile == INVALID_HANDLE_VALUE)
			return HRESULT_FROM_WIN32(GetLastError());

		if (GetFileSize(hFile, NULL) < sizeof(file_header_t)) {
			CloseHandle(hFile);
			return E_UNEXPECTED;
		}

		if (!LockFileEx(hFile, 0, 0, sizeof(file_header_t), 0, &m_ov)) {
			const HRESULT hr = HRESULT_FROM_WIN32(GetLastError());
			CloseHandle(hFile);
			return hr;
		}

		HANDLE hMap = CreateFileMapping(hFile, NULL, PAGE_READONLY, 0, 0, NULL);
		if (hMap == NULL) {
			const HRESULT hr = HRESULT_FROM_WIN32(GetLastError());
			UnlockFileEx(hFile, 0, sizeof(file_header_t), 0, &m_ov);
			CloseHandle(hFile);
			return hr;
		}

		LPVOID lpMap = MapViewOfFile(hMap, FILE_MAP_READ, 0, 0, 0);
		if (lpMap == NULL) {
			const HRESULT hr = HRESULT_FROM_WIN32(GetLastError());
			CloseHandle(hMap);
			UnlockFileEx(hFile, 0, sizeof(file_header_t), 0, &m_ov);
			CloseHandle(hFile);
			return hr;
		}

		const file_header_t* const header = static_cast<file_header_t*>(lpMap);
		if (memcmp(header->m_sig, CACHE_T::sig, sizeof(header->m_sig)) != 0) {
			const HRESULT hr = HRESULT_FROM_WIN32(E_UNEXPECTED);
			UnmapViewOfFile(lpMap);
			CloseHandle(hMap);
			UnlockFileEx(hFile, 0, sizeof(file_header_t), 0, &m_ov);
			CloseHandle(hFile);
			return hr;
		}
		if (GetFileSize(hFile, NULL) != header->m_headerSize + header->m_dataSize) {
			const HRESULT hr = HRESULT_FROM_WIN32(E_UNEXPECTED);
			UnmapViewOfFile(lpMap);
			CloseHandle(hMap);
			UnlockFileEx(hFile, 0, sizeof(file_header_t), 0, &m_ov);
			CloseHandle(hFile);
			return hr;
		}

		m_hFile = hFile;
		m_hMap = hMap;
		m_lpMap = lpMap;

		*map = LPVOID((char*)m_lpMap + header->m_headerSize);
		*size = header->m_dataSize;

		return S_OK;
	}

	//==================================================================
	//
	//==================================================================
	void Unmap() {
		if (m_lpMap == NULL)
			return;
		UnmapViewOfFile(m_lpMap);
		CloseHandle(m_hMap);
		UnlockFileEx(m_hFile, 0, sizeof(file_header_t), 0, &m_ov);
		CloseHandle(m_hFile);
	}
};

//======================================================================
//
//
//
//======================================================================
template < class CACHE_T >
class CCacheBuffer {
	LPVOID m_buffer;

	typedef CACHE_T::block_t block_t;

public:
	//==================================================================
	//
	//==================================================================
	CCacheBuffer()
		: m_buffer(NULL)
	{
	}
	
	//==================================================================
	//
	//==================================================================
	HRESULT Allocate(CACHE_T& cache, LPVOID* map, unsigned long* size) {

		const DWORD dataSize = cache.m_dataSize;
		const block_t* first_block = &cache.m_first_block;

		LPVOID const ptr = CoTaskMemAlloc(dataSize);
		if (ptr == NULL)
			return HRESULT_FROM_WIN32(GetLastError());
		char* cp = (char*)ptr;

		for (const block_t* block = first_block; block != NULL; block = block->m_next) {
			CopyMemory(cp, block->m_data, block->m_bytesInBlock);
			cp += block->m_bytesInBlock;
		}
	
		m_buffer = ptr;

		*map = m_buffer;
		*size = dataSize;

		return S_OK;
	}

	//==================================================================
	//
	//==================================================================
	void Deallocate() {
		if (m_buffer == NULL)
			return;
		CoTaskMemFree(m_buffer);
	}
};

//======================================================================
//
//
//
//======================================================================
class CCacheContext {
public:
	virtual HRESULT Open(const wchar_t* studyDir, LPVOID* dataElmPtr, unsigned long* dataElmSize, LPVOID* instDirPtr, unsigned long* instDirSize) = 0;
	virtual void Close() = 0;
};
//======================================================================
//
//
//
//======================================================================
class CCacheContextMap : public CCacheContext {
	CCacheMap< CDataElementCache > m_dataElm;
	CCacheMap< CInstanceCache > m_instDir;
public:
	//==================================================================
	//
	//
	//==================================================================
	CCacheContextMap()
		: m_dataElm(), m_instDir() {}

	//==================================================================
	//
	//
	//==================================================================
	virtual HRESULT Open(const wchar_t* studyDir, LPVOID* dataElmPtr, unsigned long* dataElmSize, LPVOID* instDirPtr, unsigned long* instDirSize) {

		LPVOID ptr1;
		unsigned long size1;
		HRESULT hr = m_dataElm.Map(studyDir, &ptr1, &size1);
		if (FAILED(hr))
			return hr;

		LPVOID ptr2;
		unsigned long size2;
		hr = m_instDir.Map(studyDir, &ptr2, &size2);
		if (FAILED(hr)) {
			m_dataElm.Unmap();
			return hr;
		}

		*dataElmPtr = ptr1;
		*dataElmSize = size1;
		*instDirPtr = ptr2;
		*instDirSize = size2;

		return S_OK;
	}

	//==================================================================
	//
	//
	//==================================================================
	virtual void Close() {
		m_dataElm.Unmap();
		m_instDir.Unmap();
	}

};

//======================================================================
//
//
//
//======================================================================
class CCacheContextBuffer : public CCacheContext {
	CCacheBuffer< CDataElementCache > m_dataElm;
	CCacheBuffer< CInstanceCache > m_instDir;
public:
	//==================================================================
	//
	//
	//==================================================================
	CCacheContextBuffer()
		: m_dataElm(), m_instDir() {}

	//==================================================================
	//
	//
	//==================================================================
	virtual HRESULT Open(const wchar_t* studyDir, LPVOID* dataElmPtr, unsigned long* dataElmSize, LPVOID* instDirPtr, unsigned long* instDirSize) {

		CDataElementCache dc;
		CInstanceCache ic;

		try {
			scan_files(studyDir, L"", dc, ic);
		} catch (HRESULT hr) {
			return hr;
		}

		LPVOID ptr1;
		unsigned long size1;
		HRESULT hr = m_dataElm.Allocate(dc, &ptr1, &size1);
		if (FAILED(hr))
			return hr;

		LPVOID ptr2;
		unsigned long size2;
		hr = m_instDir.Allocate(ic, &ptr2, &size2);
		if (FAILED(hr)) {
			m_dataElm.Deallocate();
			return hr;
		}

		*dataElmPtr = ptr1;
		*dataElmSize = size1;
		*instDirPtr = ptr2;
		*instDirSize = size2;

		return S_OK;
	}

	//==================================================================
	//
	//
	//==================================================================
	virtual void Close() {
		m_dataElm.Deallocate();
		m_instDir.Deallocate();
	}
};

//======================================================================
//
//======================================================================
HRESULT MapStudyCacheBuffer_(const wchar_t* studyDir, int study_cache_flag, LPVOID* dataElmPtr, unsigned long* dataElmSize, LPVOID* instDirPtr, unsigned long* instDirSize, LPVOID* ctx)
{
	CCacheContext* cache_ctx;
	if (study_cache_flag == 0)
		cache_ctx = new CCacheContextBuffer();
	else
		cache_ctx = new CCacheContextMap();

	const HRESULT hr = cache_ctx->Open(studyDir, dataElmPtr, dataElmSize, instDirPtr, instDirSize);
	if (FAILED(hr)) {
		cache_ctx->Close();
		delete cache_ctx;
		*ctx = NULL;
		return hr;
	}

	*ctx = (LPVOID)cache_ctx;

	return S_OK;
}

//======================================================================
//
//======================================================================
HRESULT MapStudyCacheBuffer(const wchar_t* studyLoc, LPVOID* dataElmPtr, unsigned long* dataElmSize, LPVOID* instDirPtr, unsigned long* instDirSize, LPVOID* ctx)
{
	*ctx = NULL;

	const tVOLprm* vol;
	const DWORD len = gDWConfig->mapStoragePath(studyLoc, NULL, &vol);
	if (len == 0)
		return HRESULT_FROM_WIN32(ERROR_FILE_NOT_FOUND);
	wchar_t* studyDir = (wchar_t*)_alloca(sizeof(wchar_t) * (len + 1));
	gDWConfig->mapStoragePath(studyLoc, studyDir);

	if (vol->hToken)
		if (ImpersonateLoggedOnUser(vol->hToken) == 0)
			return HRESULT_FROM_WIN32(GetLastError());

	const HRESULT hr = MapStudyCacheBuffer_(studyDir, vol->study_cache_flag, dataElmPtr, dataElmSize, instDirPtr, instDirSize, ctx);

	if (vol->hToken)
		RevertToSelf();

		return hr;
}

//======================================================================
//
//======================================================================
HRESULT UnmapStudyCacheBuffer(LPVOID ctx)
{
	CCacheContext* cache_ctx = (CCacheContext*)ctx;

	cache_ctx->Close();

	delete cache_ctx;

	return S_OK;
}

//======================================================================
//
//======================================================================
static bool StudyCacheBuffer_initialized = false;
//======================================================================
//
//======================================================================
HRESULT InitializeStudyCacheBuffer(bool logon)
{
	if (StudyCacheBuffer_initialized)
		return S_OK;

	const HRESULT hr = LoadConfig(logon);
	if (FAILED(hr))
		return hr;

	StudyCacheBuffer_initialized = true;
	
	return S_OK;
}

//======================================================================
//
//======================================================================
HRESULT UninitializeStudyCacheBuffer()
{
	if (StudyCacheBuffer_initialized) {
		UnloadConfig();
		StudyCacheBuffer_initialized = false;
	}

	return S_OK;
}



